@@ -10,6 +10,7 @@
 #include <osmocom/core/linuxlist.h>
 #include <osmocom/core/timer.h>
 #include <osmocom/core/utils.h>
+#include <osmocom/core/logging.h>
 
 /*! \defgroup fsm Finite State Machine abstraction
  *  @{
@@ -114,33 +115,93 @@ struct osmo_fsm_inst {
 		struct llist_head children;
 		/*! \ref llist_head linked to parent->proc.children */
 		struct llist_head child;
+		/*! Indicator whether osmo_fsm_inst_term() was already invoked on this instance. */
+		bool terminating;
 	} proc;
 };
 
 void osmo_fsm_log_addr(bool log_addr);
+void osmo_fsm_log_timeouts(bool log_timeouts);
+void osmo_fsm_term_safely(bool term_safely);
+void osmo_fsm_set_dealloc_ctx(void *ctx);
 
-#define LOGPFSML(fi, level, fmt, args...) \
-		LOGP((fi)->fsm->log_subsys, OSMO_MAX(level, (fi)->log_level), \
-			"%s{%s}: " fmt, \
-			osmo_fsm_inst_name(fi),				    \
-			osmo_fsm_state_name((fi)->fsm, (fi)->state), ## args)
+/*! Log using FSM instance's context, on explicit logging subsystem and level.
+ * \param fi  An osmo_fsm_inst.
+ * \param subsys  A logging subsystem, e.g. DLGLOBAL.
+ * \param level  A logging level, e.g. LOGL_INFO.
+ * \param fmt  printf-like format string.
+ * \param args  Format string arguments.
+ */
+#define LOGPFSMSL(fi, subsys, level, fmt, args...) \
+		LOGPFSMSLSRC(fi, subsys, level, __FILE__, __LINE__, fmt, ## args)
 
-#define LOGPFSM(fi, fmt, args...) \
-		LOGPFSML(fi, (fi)->log_level, fmt, ## args)
-
-#define LOGPFSMLSRC(fi, level, caller_file, caller_line, fmt, args...) \
-		LOGPSRC((fi)->fsm->log_subsys, level, \
+/*! Log using FSM instance's context, on explicit logging subsystem and level,
+ * and passing explicit source file and line information.
+ * \param fi  An osmo_fsm_inst.
+ * \param subsys  A logging subsystem, e.g. DLGLOBAL.
+ * \param level  A logging level, e.g. LOGL_INFO.
+ * \param caller_file  A string constant containing a source file path, like __FILE__.
+ * \param caller_line  A number constant containing a source file line, like __LINE__.
+ * \param fmt  printf-like format string.
+ * \param args  Format string arguments.
+ */
+#define LOGPFSMSLSRC(fi, subsys, level, caller_file, caller_line, fmt, args...) \
+		LOGPSRC(subsys, level, \
 			caller_file, caller_line, \
 			"%s{%s}: " fmt, \
 			osmo_fsm_inst_name(fi), \
-			osmo_fsm_state_name((fi)->fsm, (fi)->state), \
-			## args)
+			(fi) ? osmo_fsm_state_name((fi)->fsm, (fi)->state) : "fi=NULL", ## args)
 
+
+/*! Log using FSM instance's context, on explicit logging level.
+ * \param fi  An osmo_fsm_inst.
+ * \param level  A logging level, e.g. LOGL_INFO.
+ * \param fmt  printf-like format string.
+ * \param args  Format string arguments.
+ */
+#define LOGPFSML(fi, level, fmt, args...) \
+		LOGPFSMLSRC(fi, level, __FILE__, __LINE__, fmt, ## args)
+
+/*! Log using FSM instance's context, on explicit logging level, and with explicit source file and line info.
+ * The log subsystem to log on is obtained from the underlying FSM definition.
+ * \param fi  An osmo_fsm_inst.
+ * \param level  A logging level, e.g. LOGL_INFO.
+ * \param caller_file  A string constant containing a source file path, like __FILE__.
+ * \param caller_line  A number constant containing a source file line, like __LINE__.
+ * \param fmt  printf-like format string.
+ * \param args  Format string arguments.
+ */
+#define LOGPFSMLSRC(fi, level, caller_file, caller_line, fmt, args...) \
+		LOGPFSMSLSRC(fi, (fi) ? (fi)->fsm->log_subsys : DLGLOBAL, level, \
+			     caller_file, caller_line, fmt, ## args)
+
+/*! Log using FSM instance's context.
+ * The log level to log on is obtained from the FSM instance.
+ * The log subsystem to log on is obtained from the underlying FSM definition.
+ * \param fi  An osmo_fsm_inst.
+ * \param fmt  printf-like format string.
+ * \param args  Format string arguments.
+ */
+#define LOGPFSM(fi, fmt, args...) \
+		LOGPFSML(fi, (fi) ? (fi)->log_level : LOGL_ERROR, fmt, ## args)
+
+/*! Log using FSM instance's context, with explicit source file and line info.
+ * The log level to log on is obtained from the FSM instance.
+ * The log subsystem to log on is obtained from the underlying FSM definition.
+ * \param fi  An osmo_fsm_inst.
+ * \param caller_file  A string constant containing a source file path, like __FILE__.
+ * \param caller_line  A number constant containing a source file line, like __LINE__.
+ * \param fmt  printf-like format string.
+ * \param args  Format string arguments.
+ */
 #define LOGPFSMSRC(fi, caller_file, caller_line, fmt, args...) \
-		LOGPFSMLSRC(fi, (fi)->log_level, \
+		LOGPFSMLSRC(fi, (fi) ? (fi)->log_level : LOGL_ERROR, \
 			    caller_file, caller_line, \
 			    fmt, ## args)
 
+#define OSMO_T_FMT "%c%u"
+#define OSMO_T_FMT_ARGS(T) ((T) >= 0 ? 'T' : 'X'), ((T) >= 0 ? T : -T)
+
 int osmo_fsm_register(struct osmo_fsm *fsm);
 void osmo_fsm_unregister(struct osmo_fsm *fsm);
 struct osmo_fsm *osmo_fsm_find_by_name(const char *name);
@@ -161,6 +222,7 @@ void osmo_fsm_inst_free(struct osmo_fsm_inst *fi);
 
 int osmo_fsm_inst_update_id(struct osmo_fsm_inst *fi, const char *id);
 int osmo_fsm_inst_update_id_f(struct osmo_fsm_inst *fi, const char *fmt, ...);
+int osmo_fsm_inst_update_id_f_sanitize(struct osmo_fsm_inst *fi, char replace_with, const char *fmt, ...);
 
 const char *osmo_fsm_event_name(struct osmo_fsm *fsm, uint32_t event);
 const char *osmo_fsm_inst_name(struct osmo_fsm_inst *fi);
@@ -168,7 +230,7 @@ const char *osmo_fsm_state_name(struct osmo_fsm *fsm, uint32_t state);
 
 /*! return the name of the state the FSM instance is currently in. */
 static inline const char *osmo_fsm_inst_state_name(struct osmo_fsm_inst *fi)
-{ return osmo_fsm_state_name(fi->fsm, fi->state); }
+{ return fi ? osmo_fsm_state_name(fi->fsm, fi->state) : "NULL"; }
 
 /*! perform a state change of the given FSM instance
  *
@@ -178,11 +240,18 @@ static inline const char *osmo_fsm_inst_state_name(struct osmo_fsm_inst *fi)
  */
 #define osmo_fsm_inst_state_chg(fi, new_state, timeout_secs, T) \
 	_osmo_fsm_inst_state_chg(fi, new_state, timeout_secs, T, \
-				 __BASE_FILE__, __LINE__)
+				 __FILE__, __LINE__)
 int _osmo_fsm_inst_state_chg(struct osmo_fsm_inst *fi, uint32_t new_state,
 			     unsigned long timeout_secs, int T,
 			     const char *file, int line);
 
+#define osmo_fsm_inst_state_chg_ms(fi, new_state, timeout_ms, T) \
+	_osmo_fsm_inst_state_chg_ms(fi, new_state, timeout_ms, T, \
+				 __FILE__, __LINE__)
+int _osmo_fsm_inst_state_chg_ms(struct osmo_fsm_inst *fi, uint32_t new_state,
+				unsigned long timeout_ms, int T,
+				const char *file, int line);
+
 /*! perform a state change while keeping the current timer running.
  *
  *  This is useful to keep a timeout across several states (without having to round the
@@ -194,10 +263,33 @@ int _osmo_fsm_inst_state_chg(struct osmo_fsm_inst *fi, uint32_t new_state,
  */
 #define osmo_fsm_inst_state_chg_keep_timer(fi, new_state) \
 	_osmo_fsm_inst_state_chg_keep_timer(fi, new_state, \
-				 __BASE_FILE__, __LINE__)
+				 __FILE__, __LINE__)
 int _osmo_fsm_inst_state_chg_keep_timer(struct osmo_fsm_inst *fi, uint32_t new_state,
 					const char *file, int line);
 
+/*! perform a state change while keeping the current timer if running, or starting a timer otherwise.
+ *
+ *  This is useful to keep a timeout across several states, but to make sure that some timeout is actually running.
+ *
+ *  This is a macro that calls _osmo_fsm_inst_state_chg_keep_or_start_timer() with the given
+ *  parameters as well as the caller's source file and line number for logging
+ *  purposes. See there for documentation.
+ */
+#define osmo_fsm_inst_state_chg_keep_or_start_timer(fi, new_state, timeout_secs, T) \
+	_osmo_fsm_inst_state_chg_keep_or_start_timer(fi, new_state, timeout_secs, T, \
+						     __FILE__, __LINE__)
+int _osmo_fsm_inst_state_chg_keep_or_start_timer(struct osmo_fsm_inst *fi, uint32_t new_state,
+						 unsigned long timeout_secs, int T,
+						 const char *file, int line);
+
+#define osmo_fsm_inst_state_chg_keep_or_start_timer_ms(fi, new_state, timeout_ms, T) \
+	_osmo_fsm_inst_state_chg_keep_or_start_timer_ms(fi, new_state, timeout_ms, T, \
+						     __FILE__, __LINE__)
+int _osmo_fsm_inst_state_chg_keep_or_start_timer_ms(struct osmo_fsm_inst *fi, uint32_t new_state,
+						   unsigned long timeout_ms, int T,
+						   const char *file, int line);
+
+
 /*! dispatch an event to an osmocom finite state machine instance
  *
  *  This is a macro that calls _osmo_fsm_inst_dispatch() with the given
@@ -205,7 +297,7 @@ int _osmo_fsm_inst_state_chg_keep_timer(struct osmo_fsm_inst *fi, uint32_t new_s
  *  purposes. See there for documentation.
  */
 #define osmo_fsm_inst_dispatch(fi, event, data) \
-	_osmo_fsm_inst_dispatch(fi, event, data, __BASE_FILE__, __LINE__)
+	_osmo_fsm_inst_dispatch(fi, event, data, __FILE__, __LINE__)
 int _osmo_fsm_inst_dispatch(struct osmo_fsm_inst *fi, uint32_t event, void *data,
 			    const char *file, int line);
 
@@ -216,7 +308,7 @@ int _osmo_fsm_inst_dispatch(struct osmo_fsm_inst *fi, uint32_t event, void *data
  *  See there for documentation.
  */
 #define osmo_fsm_inst_term(fi, cause, data) \
-	_osmo_fsm_inst_term(fi, cause, data, __BASE_FILE__, __LINE__)
+	_osmo_fsm_inst_term(fi, cause, data, __FILE__, __LINE__)
 void _osmo_fsm_inst_term(struct osmo_fsm_inst *fi,
 			 enum osmo_fsm_term_cause cause, void *data,
 			 const char *file, int line);
@@ -228,7 +320,7 @@ void _osmo_fsm_inst_term(struct osmo_fsm_inst *fi,
  *  purposes. See there for documentation.
  */
 #define osmo_fsm_inst_term_children(fi, cause, data) \
-	_osmo_fsm_inst_term_children(fi, cause, data, __BASE_FILE__, __LINE__)
+	_osmo_fsm_inst_term_children(fi, cause, data, __FILE__, __LINE__)
 void _osmo_fsm_inst_term_children(struct osmo_fsm_inst *fi,
 				  enum osmo_fsm_term_cause cause,
 				  void *data,
