@@ -33,7 +33,7 @@
 #define GSMTAP_TYPE_UM		0x01
 #define GSMTAP_TYPE_ABIS	0x02
 #define GSMTAP_TYPE_UM_BURST	0x03	/* raw burst bits */
-#define GSMTAP_TYPE_SIM		0x04
+#define GSMTAP_TYPE_SIM		0x04	/* ISO 7816 smart card interface */
 #define GSMTAP_TYPE_TETRA_I1		0x05	/* tetra air interface */
 #define GSMTAP_TYPE_TETRA_I1_BURST	0x06	/* tetra air interface */
 #define GSMTAP_TYPE_WMX_BURST	0x07    /* WiMAX burst */
@@ -82,8 +82,8 @@
 #define GSMTAP_CHANNEL_SDCCH	0x06
 #define GSMTAP_CHANNEL_SDCCH4	0x07
 #define GSMTAP_CHANNEL_SDCCH8	0x08
-#define GSMTAP_CHANNEL_TCH_F	0x09
-#define GSMTAP_CHANNEL_TCH_H	0x0a
+#define GSMTAP_CHANNEL_FACCH_F	0x09	/* Actually, it's FACCH/F (signaling) */
+#define GSMTAP_CHANNEL_FACCH_H	0x0a	/* Actually, it's FACCH/H (signaling) */
 #define GSMTAP_CHANNEL_PACCH	0x0b
 #define GSMTAP_CHANNEL_CBCH52	0x0c
 #define GSMTAP_CHANNEL_PDTCH	0x0d
@@ -91,6 +91,10 @@
 #define GSMTAP_CHANNEL_PDCH	GSMTAP_CHANNEL_PDTCH
 #define GSMTAP_CHANNEL_PTCCH	0x0e
 #define GSMTAP_CHANNEL_CBCH51	0x0f
+#define GSMTAP_CHANNEL_VOICE_F	0x10	/* voice codec payload (FR/EFR/AMR) */
+#define GSMTAP_CHANNEL_VOICE_H	0x11	/* voice codec payload (HR/AMR) */
+#define GSMTAP_CHANNEL_TCH_F	GSMTAP_CHANNEL_FACCH_F	/* We used the wrong naming in 2008 when we were young */
+#define GSMTAP_CHANNEL_TCH_H	GSMTAP_CHANNEL_FACCH_H	/* We used the wrong naming in 2008 when we were young */
 
 /* GPRS Coding Scheme CS1..4 */
 #define GSMTAP_GPRS_CS_BASE	0x20
@@ -103,6 +107,18 @@
 
 /* ====== DO NOT MAKE UNAPPROVED MODIFICATIONS HERE ===== */
 
+/* sub-types for GSMTAP_TYPE_SIM */
+#define GSMTAP_SIM_APDU		0x00 /* APDU data (complete APDU) */
+#define GSMTAP_SIM_ATR		0x01 /* card ATR data */
+#define GSMTAP_SIM_PPS_REQ		0x02 /* PPS request data */
+#define GSMTAP_SIM_PPS_RSP		0x03 /* PPS response data */
+#define GSMTAP_SIM_TPDU_HDR		0x04 /* TPDU command header */
+#define GSMTAP_SIM_TPDU_CMD		0x05 /* TPDU command body */
+#define GSMTAP_SIM_TPDU_RSP		0x06 /* TPDU response body */
+#define GSMTAP_SIM_TPDU_SW		0x07 /* TPDU response trailer */
+
+/* ====== DO NOT MAKE UNAPPROVED MODIFICATIONS HERE ===== */
+
 /* sub-types for TYPE_TETRA_AIR */
 #define GSMTAP_TETRA_BSCH	0x01
 #define GSMTAP_TETRA_AACH	0x02
@@ -233,20 +249,43 @@ enum {
 
 /* LTE RRC message types */
 enum {
-    GSMTAP_LTE_RRC_SUB_DL_CCCH_Message = 0,
-    GSMTAP_LTE_RRC_SUB_DL_DCCH_Message,
-    GSMTAP_LTE_RRC_SUB_UL_CCCH_Message,
-    GSMTAP_LTE_RRC_SUB_UL_DCCH_Message,
-    GSMTAP_LTE_RRC_SUB_BCCH_BCH_Message,
-    GSMTAP_LTE_RRC_SUB_BCCH_DL_SCH_Message,
-    GSMTAP_LTE_RRC_SUB_PCCH_Message,
-    GSMTAP_LTE_RRC_SUB_MCCH_Message,
+	GSMTAP_LTE_RRC_SUB_DL_CCCH_Message = 0,
+	GSMTAP_LTE_RRC_SUB_DL_DCCH_Message,
+	GSMTAP_LTE_RRC_SUB_UL_CCCH_Message,
+	GSMTAP_LTE_RRC_SUB_UL_DCCH_Message,
+	GSMTAP_LTE_RRC_SUB_BCCH_BCH_Message,
+	GSMTAP_LTE_RRC_SUB_BCCH_DL_SCH_Message,
+	GSMTAP_LTE_RRC_SUB_PCCH_Message,
+	GSMTAP_LTE_RRC_SUB_MCCH_Message,
+	GSMTAP_LTE_RRC_SUB_BCCH_BCH_Message_MBMS,
+	GSMTAP_LTE_RRC_SUB_BCCH_DL_SCH_Message_BR,
+	GSMTAP_LTE_RRC_SUB_BCCH_DL_SCH_Message_MBMS,
+	GSMTAP_LTE_RRC_SUB_SC_MCCH_Message,
+	GSMTAP_LTE_RRC_SUB_SBCCH_SL_BCH_Message,
+	GSMTAP_LTE_RRC_SUB_SBCCH_SL_BCH_Message_V2X,
+	GSMTAP_LTE_RRC_SUB_DL_CCCH_Message_NB,
+	GSMTAP_LTE_RRC_SUB_DL_DCCH_Message_NB,
+	GSMTAP_LTE_RRC_SUB_UL_CCCH_Message_NB,
+	GSMTAP_LTE_RRC_SUB_UL_DCCH_Message_NB,
+	GSMTAP_LTE_RRC_SUB_BCCH_BCH_Message_NB,
+	GSMTAP_LTE_RRC_SUB_BCCH_BCH_Message_TDD_NB,
+	GSMTAP_LTE_RRC_SUB_BCCH_DL_SCH_Message_NB,
+	GSMTAP_LTE_RRC_SUB_PCCH_Message_NB,
+	GSMTAP_LTE_RRC_SUB_SC_MCCH_Message_NB,
 
-    GSMTAP_LTE_RRC_SUB_MAX
+	GSMTAP_LTE_RRC_SUB_MAX
+};
+
+/* LTE NAS message types */
+enum {
+	GSMTAP_LTE_NAS_PLAIN = 0,
+	GSMTAP_LTE_NAS_SEC_HEADER,
+
+	GSMTAP_LTE_NAS_SUB_MAX
 };
 
 /* ====== DO NOT MAKE UNAPPROVED MODIFICATIONS HERE ===== */
-/*! Structure of the GTMTAP pseudo-header */
+/*! Structure of the GSMTAP pseudo-header */
 struct gsmtap_hdr {
 	uint8_t version;	/*!< version, set to 0x01 currently */
 	uint8_t hdr_len;	/*!< length in number of 32bit words */
@@ -266,7 +305,7 @@ struct gsmtap_hdr {
 
 } __attribute__((packed));
 
-/*! Structure of the GTMTAP libosmocore logging header */
+/*! Structure of the GSMTAP libosmocore logging header */
 struct gsmtap_osmocore_log_hdr {
 	struct {
 		uint32_t sec;
@@ -283,3 +322,26 @@ struct gsmtap_osmocore_log_hdr {
 		uint32_t line_nr;/*!< line number */
 	} src_file;
 } __attribute__((packed));
+
+/*! First byte of type==GSMTAP_TYPE_UM sub_type==GSMTAP_CHANNEL_VOICE payload */
+enum gsmtap_um_voice_type {
+	/*! 1 byte TOC + 112 bits (14 octets) = 15 octets payload;
+	 *  Reference is RFC5993 Section 5.2.1 + 3GPP TS 46.030 Annex B */
+	GSMTAP_UM_VOICE_HR,
+	/*! 33 payload bytes; Reference is RFC3551 Section 4.5.8.1 */
+	GSMTAP_UM_VOICE_FR,
+	/*! 31 payload bytes; Reference is RFC3551 Section 4.5.9 + ETSI TS 101 318 */
+	GSMTAP_UM_VOICE_EFR,
+	/*! 1 byte TOC + 5..31 bytes = 6..32 bytes payload; RFC4867 octet-aligned */
+	GSMTAP_UM_VOICE_AMR,
+	/* TODO: Revisit the types below; their usage; ... */
+	GSMTAP_UM_VOICE_AMR_SID_BAD,
+	GSMTAP_UM_VOICE_AMR_ONSET,
+	GSMTAP_UM_VOICE_AMR_RATSCCH,
+	GSMTAP_UM_VOICE_AMR_SID_UPDATE_INH,
+	GSMTAP_UM_VOICE_AMR_SID_FIRST_P1,
+	GSMTAP_UM_VOICE_AMR_SID_FIRST_P2,
+	GSMTAP_UM_VOICE_AMR_SID_FIRST_INH,
+	GSMTAP_UM_VOICE_AMR_RATSCCH_MARKER,
+	GSMTAP_UM_VOICE_AMR_RATSCCH_DATA,
+};
