@@ -16,7 +16,9 @@
 /*! Maximum number of logging filters */
 #define LOG_MAX_FILTERS	8
 
+#ifndef DEBUG
 #define DEBUG
+#endif
 
 #ifdef DEBUG
 /*! Log a debug message through the Osmocom logging framework
@@ -55,17 +57,17 @@ void logp(int subsys, const char *file, int line, int cont, const char *format,
 #define LOGPC(ss, level, fmt, args...) \
 	do { \
 		if (log_check_level(ss, level)) \
-			logp2(ss, level, __BASE_FILE__, __LINE__, 1, fmt, ##args); \
+			logp2(ss, level, __FILE__, __LINE__, 1, fmt, ##args); \
 	} while(0)
 
 /*! Log through the Osmocom logging framework with explicit source.
- *  If caller_file is passed as NULL, __BASE_FILE__ and __LINE__ are used
+ *  If caller_file is passed as NULL, __FILE__ and __LINE__ are used
  *  instead of caller_file and caller_line (so that this macro here defines
  *  both cases in the same place, and to catch cases where callers fail to pass
  *  a non-null filename string).
  *  \param[in] ss logging subsystem (e.g. \ref DLGLOBAL)
  *  \param[in] level logging level (e.g. \ref LOGL_NOTICE)
- *  \param[in] caller_file caller's source file string (e.g. __BASE_FILE__)
+ *  \param[in] caller_file caller's source file string (e.g. __FILE__)
  *  \param[in] caller_line caller's source line nr (e.g. __LINE__)
  *  \param[in] fmt format string
  *  \param[in] args variable argument list
@@ -74,13 +76,13 @@ void logp(int subsys, const char *file, int line, int cont, const char *format,
 	LOGPSRCC(ss, level, caller_file, caller_line, 0, fmt, ##args)
 
 /*! Log through the Osmocom logging framework with explicit source.
- *  If caller_file is passed as NULL, __BASE_FILE__ and __LINE__ are used
+ *  If caller_file is passed as NULL, __FILE__ and __LINE__ are used
  *  instead of caller_file and caller_line (so that this macro here defines
  *  both cases in the same place, and to catch cases where callers fail to pass
  *  a non-null filename string).
  *  \param[in] ss logging subsystem (e.g. \ref DLGLOBAL)
  *  \param[in] level logging level (e.g. \ref LOGL_NOTICE)
- *  \param[in] caller_file caller's source file string (e.g. __BASE_FILE__)
+ *  \param[in] caller_file caller's source file string (e.g. __FILE__)
  *  \param[in] caller_line caller's source line nr (e.g. __LINE__)
  *  \param[in] cont continuation (1) or new line (0)
  *  \param[in] fmt format string
@@ -92,7 +94,7 @@ void logp(int subsys, const char *file, int line, int cont, const char *format,
 			if (caller_file) \
 				logp2(ss, level, caller_file, caller_line, cont, fmt, ##args); \
 			else \
-				logp2(ss, level, __BASE_FILE__, __LINE__, cont, fmt, ##args); \
+				logp2(ss, level, __FILE__, __LINE__, cont, fmt, ##args); \
 		}\
 	} while(0)
 
@@ -103,7 +105,7 @@ void logp(int subsys, const char *file, int line, int cont, const char *format,
 #define LOGL_ERROR	7	/*!< error condition, requires user action */
 #define LOGL_FATAL	8	/*!< fatal, program aborted */
 
-/* logging levels defined by the library itself */
+/* logging subsystems defined by the library itself */
 #define DLGLOBAL	-1	/*!< global logging */
 #define DLLAPD		-2	/*!< LAPD implementation */
 #define DLINP		-3	/*!< (A-bis) Input sub-system */
@@ -122,7 +124,27 @@ void logp(int subsys, const char *file, int line, int cont, const char *format,
 #define DLM3UA		-16	/*!< Osmocom M3UA */
 #define DLMGCP		-17	/*!< Osmocom MGCP */
 #define DLJIBUF		-18	/*!< Osmocom Jitter Buffer */
-#define OSMO_NUM_DLIB	18	/*!< Number of logging sub-systems in libraries */
+#define DLRSPRO		-19	/*!< Osmocom Remote SIM Protocol */
+#define OSMO_NUM_DLIB	19	/*!< Number of logging sub-systems in libraries */
+
+/* Colors that can be used in log_info_cat.color */
+#define OSMO_LOGCOLOR_NORMAL NULL
+#define OSMO_LOGCOLOR_RED "\033[1;31m"
+#define OSMO_LOGCOLOR_GREEN "\033[1;32m"
+#define OSMO_LOGCOLOR_YELLOW "\033[1;33m"
+#define OSMO_LOGCOLOR_BLUE "\033[1;34m"
+#define OSMO_LOGCOLOR_PURPLE "\033[1;35m"
+#define OSMO_LOGCOLOR_CYAN "\033[1;36m"
+#define OSMO_LOGCOLOR_DARKRED "\033[31m"
+#define OSMO_LOGCOLOR_DARKGREEN "\033[32m"
+#define OSMO_LOGCOLOR_DARKYELLOW "\033[33m"
+#define OSMO_LOGCOLOR_DARKBLUE "\033[34m"
+#define OSMO_LOGCOLOR_DARKPURPLE "\033[35m"
+#define OSMO_LOGCOLOR_DARKCYAN "\033[36m"
+#define OSMO_LOGCOLOR_DARKGREY "\033[1;30m"
+#define OSMO_LOGCOLOR_GREY "\033[37m"
+#define OSMO_LOGCOLOR_BRIGHTWHITE "\033[1;37m"
+#define OSMO_LOGCOLOR_END "\033[0;m"
 
 /*! Configuration of single log category / sub-system */
 struct log_category {
@@ -151,6 +173,7 @@ enum log_ctx_index {
 	LOG_CTX_GB_BVC,
 	LOG_CTX_BSC_SUBSCR,
 	LOG_CTX_VLR_SUBSCR,
+	LOG_CTX_L1_SAPI,
 	_LOG_CTX_COUNT
 };
 
@@ -163,6 +186,7 @@ enum log_filter_index {
 	LOG_FLT_GB_BVC,
 	LOG_FLT_BSC_SUBSCR,
 	LOG_FLT_VLR_SUBSCR,
+	LOG_FLT_L1_SAPI,
 	_LOG_FLT_COUNT
 };
 
@@ -228,6 +252,12 @@ enum log_filename_type {
 	LOG_FILENAME_BASENAME,
 };
 
+/*! Where on a log line source file and line should be logged. */
+enum log_filename_pos {
+	LOG_FILENAME_POS_HEADER_END,
+	LOG_FILENAME_POS_LINE_END,
+};
+
 /*! structure representing a logging target */
 struct log_target {
         struct llist_head entry;		/*!< linked list */
@@ -313,6 +343,8 @@ struct log_target {
 	bool print_category_hex;
 	/* Should we print the source file and line, and in which way? */
 	enum log_filename_type print_filename2;
+	/* Where on a log line to put the source file info. */
+	enum log_filename_pos print_filename_pos;
 };
 
 /* use the above macros */
@@ -335,6 +367,7 @@ void log_set_print_extended_timestamp(struct log_target *target, int);
 void log_set_print_timestamp(struct log_target *target, int);
 void log_set_print_filename(struct log_target *target, int);
 void log_set_print_filename2(struct log_target *target, enum log_filename_type lft);
+void log_set_print_filename_pos(struct log_target *target, enum log_filename_pos pos);
 void log_set_print_category(struct log_target *target, int);
 void log_set_print_category_hex(struct log_target *target, int);
 void log_set_print_level(struct log_target *target, int);
@@ -364,11 +397,20 @@ int log_targets_reopen(void);
 void log_add_target(struct log_target *target);
 void log_del_target(struct log_target *target);
 
-/* Generate command string for VTY use */
-const char *log_vty_command_string() OSMO_DEPRECATED_OUTSIDE_LIBOSMOCORE;
-const char *log_vty_command_description() OSMO_DEPRECATED_OUTSIDE_LIBOSMOCORE;
-
 struct log_target *log_target_find(int type, const char *fname);
-extern struct llist_head osmo_log_target_list;
+
+void log_enable_multithread(void);
+
+void log_tgt_mutex_lock_impl(void);
+void log_tgt_mutex_unlock_impl(void);
+#define LOG_MTX_DEBUG 0
+#if LOG_MTX_DEBUG
+	#include <pthread.h>
+	#define log_tgt_mutex_lock() do { fprintf(stderr, "[%lu] %s:%d [%s] lock\n", pthread_self(), __FILE__, __LINE__, __func__); log_tgt_mutex_lock_impl(); } while (0)
+	#define log_tgt_mutex_unlock() do { fprintf(stderr, "[%lu] %s:%d [%s] unlock\n", pthread_self(), __FILE__, __LINE__, __func__); log_tgt_mutex_unlock_impl(); } while (0)
+#else
+	#define log_tgt_mutex_lock() log_tgt_mutex_lock_impl()
+	#define log_tgt_mutex_unlock() log_tgt_mutex_unlock_impl()
+#endif
 
 /*! @} */
