@@ -60,6 +60,7 @@ struct msgb {
 	unsigned char _data[0]; /*!< optional immediate data array */
 };
 
+extern struct msgb *msgb_alloc_c(const void *ctx, uint16_t size, const char *name);
 extern struct msgb *msgb_alloc(uint16_t size, const char *name);
 extern void msgb_free(struct msgb *m);
 extern void msgb_enqueue(struct llist_head *queue, struct msgb *msg);
@@ -67,9 +68,12 @@ extern struct msgb *msgb_dequeue(struct llist_head *queue);
 extern void msgb_reset(struct msgb *m);
 uint16_t msgb_length(const struct msgb *msg);
 extern const char *msgb_hexdump(const struct msgb *msg);
+char *msgb_hexdump_buf(char *buf, size_t buf_len, const struct msgb *msg);
+char *msgb_hexdump_c(const void *ctx, const struct msgb *msg);
 extern int msgb_resize_area(struct msgb *msg, uint8_t *area,
 	int old_size, int new_size);
 extern struct msgb *msgb_copy(const struct msgb *msg, const char *name);
+extern struct msgb *msgb_copy_c(const void *ctx, const struct msgb *msg, const char *name);
 static int msgb_test_invariant(const struct msgb *msg) __attribute__((pure));
 
 /*! Free all msgbs from a queue built with msgb_enqueue().
@@ -81,6 +85,40 @@ static inline void msgb_queue_free(struct llist_head *queue)
 	while ((msg = msgb_dequeue(queue))) msgb_free(msg);
 }
 
+/*! Enqueue message buffer to tail of a queue and increment queue size counter
+ * \param[in] queue linked list header of queue
+ * \param[in] msg message buffer to be added to the queue
+ * \param[in] count pointer to variable holding size of the queue
+ *
+ * The function will append the specified message buffer \a msg to the queue
+ * implemented by \ref llist_head \a queue using function \ref msgb_enqueue_count,
+ * then increment \a count
+ */
+static inline void msgb_enqueue_count(struct llist_head *queue, struct msgb *msg,
+					unsigned int *count)
+{
+	msgb_enqueue(queue, msg);
+	(*count)++;
+}
+
+/*! Dequeue message buffer from head of queue and decrement queue size counter
+ * \param[in] queue linked list header of queue
+ * \param[in] count pointer to variable holding size of the queue
+ * \returns message buffer (if any) or NULL if queue empty
+ *
+ * The function will remove the first message buffer from the queue
+ * implemented by \ref llist_head \a queue using function \ref msgb_enqueue_count,
+ * and decrement \a count, all if queue is not empty.
+ */
+static inline struct msgb *msgb_dequeue_count(struct llist_head *queue,
+						unsigned int *count)
+{
+	struct msgb *msg = msgb_dequeue(queue);
+	if (msg)
+		(*count)--;
+	return msg;
+}
+
 #ifdef MSGB_DEBUG
 #include <osmocom/core/panic.h>
 #define MSGB_ABORT(msg, fmt, args ...) do {		\
@@ -96,8 +134,10 @@ static inline void msgb_queue_free(struct llist_head *queue)
 #define msgb_l2(m)	((void *)(m->l2h))
 /*! obtain L3 header of msgb */
 #define msgb_l3(m)	((void *)(m->l3h))
+/*! obtain L4 header of msgb */
+#define msgb_l4(m)	((void *)(m->l4h))
 /*! obtain SMS header of msgb */
-#define msgb_sms(m)	((void *)(m->l4h))
+#define msgb_sms(m)	msgb_l4(m)
 
 /*! determine length of L1 message
  *  \param[in] msgb message buffer
@@ -135,6 +175,18 @@ static inline unsigned int msgb_l3len(const struct msgb *msgb)
 	return msgb->tail - (uint8_t *)msgb_l3(msgb);
 }
 
+/*! determine length of L4 message
+ *  \param[in] msgb message buffer
+ *  \returns size of L4 message in bytes
+ *
+ * This function computes the number of bytes between the tail of the
+ * message and the layer 4 header.
+ */
+static inline unsigned int msgb_l4len(const struct msgb *msgb)
+{
+	return msgb->tail - (uint8_t *)msgb_sms(msgb);
+}
+
 /*! determine the length of the header
  *  \param[in] msgb message buffer
  *  \returns number of bytes between start of buffer and start of msg
@@ -187,7 +239,11 @@ static inline unsigned char *msgb_put(struct msgb *msgb, unsigned int len)
 {
 	unsigned char *tmp = msgb->tail;
 	if (msgb_tailroom(msgb) < (int) len)
-		MSGB_ABORT(msgb, "Not enough tailroom msgb_put (%u < %u)\n",
+		MSGB_ABORT(msgb, "Not enough tailroom msgb_put"
+			   " (allocated %u, head at %u, len %u, tailroom %u < want tailroom %u)\n",
+			   msgb->data_len - sizeof(struct msgb),
+			   msgb->head - msgb->_data,
+			   msgb->len,
 			   msgb_tailroom(msgb), len);
 	msgb->tail += len;
 	msgb->len += len;
@@ -283,15 +339,20 @@ static inline uint32_t msgb_get_u32(struct msgb *msgb)
 static inline unsigned char *msgb_push(struct msgb *msgb, unsigned int len)
 {
 	if (msgb_headroom(msgb) < (int) len)
-		MSGB_ABORT(msgb, "Not enough headroom msgb_push (%u < %u)\n",
-			   msgb_headroom(msgb), len);
+		MSGB_ABORT(msgb, "Not enough headroom msgb_push"
+			   " (allocated %u, head at %u < want headroom %u, len %u, tailroom %u)\n",
+			   msgb->data_len - sizeof(struct msgb),
+			   msgb->head - msgb->_data,
+			   len,
+			   msgb->len,
+			   msgb_tailroom(msgb));
 	msgb->data -= len;
 	msgb->len += len;
 	return msgb->data;
 }
 
 /*! prepend a uint8 value to the head of the message
- *  \param[in] msgb message buffer
+ *  \param[in] msg message buffer
  *  \param[in] word unsigned 8bit byte to be prepended
  */
 static inline void msgb_push_u8(struct msgb *msg, uint8_t word)
@@ -301,7 +362,7 @@ static inline void msgb_push_u8(struct msgb *msg, uint8_t word)
 }
 
 /*! prepend a uint16 value to the head of the message
- *  \param[in] msgb message buffer
+ *  \param[in] msg message buffer
  *  \param[in] word unsigned 16bit byte to be prepended
  */
 static inline void msgb_push_u16(struct msgb *msg, uint16_t word)
@@ -311,7 +372,7 @@ static inline void msgb_push_u16(struct msgb *msg, uint16_t word)
 }
 
 /*! prepend a uint32 value to the head of the message
- *  \param[in] msgb message buffer
+ *  \param[in] msg message buffer
  *  \param[in] word unsigned 32bit byte to be prepended
  */
 static inline void msgb_push_u32(struct msgb *msg, uint32_t word)
@@ -320,6 +381,15 @@ static inline void msgb_push_u32(struct msgb *msg, uint32_t word)
 	osmo_store32be(word, space);
 }
 
+static inline unsigned char *msgb_push_tl(struct msgb *msgb, uint8_t tag)
+{
+	uint8_t *data = msgb_push(msgb, 2);
+
+	data[0] = tag;
+	data[1] = msgb->len - 2;
+	return data;
+}
+
 /*! remove (pull) a header from the front of the message buffer
  *  \param[in] msgb message buffer
  *  \param[in] len number of octets to be pulled
@@ -339,7 +409,7 @@ static inline unsigned char *msgb_pull(struct msgb *msgb, unsigned int len)
 }
 
 /*! remove (pull) all headers in front of l3h from the message buffer.
- *  \param[in] msgb message buffer with a valid l3h
+ *  \param[in] msg message buffer with a valid l3h
  *  \returns pointer to new start of msgb (l3h)
  *
  * This function moves the \a data pointer of the \ref msgb further back
@@ -354,7 +424,7 @@ static inline unsigned char *msgb_pull_to_l3(struct msgb *msg)
 }
 
 /*! remove (pull) all headers in front of l2h from the message buffer.
- *  \param[in] msgb message buffer with a valid l2h
+ *  \param[in] msg message buffer with a valid l2h
  *  \returns pointer to new start of msgb (l2h)
  *
  * This function moves the \a data pointer of the \ref msgb further back
@@ -443,6 +513,29 @@ static inline int msgb_l3trim(struct msgb *msg, int l3len)
 	return msgb_trim(msg, (msg->l3h - msg->data) + l3len);
 }
 
+/*! Allocate message buffer with specified headroom from specified talloc context.
+ *  \param[in] ctx talloc context from which to allocate
+ *  \param[in] size size in bytes, including headroom
+ *  \param[in] headroom headroom in bytes
+ *  \param[in] name human-readable name
+ *  \returns allocated message buffer with specified headroom
+ *
+ * This function is a convenience wrapper around \ref msgb_alloc
+ * followed by \ref msgb_reserve in order to create a new \ref msgb with
+ * user-specified amount of headroom.
+ */
+static inline struct msgb *msgb_alloc_headroom_c(const void *ctx, int size, int headroom,
+						 const char *name)
+{
+	osmo_static_assert(size >= headroom, headroom_bigger);
+
+	struct msgb *msg = msgb_alloc_c(ctx, size, name);
+	if (msg)
+		msgb_reserve(msg, headroom);
+	return msg;
+}
+
+
 /*! Allocate message buffer with specified headroom
  *  \param[in] size size in bytes, including headroom
  *  \param[in] headroom headroom in bytes
@@ -456,7 +549,7 @@ static inline int msgb_l3trim(struct msgb *msg, int l3len)
 static inline struct msgb *msgb_alloc_headroom(int size, int headroom,
 						const char *name)
 {
-	osmo_static_assert(size > headroom, headroom_bigger);
+	osmo_static_assert(size >= headroom, headroom_bigger);
 
 	struct msgb *msg = msgb_alloc(size, name);
 	if (msg)
@@ -503,6 +596,145 @@ static inline int msgb_test_invariant(const struct msgb *msg)
 	return lbound <= msg->head +  msg->data_len;
 }
 
+
+/* msgb data comparison helpers */
+
+/*! Compare: check data in msgb against given data
+ *  \param[in] msg message buffer
+ *  \param[in] data expected data
+ *  \param[in] len length of data
+ *  \returns boolean indicating whether msgb content is equal to the given data
+ */
+#define msgb_eq_data(msg, data, len)				\
+	_msgb_eq(__FILE__, __LINE__, __func__, 0, msg, data, len, false)
+
+/*! Compare: check L1 data in msgb against given data
+ *  \param[in] msg message buffer
+ *  \param[in] data expected L1 data
+ *  \param[in] len length of data
+ *  \returns boolean indicating whether msgb L1 content is equal to the given data
+ */
+#define msgb_eq_l1_data(msg, data, len)				\
+	_msgb_eq(__FILE__, __LINE__, __func__, 1, msg, data, len, false)
+
+/*! Compare: check L2 data in msgb against given data
+ *  \param[in] msg message buffer
+ *  \param[in] data expected L2 data
+ *  \param[in] len length of data
+ *  \returns boolean indicating whether msgb L2 content is equal to the given data
+ */
+#define msgb_eq_l2_data(msg, data, len)				\
+	_msgb_eq(__FILE__, __LINE__, __func__, 2, msg, data, len, false)
+
+/*! Compare: check L3 data in msgb against given data
+ *  \param[in] msg message buffer
+ *  \param[in] data expected L3 data
+ *  \param[in] len length of data
+ *  \returns boolean indicating whether msgb L3 content is equal to the given data
+ */
+#define msgb_eq_l3_data(msg, data, len)				\
+	_msgb_eq(__FILE__, __LINE__, __func__, 3, msg, data, len, false)
+
+/*! Compare: check L4 data in msgb against given data
+ *  \param[in] msg message buffer
+ *  \param[in] data expected L4 data
+ *  \param[in] len length of data
+ *  \returns boolean indicating whether msgb L4 content is equal to the given data
+ */
+#define msgb_eq_l4_data(msg, data, len)				\
+	_msgb_eq(__FILE__, __LINE__, __func__, 4, msg, data, len, false)
+
+
+/* msgb test/debug helpers */
+
+/*! Compare and print: check data in msgb against given data and print errors if any
+ *  \param[in] msg message buffer
+ *  \param[in] data expected data
+ *  \param[in] len length of data
+ *  \returns boolean indicating whether msgb content is equal to the given data
+ */
+#define msgb_eq_data_print(msg, data, len)				\
+	_msgb_eq(__FILE__, __LINE__, __func__, 0, msg, data, len, true)
+
+/*! Compare and print: check L1 data in msgb against given data and print errors if any
+ *  \param[in] msg message buffer
+ *  \param[in] data expected L1 data
+ *  \param[in] len length of data
+ *  \returns boolean indicating whether msgb L1 content is equal to the given data
+ */
+#define msgb_eq_l1_data_print(msg, data, len)				\
+	_msgb_eq(__FILE__, __LINE__, __func__, 1, msg, data, len, true)
+
+/*! Compare and print: check L2 data in msgb against given data and print errors if any
+ *  \param[in] msg message buffer
+ *  \param[in] data expected L2 data
+ *  \param[in] len length of data
+ *  \returns boolean indicating whether msgb L2 content is equal to the given data
+ */
+#define msgb_eq_l2_data_print(msg, data, len)				\
+	_msgb_eq(__FILE__, __LINE__, __func__, 2, msg, data, len, true)
+
+/*! Compare and print: check L3 data in msgb against given data and print errors if any
+ *  \param[in] msg message buffer
+ *  \param[in] data expected L3 data
+ *  \param[in] len length of data
+ *  \returns boolean indicating whether msgb L3 content is equal to the given data
+ */
+#define msgb_eq_l3_data_print(msg, data, len)				\
+	_msgb_eq(__FILE__, __LINE__, __func__, 3, msg, data, len, true)
+
+
+/*! Compare and print: check L4 data in msgb against given data and print errors if any
+ *  \param[in] msg message buffer
+ *  \param[in] data expected L4 data
+ *  \param[in] len length of data
+ *  \returns boolean indicating whether msgb L4 content is equal to the given data
+ */
+#define msgb_eq_l4_data_print(msg, data, len)				\
+	_msgb_eq(__FILE__, __LINE__, __func__, 4, msg, data, len, true)
+
+bool _msgb_eq(const char *file, size_t line, const char *func, uint8_t level,
+	      const struct msgb *msg, const uint8_t *data, size_t len, bool print);
+
+
+/* msgb data comparison */
+
+/*! Compare msgbs
+ *  \param[in] msg1 message buffer
+ *  \param[in] msg2 reference message buffer
+ *  \returns boolean indicating whether msgb content is equal
+ */
+#define msgb_eq(msg1, msg2) msgb_eq_data(msg1, msgb_data(msg2), msgb_length(msg2))
+
+/*! Compare msgbs L1 content
+ *  \param[in] msg1 message buffer
+ *  \param[in] msg2 reference message buffer
+ *  \returns boolean indicating whether msgb L1 content is equal
+ */
+#define msgb_eq_l1(msg1, msg2) msgb_eq_l1_data(msg1, msgb_l1(msg2), msgb_l1len(msg2))
+
+/*! Compare msgbs L2 content
+ *  \param[in] msg1 message buffer
+ *  \param[in] msg2 reference message buffer
+ *  \returns boolean indicating whether msgb L2 content is equal
+ */
+#define msgb_eq_l2(msg1, msg2) msgb_eq_l2_data(msg1, msgb_l2(msg2), msgb_l2len(msg2))
+
+/*! Compare msgbs L3 content
+ *  \param[in] msg1 message buffer
+ *  \param[in] msg2 reference message buffer
+ *  \returns boolean indicating whether msgb L3 content is equal
+ */
+#define msgb_eq_l3(msg1, msg2) msgb_eq_l3_data(msg1, msgb_l3(msg2), msgb_l3len(msg2))
+
+/*! Compare msgbs L4 content
+ *  \param[in] msg1 message buffer
+ *  \param[in] msg2 reference message buffer
+ *  \returns boolean indicating whether msgb L4 content is equal
+ */
+#define msgb_eq_l4(msg1, msg2) msgb_eq_l4_data(msg1, msgb_l4(msg2), msgb_l4len(msg2))
+
+
 /* non inline functions to ease binding */
 
 uint8_t *msgb_data(const struct msgb *msg);
@@ -511,6 +743,13 @@ void *msgb_talloc_ctx_init(void *root_ctx, unsigned int pool_size);
 void msgb_set_talloc_ctx(void *ctx) OSMO_DEPRECATED("Use msgb_talloc_ctx_init() instead");
 int msgb_printf(struct msgb *msgb, const char *format, ...);
 
+static inline const char *msgb_hexdump_l1(const struct msgb *msg)
+{
+	if (!msgb_l1(msg) || !(msgb_l1len(msg)))
+		return "[]";
+	return osmo_hexdump((const unsigned char *) msgb_l1(msg), msgb_l1len(msg));
+}
+
 static inline const char *msgb_hexdump_l2(const struct msgb *msg)
 {
 	if (!msgb_l2(msg) || !(msgb_l2len(msg)))
@@ -525,4 +764,11 @@ static inline const char *msgb_hexdump_l3(const struct msgb *msg)
 	return osmo_hexdump((const unsigned char*) msgb_l3(msg), msgb_l3len(msg));
 }
 
+static inline const char *msgb_hexdump_l4(const struct msgb *msg)
+{
+	if (!msgb_l4(msg) || !(msgb_l4len(msg)))
+		return "[]";
+	return osmo_hexdump((const unsigned char*) msgb_l4(msg), msgb_l4len(msg));
+}
+
 /*! @} */
