@@ -97,6 +97,7 @@ enum node_type {
 	L_CS7_SCCPADDR_NODE,	/*!< SS7 SCCP Address */
 	L_CS7_SCCPADDR_GT_NODE,	/*!< SS7 SCCP Global Title */
 
+	L_CPU_SCHED_NODE,		/*!< CPU Sched related options node */
 	/*
 	 * When adding new nodes to the libosmocore project, these nodes can be
 	 * used to avoid ABI changes for unrelated projects.
@@ -132,9 +133,10 @@ struct cmd_node {
 	char name[64];
 };
 
+/*! Attributes (flags) for \ref cmd_element */
 enum {
-	CMD_ATTR_DEPRECATED = 1,
-	CMD_ATTR_HIDDEN,
+	CMD_ATTR_DEPRECATED	= (1 << 0),
+	CMD_ATTR_HIDDEN		= (1 << 1),
 };
 
 /*! Structure of a command element */
@@ -303,6 +305,10 @@ struct desc {
 #define CMD_IPV6(S)        ((strcmp ((S), "X:X::X:X") == 0))
 #define CMD_IPV6_PREFIX(S) ((strcmp ((S), "X:X::X:X/M") == 0))
 
+#define VTY_IPV4_CMD "A.B.C.D"
+#define VTY_IPV6_CMD "X:X::X:X"
+#define VTY_IPV46_CMD "(" VTY_IPV4_CMD "|" VTY_IPV6_CMD ")"
+
 /* Common descriptions. */
 #define SHOW_STR "Show running system information\n"
 #define IP_STR "IP information\n"
@@ -400,4 +406,6 @@ void print_version(int print_copyright);
 
 extern void *tall_vty_cmd_ctx;
 
+int vty_dump_xml_ref(FILE *stream);
+
 /*! @} */
