@@ -45,6 +45,7 @@
 #include <osmocom/core/bits.h>
 #include <osmocom/core/bitvec.h>
 #include <osmocom/core/panic.h>
+#include <osmocom/core/utils.h>
 
 #define BITNUM_FROM_COMP(byte, bit)	((byte*8)+bit)
 
@@ -58,25 +59,18 @@ static inline unsigned int bytenum_from_bitnum(unsigned int bitnum)
 /* convert ZERO/ONE/L/H to a bitmask at given pos in a byte */
 static uint8_t bitval2mask(enum bit_value bit, uint8_t bitnum)
 {
-	int bitval;
-
 	switch (bit) {
 	case ZERO:
-		bitval = (0 << bitnum);
-		break;
+		return (0 << bitnum);
 	case ONE:
-		bitval = (1 << bitnum);
-		break;
+		return (1 << bitnum);
 	case L:
-		bitval = ((0x2b ^ (0 << bitnum)) & (1 << bitnum));
-		break;
+		return ((0x2b ^ (0 << bitnum)) & (1 << bitnum));
 	case H:
-		bitval = ((0x2b ^ (1 << bitnum)) & (1 << bitnum));
-		break;
+		return ((0x2b ^ (1 << bitnum)) & (1 << bitnum));
 	default:
 		return 0;
 	}
-	return bitval;
 }
 
 /*! check if the bit is 0 or 1 for a given position inside a bitvec
@@ -298,7 +292,7 @@ int bitvec_fill(struct bitvec *bv, unsigned int num_bits, enum bit_value fill)
 	return 0;
 }
 
-/*! pad all remaining bits up to num_bits
+/*! pad all remaining bits up to a given bit number
  *  \return 0 on success; negative otherwise */
 int bitvec_spare_padding(struct bitvec *bv, unsigned int up_to_bit)
 {
@@ -401,12 +395,12 @@ int bitvec_set_bytes(struct bitvec *bv, const uint8_t *bytes, unsigned int count
 }
 
 /*! Allocate a bit vector
- *  \param[in] size Number of bits in the vector
+ *  \param[in] size Number of bytes in the vector
  *  \param[in] ctx Context from which to allocate
  *  \return pointer to allocated vector; NULL in case of error */
 struct bitvec *bitvec_alloc(unsigned int size, TALLOC_CTX *ctx)
 {
-	struct bitvec *bv = talloc_zero(ctx, struct bitvec);
+	struct bitvec *bv = talloc(ctx, struct bitvec);
 	if (!bv)
 		return NULL;
 
@@ -425,6 +419,8 @@ struct bitvec *bitvec_alloc(unsigned int size, TALLOC_CTX *ctx)
  *  \param[in] bit vector to free */
 void bitvec_free(struct bitvec *bv)
 {
+	if (bv == NULL)
+		return;
 	talloc_free(bv->data);
 	talloc_free(bv);
 }
@@ -435,7 +431,7 @@ void bitvec_free(struct bitvec *bv)
  *  \return number of bytes (= bits) copied */
 unsigned int bitvec_pack(const struct bitvec *bv, uint8_t *buffer)
 {
-	unsigned int i = 0;
+	unsigned int i;
 	for (i = 0; i < bv->data_len; i++)
 		buffer[i] = bv->data[i];
 
@@ -448,7 +444,7 @@ unsigned int bitvec_pack(const struct bitvec *bv, uint8_t *buffer)
  *  \return number of bytes (= bits) copied */
 unsigned int bitvec_unpack(struct bitvec *bv, const uint8_t *buffer)
 {
-	unsigned int i = 0;
+	unsigned int i;
 	for (i = 0; i < bv->data_len; i++)
 		bv->data[i] = buffer[i];
 
@@ -462,17 +458,13 @@ unsigned int bitvec_unpack(struct bitvec *bv, const uint8_t *buffer)
  */
 int bitvec_unhex(struct bitvec *bv, const char *src)
 {
-	unsigned i;
-	unsigned val;
-	unsigned write_index = 0;
-	unsigned digits = bv->data_len * 2;
+	int rc;
 
-	for (i = 0; i < digits; i++) {
-		if (sscanf(src + i, "%1x", &val) < 1) {
-			return 1;
-		}
-		bitvec_write_field(bv, &write_index, val, 4);
-	}
+	rc = osmo_hexparse(src, bv->data, bv->data_len);
+	if (rc < 0) /* turn -1 into 1 in case of error */
+		return 1;
+
+	bv->cur_bit = rc * 8;
 	return 0;
 }
 
@@ -504,7 +496,7 @@ uint64_t bitvec_read_field(struct bitvec *bv, unsigned int *read_index, unsigned
  *  \param[in] bv The boolean vector to work on
  *  \param[in,out] write_index Where writing supposed to start in the vector
  *  \param[in] len How many bits to write
- *  \returns next write index or negative value on error
+ *  \returns 0 on success, negative value on error
  */
 int bitvec_write_field(struct bitvec *bv, unsigned int *write_index, uint64_t val, unsigned int len)
 {
