@@ -589,6 +589,7 @@ const struct tlv_definition abis_nm_att_tlvdef = {
 /*! org.osmocom GSM A-bis OML TLV parser definition */
 const struct tlv_definition abis_nm_osmo_att_tlvdef = {
 	.def = {
+		[NM_ATT_OSMO_NS_LINK_CFG] =	{ TLV_TYPE_TL16V },
 		[NM_ATT_OSMO_REDUCEPOWER] =	{ TLV_TYPE_TV },
 	},
 };
@@ -928,14 +929,28 @@ enum gsm_phys_chan_config abis_nm_pchan4chcomb(uint8_t chcomb)
 	return GSM_PCHAN_NONE;
 }
 
-const char *abis_nm_dump_foh(const struct abis_om_fom_hdr *foh)
+char *abis_nm_dump_foh_buf(char *buf, size_t buf_len, const struct abis_om_fom_hdr *foh)
 {
-	static char foh_buf[128];
-	snprintf(foh_buf, sizeof(foh_buf), "OC=%s(%02x) INST=(%02x,%02x,%02x)",
+	snprintf(buf, buf_len, "OC=%s(%02x) INST=(%02x,%02x,%02x)",
 		get_value_string(abis_nm_obj_class_names, foh->obj_class),
 		foh->obj_class, foh->obj_inst.bts_nr, foh->obj_inst.trx_nr,
 		foh->obj_inst.ts_nr);
-	return foh_buf;
+	return buf;
+}
+
+const char *abis_nm_dump_foh(const struct abis_om_fom_hdr *foh)
+{
+	static __thread char foh_buf[128];
+	return abis_nm_dump_foh_buf(foh_buf, sizeof(foh_buf), foh);
+}
+
+char *abis_nm_dump_foh_c(void *ctx, const struct abis_om_fom_hdr *foh)
+{
+	size_t len = 15 /* format */ + 22 /* obj_class_name */+ 4*3 /* uint8 */ + 1 /*nul*/;
+	char *buf = talloc_size(ctx, len);
+	if (!buf)
+		return NULL;
+	return abis_nm_dump_foh_buf(buf, len, foh);
 }
 
 /* this is just for compatibility reasons, it is now a macro */
