@@ -24,25 +24,38 @@
 #include <stdint.h>
 #include <stdlib.h>
 #include <string.h>
+#include <talloc.h>
 
 #include <osmocom/gsm/apn.h>
 
 #define APN_OI_GPRS_FMT	"mnc%03u.mcc%03u.gprs"
 #define APN_GPRS_FMT	"%s.mnc%03u.mcc%03u.gprs"
 
-static char apn_strbuf[APN_MAXLEN+1];
+static __thread char apn_strbuf[APN_MAXLEN+1];
+
+char *osmo_apn_qualify_buf(char *buf, size_t buf_len, unsigned int mcc, unsigned int mnc, const char *ni)
+{
+	snprintf(buf, buf_len-1, APN_GPRS_FMT, ni, mnc, mcc);
+	buf[buf_len-1] = '\0';
+
+	return buf;
+}
 
 char *osmo_apn_qualify(unsigned int mcc, unsigned int mnc, const char *ni)
 {
-	snprintf(apn_strbuf, sizeof(apn_strbuf)-1, APN_GPRS_FMT,
-		ni, mnc, mcc);
-	apn_strbuf[sizeof(apn_strbuf)-1] = '\0';
-
-	return apn_strbuf;
+	return osmo_apn_qualify_buf(apn_strbuf, sizeof(apn_strbuf), mcc, mnc, ni);
 }
 
-char *osmo_apn_qualify_from_imsi(const char *imsi,
-				 const char *ni, int have_3dig_mnc)
+char *osmo_apn_qualify_c(const void *ctx, unsigned int mcc, unsigned int mnc, const char *ni)
+{
+	char *buf = talloc_size(ctx, APN_MAXLEN);
+	if (!buf)
+		return NULL;
+	return osmo_apn_qualify_buf(buf, APN_MAXLEN, mcc, mnc, ni);
+}
+
+char *osmo_apn_qualify_from_imsi_buf(char *buf, size_t buf_len, const char *imsi,
+				     const char *ni, int have_3dig_mnc)
 {
 	char cbuf[3+1], nbuf[3+1];
 
@@ -56,7 +69,21 @@ char *osmo_apn_qualify_from_imsi(const char *imsi,
 		strncpy(nbuf, imsi+3, 2);
 		nbuf[2] = '\0';
 	}
-	return osmo_apn_qualify(atoi(cbuf), atoi(nbuf), ni);
+	return osmo_apn_qualify_buf(buf, buf_len, atoi(cbuf), atoi(nbuf), ni);
+}
+
+char *osmo_apn_qualify_from_imsi(const char *imsi,
+				 const char *ni, int have_3dig_mnc)
+{
+	return osmo_apn_qualify_from_imsi_buf(apn_strbuf, sizeof(apn_strbuf), imsi, ni, have_3dig_mnc);
+}
+
+char *osmo_apn_qualify_from_imsi_c(const void *ctx, const char *imsi, const char *ni, int have_3dig_mnc)
+{
+	char *buf = talloc_size(ctx, APN_MAXLEN);
+	if (!buf)
+		return NULL;
+	return osmo_apn_qualify_from_imsi_buf(buf, APN_MAXLEN, imsi, ni, have_3dig_mnc);
 }
 
 /**
@@ -68,7 +95,7 @@ char *osmo_apn_qualify_from_imsi(const char *imsi,
  *
  * \returns out_str on success and NULL otherwise
  */
-char * osmo_apn_to_str(char *out_str, const uint8_t *apn_enc, size_t apn_enc_len)
+char *osmo_apn_to_str(char *out_str, const uint8_t *apn_enc, size_t apn_enc_len)
 {
 	char *str = out_str;
 	size_t rest_chars = apn_enc_len;
