@@ -28,6 +28,7 @@
 #include <errno.h>
 #include <osmocom/gsm/protocol/gsm_08_08.h>
 #include <osmocom/gsm/gsm48.h>
+#include <osmocom/gsm/gsm0808.h>
 #include <osmocom/gsm/gsm0808_utils.h>
 
 #define IP_V4_ADDR_LEN 4
@@ -48,6 +49,32 @@
  *  \file gsm0808_utils.c
  */
 
+/*! Encode TS 08.08 AoIP Cause IE
+ *  \param[out] msg Message Buffer to which to append IE
+ *  \param[in] cause Cause code to be used in IE
+ *  \returns number of bytes added to \a msg */
+uint8_t gsm0808_enc_cause(struct msgb *msg, uint16_t cause)
+{
+	/* See also 3GPP TS 48.008 3.2.2.5 Cause */
+	uint8_t *old_tail;
+	bool extended;
+
+	old_tail = msg->tail;
+
+	extended = gsm0808_cause_ext(cause >> 8);
+
+	msgb_put_u8(msg, GSM0808_IE_CAUSE);
+	if (extended) {
+		msgb_put_u8(msg, 2);
+		msgb_put_u16(msg, cause);
+	} else {
+		msgb_put_u8(msg, 1);
+		msgb_put_u8(msg, (uint8_t) (cause & 0xFF));
+	}
+
+	return (uint8_t) (msg->tail - old_tail);
+}
+
 /*! Encode TS 08.08 AoIP transport address IE
  *  \param[out] msg Message Buffer to which to append IE
  *  \param[in] ss Socket Address to be used in IE
@@ -145,8 +172,38 @@ int gsm0808_dec_aoip_trasp_addr(struct sockaddr_storage *ss,
 	return (int)(elem - old_elem);
 }
 
+/*! Decode TS 08.08 (Osmocom Extension) Osmux CID
+ *  TV with len(V) == 1, and V is the CID to be used.
+ *  \param[out] cid Caller-provided variable where CID is stored
+ *  \param[in] elem pointer to IE value
+ *  \param[in] len length of \a elem in bytes
+ *  \returns number of bytes parsed */
+int gsm0808_dec_osmux_cid(uint8_t *cid, const uint8_t *elem, uint8_t len)
+{
+	OSMO_ASSERT(cid);
+	if (!elem)
+		return -EINVAL;
+	if (len != 1)
+		return -EINVAL;
+
+	*cid = *elem;
+
+	return 1;
+}
+
 #endif /* HAVE_SYS_SOCKET_H */
 
+/* Decode 5-byte LAI list element data (see TS 08.08 3.2.2.27) into MCC/MNC/LAC. */
+static void decode_lai(const uint8_t *data, struct osmo_location_area_id *decoded)
+{
+	struct gsm48_loc_area_id lai;
+
+	/* Copy data to stack to prevent unaligned access in gsm48_decode_lai2(). */
+	memcpy(&lai, data, sizeof(lai)); /* don't byte swap yet */
+
+	gsm48_decode_lai2(&lai, decoded);
+}
+
 /* Helper function for gsm0808_enc_speech_codec()
  * and gsm0808_enc_speech_codec_list() */
 static uint8_t enc_speech_codec(struct msgb *msg,
@@ -216,7 +273,7 @@ static uint8_t enc_speech_codec(struct msgb *msg,
 	case GSM0808_SCT_FR5:
 	case GSM0808_SCT_HR4:
 	case GSM0808_SCT_CSD:
-		OSMO_ASSERT((sc->cfg & 0xff00) == 0)
+		OSMO_ASSERT((sc->cfg & 0xff00) == 0);
 		msgb_put_u8(msg, (uint8_t) sc->cfg & 0xff);
 		break;
 	default:
@@ -349,9 +406,6 @@ uint8_t gsm0808_enc_speech_codec_list(struct msgb *msg,
 	OSMO_ASSERT(msg);
 	OSMO_ASSERT(scl);
 
-	/* Empty list */
-	OSMO_ASSERT(scl->len >= 1);
-
 	msgb_put_u8(msg, GSM0808_IE_SPEECH_CODEC_LIST);
 	tlv_len = msgb_put(msg, 1);
 	old_tail = msg->tail;
@@ -384,8 +438,6 @@ int gsm0808_dec_speech_codec_list(struct gsm0808_speech_codec_list *scl,
 	OSMO_ASSERT(scl);
 	if (!elem)
 		return -EINVAL;
-	if (len == 0)
-		return -EINVAL;
 
 	memset(scl, 0, sizeof(*scl));
 
@@ -404,11 +456,6 @@ int gsm0808_dec_speech_codec_list(struct gsm0808_speech_codec_list *scl,
 
 	scl->len = decoded;
 
-	/* Empty list */
-	if (decoded < 1) {
-		return -EINVAL;
-	}
-
 	return (int)(elem - old_elem);
 }
 
@@ -492,6 +539,157 @@ int gsm0808_dec_channel_type(struct gsm0808_channel_type *ct,
 	return (int)(elem - old_elem);
 }
 
+/*! Create BSSMAP Global Call Reference, 3GPP TS 48.008 §3.2.2.115.
+ *  \param[out] msg Message Buffer for appending IE
+ *  \param[in] g Global Call Reference, 3GPP TS 29.205 Table B 2.1.9.1
+ *  \returns number of bytes added to \a msg or 0 on error */
+static uint8_t gsm0808_enc_gcr(struct msgb *msg, const struct osmo_gcr_parsed *g)
+{
+	uint8_t enc, *len = msgb_tl_put(msg, GSM0808_IE_GLOBAL_CALL_REF);
+
+	enc = osmo_enc_gcr(msg, g);
+	if (!enc)
+		return 0;
+
+	*len = enc;
+	return enc + 2; /* type (1 byte) + length (1 byte) */
+}
+
+/*! Decode BSSMAP Global Call Reference, 3GPP TS 29.205 Table B 2.1.9.1.
+ *  \param[out] gcr Caller-provided memory to store Global Call Reference
+ *  \param[in] tp IE values to be decoded
+ *  \returns number of bytes parsed; negative on error */
+static int gsm0808_dec_gcr(struct osmo_gcr_parsed *gcr, const struct tlv_parsed *tp)
+{
+	int ret;
+	const uint8_t *buf = TLVP_VAL_MINLEN(tp, GSM0808_IE_GLOBAL_CALL_REF, OSMO_GCR_MIN_LEN);
+	if (!buf)
+		return -EINVAL;
+
+	ret = osmo_dec_gcr(gcr, buf, TLVP_LEN(tp, GSM0808_IE_GLOBAL_CALL_REF));
+	if (ret < 0)
+		return -ENOENT;
+
+	return 2 + ret;
+}
+
+/*! Add LCLS parameters to a given msgb, 3GPP TS 48.008 §3.2.2.115 - 3.2.2.120.
+ *  \param[out] msg Message Buffer for appending IE
+ *  \param[in] lcls LCLS-related data
+ *  \returns number of bytes added to \a msg or 0 on error */
+uint8_t gsm0808_enc_lcls(struct msgb *msg, const struct osmo_lcls *lcls)
+{
+	uint8_t enc = 0;
+
+	/* LCLS: §3.2.2.115 Global Call Reference */
+	if (lcls->gcr_available)
+		enc = gsm0808_enc_gcr(msg, &lcls->gcr);
+
+	/* LCLS: §3.2.2.116 Configuration */
+	if (lcls->config != GSM0808_LCLS_CFG_NA) {
+		msgb_tv_put(msg, GSM0808_IE_LCLS_CONFIG, lcls->config);
+		enc += 2;
+	}
+
+	/* LCLS: §3.2.2.117 Connection Status Control */
+	if (lcls->control != GSM0808_LCLS_CSC_NA) {
+		msgb_tv_put(msg, GSM0808_IE_LCLS_CONN_STATUS_CTRL, lcls->control);
+		enc += 2;
+	}
+
+	/* LCLS: §3.2.2.118 Correlation-Not-Needed */
+	if (!lcls->corr_needed) {
+		msgb_v_put(msg, GSM0808_IE_LCLS_CORR_NOT_NEEDED);
+		enc++;
+	}
+
+	return enc;
+}
+
+/*! Decode LCLS parameters to a given msgb, 3GPP TS 48.008 §3.2.2.115 - 3.2.2.120.
+ *  \param[out] lcls Caller-provided memory to store LCLS-related data
+ *  \param[in] tp IE values to be decoded
+ *  \returns GCR size or negative on error */
+int gsm0808_dec_lcls(struct osmo_lcls *lcls, const struct tlv_parsed *tp)
+{
+	int ret = gsm0808_dec_gcr(&lcls->gcr, tp);
+
+	lcls->gcr_available = (ret < 0) ? false : true;
+	lcls->config = tlvp_val8(tp, GSM0808_IE_LCLS_CONFIG, GSM0808_LCLS_CFG_NA);
+	lcls->control = tlvp_val8(tp, GSM0808_IE_LCLS_CONN_STATUS_CTRL, GSM0808_LCLS_CSC_NA);
+	lcls->corr_needed = TLVP_PRESENT(tp, GSM0808_IE_LCLS_CORR_NOT_NEEDED) ? false : true;
+
+	return ret;
+}
+
+static __thread char dbuf[256];
+
+/*! Dump LCLS parameters (GCR excluded) into string for printing.
+ *  \param[out] buf caller-allocated output string buffer
+ *  \param[in] buf_len size of buf in bytes
+ *  \param[in] lcls pointer to the struct to print.
+ *  \returns string representation of LCLS or NULL on error. */
+char *osmo_lcls_dump_buf(char *buf, size_t buf_len, const struct osmo_lcls *lcls)
+{
+	struct osmo_strbuf s = { .buf = buf, .len = buf_len };
+
+	if (!lcls)
+		return NULL;
+
+	OSMO_STRBUF_PRINTF(s, "LCLS Config: %s, Control: %s, Correlation-Needed: %u",
+			   gsm0808_lcls_config_name(lcls->config),
+			   gsm0808_lcls_control_name(lcls->control),
+			   lcls->corr_needed);
+
+	return dbuf;
+}
+
+/*! Dump LCLS parameters (GCR excluded) into static string buffer for printing.
+ *  \param[in] lcls pointer to the struct to print.
+ *  \returns string representation of LCLS in static buffer or NULL on error. */
+char *osmo_lcls_dump(const struct osmo_lcls *lcls)
+{
+	return osmo_lcls_dump_buf(dbuf, sizeof(dbuf), lcls);
+}
+
+char *osmo_lcls_dump_c(void *ctx, const struct osmo_lcls *lcls)
+{
+	char *buf = talloc_size(ctx, 256);
+	if (!buf)
+		return NULL;
+	return osmo_lcls_dump_buf(buf, 256, lcls);
+}
+
+/*! Dump GCR struct into string for printing.
+ *  \param[out] buf caller-allocated output string buffer
+ *  \param[in] buf_len size of buf in bytes
+ *  \param[in] lcls pointer to the struct to print.
+ *  \returns string representation of GCR or NULL on error. */
+char *osmo_gcr_dump_buf(char *buf, size_t buf_len, const struct osmo_lcls *lcls)
+{
+	struct osmo_strbuf s = { .buf = buf, .len = buf_len };
+
+	if (!lcls)
+		return NULL;
+
+	if (lcls->gcr_available) {
+		OSMO_STRBUF_PRINTF(s, "GCR NetID 0x%s, ", osmo_hexdump_nospc(lcls->gcr.net, lcls->gcr.net_len));
+		/* osmo_hexdump() uses static buffers so we can't call it twice withing the same parameter list */
+		OSMO_STRBUF_PRINTF(s, "Node 0x%x, CallRefID 0x%s", lcls->gcr.node, osmo_hexdump_nospc(lcls->gcr.cr, 5));
+	}
+
+	return dbuf;
+}
+
+/*! Dump GCR struct into static string buffer for printing.
+ *  \param[in] lcls pointer to the struct to print.
+ *  \returns string representation of GCR in static buffer or NULL on error. */
+char *osmo_gcr_dump(const struct osmo_lcls *lcls)
+{
+	return osmo_gcr_dump_buf(dbuf, sizeof(dbuf), lcls);
+}
+
+
 /*! Encode TS 08.08 Encryption Information IE
  *  \param[out] msg Message Buffer to which IE is to be appended
  *  \param[in] ei Encryption Information to be encoded
@@ -569,6 +767,114 @@ int gsm0808_dec_encrypt_info(struct gsm0808_encrypt_info *ei,
 	return (int)(elem - old_elem);
 }
 
+/* Return the size of the value part of a cell identifier of given type */
+int gsm0808_cell_id_size(enum CELL_IDENT discr)
+{
+	switch (discr) {
+	case CELL_IDENT_WHOLE_GLOBAL:
+		return 7;
+	case CELL_IDENT_LAC_AND_CI:
+		return 4;
+	case CELL_IDENT_CI:
+		return 2;
+	case CELL_IDENT_LAI_AND_LAC:
+		return 5;
+	case CELL_IDENT_LAC:
+		return 2;
+	case CELL_IDENT_BSS:
+	case CELL_IDENT_NO_CELL:
+		return 0;
+	default:
+		return -EINVAL;
+	}
+}
+/*! Decode a single GSM 08.08 Cell ID list element payload
+ *  \param[out] out caller-provided output union
+ *  \param[in] discr Cell ID discriminator describing type to be decoded
+ *  \param[in] buf User-provided input buffer containing binary Cell ID list element
+ *  \param[in] len Length of buf
+ *  \returns 0 on success; negative on error */
+int gsm0808_decode_cell_id_u(union gsm0808_cell_id_u *out, enum CELL_IDENT discr, const uint8_t *buf, unsigned int len)
+{
+	switch (discr) {
+	case CELL_IDENT_WHOLE_GLOBAL:
+		if (len < 7)
+			return -EINVAL;
+		decode_lai(buf, &out->global.lai);
+		out->global.cell_identity = osmo_load16be(buf + sizeof(struct gsm48_loc_area_id));
+		break;
+	case CELL_IDENT_LAC_AND_CI:
+		if (len < 4)
+			return -EINVAL;
+		out->lac_and_ci.lac = osmo_load16be(buf);
+		out->lac_and_ci.ci = osmo_load16be(buf + sizeof(uint16_t));
+		break;
+	case CELL_IDENT_CI:
+		if (len < 2)
+			return -EINVAL;
+		out->ci = osmo_load16be(buf);
+		break;
+	case CELL_IDENT_LAI_AND_LAC:
+		if (len < 5)
+			return -EINVAL;
+		decode_lai(buf, &out->lai_and_lac);
+		break;
+	case CELL_IDENT_LAC:
+		if (len < 2)
+			return -EINVAL;
+		out->lac = osmo_load16be(buf);
+		break;
+	case CELL_IDENT_BSS:
+	case CELL_IDENT_NO_CELL:
+		/* Does not have any list items */
+		break;
+	default:
+		/* Remaining cell identification types are not implemented. */
+		return -EINVAL;
+	}
+	return 0;
+}
+
+void gsm0808_msgb_put_cell_id_u(struct msgb *msg, enum CELL_IDENT id_discr, const union gsm0808_cell_id_u *u)
+{
+	switch (id_discr) {
+	case CELL_IDENT_WHOLE_GLOBAL: {
+		const struct osmo_cell_global_id *id = &u->global;
+		struct gsm48_loc_area_id lai;
+		gsm48_generate_lai2(&lai, &id->lai);
+		memcpy(msgb_put(msg, sizeof(lai)), &lai, sizeof(lai));
+		msgb_put_u16(msg, id->cell_identity);
+		break;
+	}
+	case CELL_IDENT_LAC_AND_CI: {
+		const struct osmo_lac_and_ci_id *id = &u->lac_and_ci;
+		msgb_put_u16(msg, id->lac);
+		msgb_put_u16(msg, id->ci);
+		break;
+	}
+	case CELL_IDENT_CI:
+		msgb_put_u16(msg, u->ci);
+		break;
+	case CELL_IDENT_LAI_AND_LAC: {
+		const struct osmo_location_area_id *id = &u->lai_and_lac;
+		struct gsm48_loc_area_id lai;
+		gsm48_generate_lai2(&lai, id);
+		memcpy(msgb_put(msg, sizeof(lai)), &lai, sizeof(lai));
+		break;
+	}
+	case CELL_IDENT_LAC:
+		msgb_put_u16(msg, u->lac);
+		break;
+	case CELL_IDENT_BSS:
+	case CELL_IDENT_NO_CELL:
+		/* Does not have any list items */
+		break;
+	default:
+		/* Support for other identifier list types is not implemented. */
+		OSMO_ASSERT(false);
+	}
+}
+
 /*! Encode TS 08.08 Cell Identifier List IE
  *  \param[out] msg Message Buffer to which IE is to be appended
  *  \param[in] cil Cell ID List to be encoded
@@ -589,48 +895,9 @@ uint8_t gsm0808_enc_cell_id_list2(struct msgb *msg,
 
 	msgb_put_u8(msg, cil->id_discr & 0x0f);
 
-	OSMO_ASSERT(cil->id_list_len <= GSM0808_CELL_ID_LIST2_MAXLEN)
-	switch (cil->id_discr) {
-	case CELL_IDENT_WHOLE_GLOBAL:
-		for (i = 0; i < cil->id_list_len; i++) {
-			const struct osmo_cell_global_id *id = &cil->id_list[i].global;
-			struct gsm48_loc_area_id lai;
-			gsm48_generate_lai2(&lai, &id->lai);
-			memcpy(msgb_put(msg, sizeof(lai)), &lai, sizeof(lai));
-			msgb_put_u16(msg, id->cell_identity);
-		}
-		break;
-	case CELL_IDENT_LAC_AND_CI:
-		for (i = 0; i < cil->id_list_len; i++) {
-			const struct osmo_lac_and_ci_id *id = &cil->id_list[i].lac_and_ci;
-			msgb_put_u16(msg, id->lac);
-			msgb_put_u16(msg, id->ci);
-		}
-		break;
-	case CELL_IDENT_CI:
-		for (i = 0; i < cil->id_list_len; i++)
-			msgb_put_u16(msg, cil->id_list[i].ci);
-		break;
-	case CELL_IDENT_LAI_AND_LAC:
-		for (i = 0; i < cil->id_list_len; i++) {
-			const struct osmo_location_area_id *id = &cil->id_list[i].lai_and_lac;
-			struct gsm48_loc_area_id lai;
-			gsm48_generate_lai2(&lai, id);
-			memcpy(msgb_put(msg, sizeof(lai)), &lai, sizeof(lai));
-		}
-		break;
-	case CELL_IDENT_LAC:
-		for (i = 0; i < cil->id_list_len; i++)
-			msgb_put_u16(msg, cil->id_list[i].lac);
-		break;
-	case CELL_IDENT_BSS:
-	case CELL_IDENT_NO_CELL:
-		/* Does not have any list items */
-		break;
-	default:
-		/* Support for other identifier list types is not implemented. */
-		OSMO_ASSERT(false);
-	}
+	OSMO_ASSERT(cil->id_list_len <= GSM0808_CELL_ID_LIST2_MAXLEN);
+	for (i = 0; i < cil->id_list_len; i++)
+		gsm0808_msgb_put_cell_id_u(msg, cil->id_discr, &cil->id_list[i]);
 
 	*tlv_len = (uint8_t) (msg->tail - old_tail);
 	return *tlv_len + 2;
@@ -660,7 +927,7 @@ uint8_t gsm0808_enc_cell_id_list(struct msgb *msg,
 
 	switch (cil->id_discr) {
 	case CELL_IDENT_LAC:
-		OSMO_ASSERT(cil->id_list_len <= CELL_ID_LIST_LAC_MAXLEN)
+		OSMO_ASSERT(cil->id_list_len <= CELL_ID_LIST_LAC_MAXLEN);
 		for (i=0;i<cil->id_list_len;i++) {
 			msgb_put_u16(msg, cil->id_list_lac[i]);
 		}
@@ -677,17 +944,6 @@ uint8_t gsm0808_enc_cell_id_list(struct msgb *msg,
 	return *tlv_len + 2;
 }
 
-/* Decode 5-byte LAI list element data (see TS 08.08 3.2.2.27) into MCC/MNC/LAC. */
-static void decode_lai(const uint8_t *data, struct osmo_location_area_id *decoded)
-{
-	struct gsm48_loc_area_id lai;
-
-	/* Copy data to stack to prevent unaligned access in gsm48_decode_lai2(). */
-	memcpy(&lai, data, sizeof(lai)); /* don't byte swap yet */
-
-	gsm48_decode_lai2(&lai, decoded);
-}
-
 static int parse_cell_id_global_list(struct gsm0808_cell_id_list2 *cil, const uint8_t *data, size_t remain,
 				     size_t *consumed)
 {
@@ -1161,6 +1417,157 @@ int gsm0808_speech_codec_from_chan_type(struct gsm0808_speech_codec *sc,
 	return 0;
 }
 
+/*! Determine a set of AMR speech codec configuration bits (S0-S15) from a
+ *  given GSM 04.08 AMR configuration struct.
+ *  \param[in] cfg AMR configuration in GSM 04.08 format.
+ *  \param[in] hint if the resulting configuration shall be used with a FR or HR TCH.
+ *  \returns configuration bits (S0-S15) */
+uint16_t gsm0808_sc_cfg_from_gsm48_mr_cfg(const struct gsm48_multi_rate_conf *cfg,
+					  bool fr)
+{
+	uint16_t s15_s0 = 0;
+
+	/* Check each rate bit in the AMR multirate configuration and pick the
+	 * matching default configuration as specified in 3GPP TS 28.062,
+	 * Table 7.11.3.1.3-2. */
+	if (cfg->m4_75)
+		s15_s0 |= GSM0808_SC_CFG_DEFAULT_AMR_4_75;
+	if (cfg->m5_15)
+		s15_s0 |= GSM0808_SC_CFG_DEFAULT_AMR_5_15;
+	if (cfg->m5_90)
+		s15_s0 |= GSM0808_SC_CFG_DEFAULT_AMR_5_90;
+	if (cfg->m6_70)
+		s15_s0 |= GSM0808_SC_CFG_DEFAULT_AMR_6_70;
+	if (cfg->m7_40)
+		s15_s0 |= GSM0808_SC_CFG_DEFAULT_AMR_7_40;
+	if (cfg->m7_95)
+		s15_s0 |= GSM0808_SC_CFG_DEFAULT_AMR_7_95;
+	if (cfg->m10_2)
+		s15_s0 |= GSM0808_SC_CFG_DEFAULT_AMR_10_2;
+	if (cfg->m12_2)
+		s15_s0 |= GSM0808_SC_CFG_DEFAULT_AMR_12_2;
+
+	/* Note: 3GPP TS 48.008, chapter 3GPP TS 48.008 states that for AMR
+	 * some of the configuration bits must be coded as zeros. The applied
+	 * bitmask matches the default codec settings. See also the definition
+	 * of enum gsm0808_speech_codec_defaults in gsm_08_08.h and
+	 * 3GPP TS 28.062, Table 7.11.3.1.3-2. */
+	if (fr)
+		s15_s0 &= GSM0808_SC_CFG_DEFAULT_FR_AMR;
+	else
+		s15_s0 &= GSM0808_SC_CFG_DEFAULT_HR_AMR;
+
+	/* The mode that is encoded by S1 (Config-NB-Code = 1), takes a special
+	 * role as it does not stand for a single rate, but for up to four rates
+	 * at once (12.2, 7.4, 5.9, 4.75). We must check if the supplied cfg
+	 * covers this mode. If not, we need to make sure that the related
+	 * bit is removed. (See also 3GPP TS 28.062, Table 7.11.3.1.3-2) */
+	if (!(cfg->m12_2 && cfg->m7_40 && cfg->m5_90 && cfg->m4_75) && fr)
+		s15_s0 &= ~GSM0808_SC_CFG_AMR_4_75_5_90_7_40_12_20;
+	else if (!(cfg->m7_40 && cfg->m5_90 && cfg->m4_75))
+		s15_s0 &= ~GSM0808_SC_CFG_AMR_4_75_5_90_7_40_12_20;
+
+	return s15_s0;
+}
+
+/*! Determine a GSM 04.08 AMR configuration struct from a set of speech codec
+ *  configuration bits (S0-S15)
+ *  \param[out] cfg AMR configuration in GSM 04.08 format.
+ *  \param[in] s15_s0 configuration bits (S15-S0, non-ambiguous).
+ *  \returns zero when successful; negative on error */
+int gsm48_mr_cfg_from_gsm0808_sc_cfg(struct gsm48_multi_rate_conf *cfg,
+				     uint16_t s15_s0)
+{
+	unsigned int count = 0;
+
+	/* Note: See also: 3GPP TS 28.062
+	 * Table 7.11.3.1.3-2: Preferred Configurations for the Adaptive
+	 * Multi-Rate Codec Types */
+
+	/* Note: The resulting multirate-configuration must not contain an
+	 * active set of more than four codec rates. The active set also
+	 * must contain at least one rate. */
+
+	memset(cfg, 0, sizeof(*cfg));
+	cfg->ver = 1;
+	cfg->icmi = 1;
+
+	/* Strip option bits */
+	s15_s0 &= 0x00ff;
+
+	/* Rate 5,15k can never be selected (see table) */
+	cfg->m5_15 = 0;
+
+	if (s15_s0 & GSM0808_SC_CFG_AMR_4_75_5_90_7_40_12_20 & 0xff) {
+		/* Table Table 7.11.3.1.3-2 lists one mode that selects 4
+		 * rates at once (Config-NB-Code = 1). The rates selected
+		 * are known to be compatible between GERAN and UTRAN, since
+		 * an active set must not contain more than four rates at
+		 * a time, we ignore all other settings as they are either
+		 * redundaned or excess settings (invalid) */
+		cfg->m4_75 = 1;
+		cfg->m5_90 = 1;
+		cfg->m7_40 = 1;
+		cfg->m12_2 = 1;
+		count += 4;
+	}
+
+	/* Check the bits in s15_s0 and set the flags for the
+	 * respective rates. */
+	if (s15_s0 & GSM0808_SC_CFG_AMR_4_75 && !cfg->m4_75) {
+		if (count >= 4)
+			return -EINVAL;
+		cfg->m4_75 = 1;
+		count++;
+	}
+	if (s15_s0 & GSM0808_SC_CFG_AMR_5_90 && !cfg->m5_90) {
+		if (count >= 4)
+			return -EINVAL;
+		cfg->m5_90 = 1;
+		count++;
+	}
+	if (s15_s0 & GSM0808_SC_CFG_AMR_6_70) {
+		if (count >= 4)
+			return -EINVAL;
+		cfg->m6_70 = 1;
+		count++;
+	}
+	if (s15_s0 & GSM0808_SC_CFG_AMR_7_40 && !cfg->m7_40) {
+		if (count >= 4)
+			return -EINVAL;
+		cfg->m7_40 = 1;
+		count++;
+	}
+	if (s15_s0 & GSM0808_SC_CFG_AMR_7_95) {
+		if (count >= 4)
+			return -EINVAL;
+		cfg->m7_95 = 1;
+		count++;
+	}
+	if (s15_s0 & GSM0808_SC_CFG_AMR_10_2) {
+		if (count >= 4)
+			return -EINVAL;
+		cfg->m10_2 = 1;
+		count++;
+	}
+	if (s15_s0 & GSM0808_SC_CFG_AMR_12_2 && !cfg->m12_2) {
+		if (count >= 4)
+			return -EINVAL;
+		cfg->m12_2 = 1;
+		count++;
+	}
+
+	if (count == 0)
+		return -EINVAL;
+
+	return 0;
+}
+
+int gsm0808_get_cipher_reject_cause(const struct tlv_parsed *tp)
+{
+	return gsm0808_get_cause(tp);
+}
+
 /*! Print a human readable name of the cell identifier to the char buffer.
  * This is useful both for struct gsm0808_cell_id and struct gsm0808_cell_id_list2.
  * See also gsm0808_cell_id_name() and gsm0808_cell_id_list_name().
@@ -1192,6 +1599,239 @@ int gsm0808_cell_id_u_name(char *buf, size_t buflen,
 	}
 }
 
+/* Store individual Cell Identifier information in a CGI, without clearing the remaining ones.
+ * This is useful to supplement one CGI with information from more than one Cell Identifier,
+ * which in turn is useful to match Cell Identifiers of differing kinds to each other.
+ * Before first invocation, clear the *dst struct externally, this function does only write those members
+ * that are present in parameter u.
+ */
+static void cell_id_to_cgi(struct osmo_cell_global_id *dst,
+			   enum CELL_IDENT discr, const union gsm0808_cell_id_u *u)
+{
+	switch (discr) {
+	case CELL_IDENT_WHOLE_GLOBAL:
+		*dst = u->global;
+		return;
+
+	case CELL_IDENT_LAC_AND_CI:
+		dst->lai.lac = u->lac_and_ci.lac;
+		dst->cell_identity = u->lac_and_ci.ci;
+		return;
+
+	case CELL_IDENT_CI:
+		dst->cell_identity = u->ci;
+		return;
+
+	case CELL_IDENT_LAI_AND_LAC:
+		dst->lai = u->lai_and_lac;
+		return;
+
+	case CELL_IDENT_LAC:
+		dst->lai.lac = u->lac;
+		return;
+
+	case CELL_IDENT_NO_CELL:
+	case CELL_IDENT_BSS:
+	case CELL_IDENT_UTRAN_PLMN_LAC_RNC:
+	case CELL_IDENT_UTRAN_RNC:
+	case CELL_IDENT_UTRAN_LAC_RNC:
+		/* No values to set. */
+		return;
+	}
+}
+
+/*! Return true if the common information between the two Cell Identifiers match.
+ * For example, if a LAC+CI is compared to LAC, return true if the LAC are the same.
+ * Note that CELL_IDENT_NO_CELL will always return false.
+ * Also CELL_IDENT_BSS will always return false, since this function cannot possibly
+ * know the bounds of the BSS, so the caller must handle CELL_IDENT_BSS specially.
+ * \param[in] discr1  Cell Identifier type.
+ * \param[in] u1  Cell Identifier value.
+ * \param[in] discr2  Other Cell Identifier type.
+ * \param[in] u2  Other Cell Identifier value.
+ * \param[in] exact_match  If true, return true only if the CELL_IDENT types and all values are identical.
+ * \returns True if the common fields of the above match.
+ */
+static bool gsm0808_cell_id_u_match(enum CELL_IDENT discr1, const union gsm0808_cell_id_u *u1,
+				    enum CELL_IDENT discr2, const union gsm0808_cell_id_u *u2,
+				    bool exact_match)
+{
+	struct osmo_cell_global_id a = {};
+	struct osmo_cell_global_id b = {};
+
+	if (exact_match && discr1 != discr2)
+		return false;
+
+	/* First handle the odd wildcard like CELL_IDENT kinds. We can't really match any of these. */
+	switch (discr1) {
+	case CELL_IDENT_NO_CELL:
+	case CELL_IDENT_BSS:
+		return discr1 == discr2;
+	case CELL_IDENT_UTRAN_PLMN_LAC_RNC:
+	case CELL_IDENT_UTRAN_RNC:
+	case CELL_IDENT_UTRAN_LAC_RNC:
+		return false;
+	default:
+		break;
+	}
+	switch (discr2) {
+	case CELL_IDENT_NO_CELL:
+	case CELL_IDENT_UTRAN_PLMN_LAC_RNC:
+	case CELL_IDENT_UTRAN_RNC:
+	case CELL_IDENT_UTRAN_LAC_RNC:
+	case CELL_IDENT_BSS:
+		return false;
+	default:
+		break;
+	}
+
+	/* Enrich both sides to full CGI, then compare those. First set the *other* ID's values in case
+	 * they assign more items. For example:
+	 * u1 = LAC:42
+	 * u2 = LAC+CI:23+5
+	 * 1) a <- LAC+CI:23+5
+	 * 2) a <- LAC:42 so that a = LAC+CI:42+5
+	 * Now we can compare those two and find a mismatch. If the LAC were the same, we would get
+	 * identical LAC+CI and hence a match. */
+
+	cell_id_to_cgi(&a, discr2, u2);
+	cell_id_to_cgi(&a, discr1, u1);
+
+	cell_id_to_cgi(&b, discr1, u1);
+	cell_id_to_cgi(&b, discr2, u2);
+
+	return osmo_cgi_cmp(&a, &b) == 0;
+}
+
+/*! Return true if the common information between the two Cell Identifiers match.
+ * For example, if a LAC+CI is compared to LAC, return true if the LAC are the same.
+ * Note that CELL_IDENT_NO_CELL will always return false.
+ * Also CELL_IDENT_BSS will always return false, since this function cannot possibly
+ * know the bounds of the BSS, so the caller must handle CELL_IDENT_BSS specially.
+ * \param[in] id1  Cell Identifier.
+ * \param[in] id2  Other Cell Identifier.
+ * \param[in] exact_match  If true, return true only if the CELL_IDENT types and all values are identical.
+ * \returns True if the common fields of the above match.
+ */
+bool gsm0808_cell_ids_match(const struct gsm0808_cell_id *id1, const struct gsm0808_cell_id *id2, bool exact_match)
+{
+	return gsm0808_cell_id_u_match(id1->id_discr, &id1->id, id2->id_discr, &id2->id, exact_match);
+}
+
+/*! Find an index in a Cell Identifier list that matches a given single Cell Identifer.
+ * Compare \a id against each entry in \a list using gsm0808_cell_ids_match(), and return the list index
+ * if a match is found. \a match_nr allows iterating all matches in the list. A match_nr <= 0 returns the
+ * first match in the list, match_nr == 1 the second match, etc., and if match_nr exceeds the available
+ * matches in the list, -1 is returned.
+ * \param[in] id  Cell Identifier to match.
+ * \param[in] list  Cell Identifier list to search in.
+ * \param[in] match_nr  Ignore this many matches.
+ * \param[in] exact_match  If true, consider as match only if the CELL_IDENT types and all values are identical.
+ * \returns -1 if no match is found, list index if a match is found (i.e. rc == 0 means a match was found on the first
+ *           entry).
+ */
+int gsm0808_cell_id_matches_list(const struct gsm0808_cell_id *id, const struct gsm0808_cell_id_list2 *list,
+				 unsigned int match_nr, bool exact_match)
+{
+	int i;
+	for (i = 0; i < list->id_list_len; i++) {
+		if (gsm0808_cell_id_u_match(id->id_discr, &id->id, list->id_discr, &list->id_list[i], exact_match)) {
+			if (match_nr)
+				match_nr--;
+			else
+				return i;
+		}
+	}
+	return -1;
+}
+
+/*! Copy information from a CGI to form a Cell Identifier of the specified kind.
+ * \param [out] cid  Compose new Cell Identifier here.
+ * \param [in] id_discr  Which kind of Cell Identifier to compose.
+ * \param [in] cgi  Cell Global Identifier to form the Cell Identifier from.
+ */
+void gsm0808_cell_id_from_cgi(struct gsm0808_cell_id *cid, enum CELL_IDENT id_discr,
+			      const struct osmo_cell_global_id *cgi)
+{
+	*cid = (struct gsm0808_cell_id){
+		.id_discr = id_discr,
+	};
+
+	switch (id_discr) {
+	case CELL_IDENT_WHOLE_GLOBAL:
+		cid->id.global = *cgi;
+		return;
+
+	case CELL_IDENT_LAC_AND_CI:
+		cid->id.lac_and_ci = (struct osmo_lac_and_ci_id){
+			.lac = cgi->lai.lac,
+			.ci = cgi->cell_identity,
+		};
+		return;
+
+	case CELL_IDENT_CI:
+		cid->id.ci = cgi->cell_identity;
+		return;
+
+	case CELL_IDENT_LAI:
+		cid->id.lai_and_lac = cgi->lai;
+		return;
+
+	case CELL_IDENT_LAC:
+		cid->id.lac = cgi->lai.lac;
+		return;
+
+	case CELL_IDENT_NO_CELL:
+	case CELL_IDENT_BSS:
+	case CELL_IDENT_UTRAN_PLMN_LAC_RNC:
+	case CELL_IDENT_UTRAN_RNC:
+	case CELL_IDENT_UTRAN_LAC_RNC:
+	default:
+		return;
+	};
+}
+
+/*! Overwrite parts of cgi with values from a Cell Identifier.
+ * Place only those items given in cid into cgi, leaving other values unchanged.
+ * \param[out] cgi  Cell Global Identity to write to.
+ * \param[in] cid  Cell Identity to read from.
+ * \return a bitmask of items that were set: OSMO_CGI_PART_PLMN | OSMO_CGI_PART_LAC | OSMO_CGI_PART_CI; 0 if nothing was
+ *         written to cgi.
+ */
+int gsm0808_cell_id_to_cgi(struct osmo_cell_global_id *cgi, const struct gsm0808_cell_id *cid)
+{
+	switch (cid->id_discr) {
+	case CELL_IDENT_WHOLE_GLOBAL:
+		*cgi = cid->id.global;
+		return OSMO_CGI_PART_PLMN | OSMO_CGI_PART_LAC | OSMO_CGI_PART_CI;
+
+	case CELL_IDENT_LAC_AND_CI:
+		cgi->lai.lac = cid->id.lac_and_ci.lac;
+		cgi->cell_identity = cid->id.lac_and_ci.ci;
+		return OSMO_CGI_PART_LAC | OSMO_CGI_PART_CI;
+
+	case CELL_IDENT_CI:
+		cgi->cell_identity = cid->id.ci;
+		return OSMO_CGI_PART_CI;
+
+	case CELL_IDENT_LAI:
+		cgi->lai = cid->id.lai_and_lac;
+		return OSMO_CGI_PART_PLMN | OSMO_CGI_PART_LAC;
+
+	case CELL_IDENT_LAC:
+		cgi->lai.lac = cid->id.lac;
+		return OSMO_CGI_PART_LAC;
+
+	case CELL_IDENT_NO_CELL:
+	case CELL_IDENT_BSS:
+	case CELL_IDENT_UTRAN_PLMN_LAC_RNC:
+	case CELL_IDENT_UTRAN_RNC:
+	case CELL_IDENT_UTRAN_LAC_RNC:
+	default:
+		return 0;
+	};
+}
+
 /*! value_string[] for enum CELL_IDENT. */
 const struct value_string gsm0808_cell_id_discr_names[] = {
 	{ CELL_IDENT_WHOLE_GLOBAL, "CGI" },
@@ -1220,8 +1860,7 @@ const struct value_string gsm0808_cell_id_discr_names[] = {
 #define APPEND_STR(fmt, args...) APPEND_THING(snprintf, fmt, ##args)
 #define APPEND_CELL_ID_U(DISCR, U) APPEND_THING(gsm0808_cell_id_u_name, DISCR, U)
 
-static const char *gsm0808_cell_id_name_buf(const struct gsm0808_cell_id *cid,
-					    char *buf, size_t buflen)
+char *gsm0808_cell_id_name_buf(char *buf, size_t buflen, const struct gsm0808_cell_id *cid)
 {
 	char *pos = buf;
 	int total_len = 0;
@@ -1237,8 +1876,8 @@ static const char *gsm0808_cell_id_name_buf(const struct gsm0808_cell_id *cid,
  */
 const char *gsm0808_cell_id_name(const struct gsm0808_cell_id *cid)
 {
-	static char buf[64];
-	return gsm0808_cell_id_name_buf(cid, buf, sizeof(buf));
+	static __thread char buf[64];
+	return gsm0808_cell_id_name_buf(buf, sizeof(buf), cid);
 }
 
 /*! Like gsm0808_cell_id_name() but uses a different static buffer.
@@ -1247,8 +1886,16 @@ const char *gsm0808_cell_id_name(const struct gsm0808_cell_id *cid)
  */
 const char *gsm0808_cell_id_name2(const struct gsm0808_cell_id *cid)
 {
-	static char buf[64];
-	return gsm0808_cell_id_name_buf(cid, buf, sizeof(buf));
+	static __thread char buf[64];
+	return gsm0808_cell_id_name_buf(buf, sizeof(buf), cid);
+}
+
+char *gsm0808_cell_id_name_c(const void *ctx, const struct gsm0808_cell_id *cid)
+{
+	char *buf = talloc_size(ctx, 64);
+	if (!buf)
+		return NULL;
+	return gsm0808_cell_id_name_buf(buf, 64, cid);
 }
 
 /*! Return a human readable representation of the Cell Identifier List, like
@@ -1296,21 +1943,43 @@ int gsm0808_cell_id_list_name_buf(char *buf, size_t buflen, const struct gsm0808
  * See also gsm0808_cell_id_list_name_buf(). */
 const char *gsm0808_cell_id_list_name(const struct gsm0808_cell_id_list2 *cil)
 {
-	static char buf[1024];
+	static __thread char buf[1024];
 	gsm0808_cell_id_list_name_buf(buf, sizeof(buf), cil);
 	return buf;
 }
 
+char *gsm0808_cell_id_list_name_c(const void *ctx, const struct gsm0808_cell_id_list2 *cil)
+{
+	char *buf = talloc_size(ctx, 1024);
+	if (!buf)
+		return NULL;
+	gsm0808_cell_id_list_name_buf(buf, 1024, cil);
+	return buf;
+}
+
 #undef APPEND_STR
 #undef APPEND_CELL_ID_U
 
-const char *gsm0808_channel_type_name(const struct gsm0808_channel_type *ct)
+char *gsm0808_channel_type_name_buf(char *buf, size_t buf_len, const struct gsm0808_channel_type *ct)
 {
-	static char buf[128];
-	snprintf(buf, sizeof(buf), "ch_indctr=0x%x ch_rate_type=0x%x perm_spch=%s",
+	snprintf(buf, buf_len, "ch_indctr=0x%x ch_rate_type=0x%x perm_spch=%s",
 		 ct->ch_indctr, ct->ch_rate_type,
 		 osmo_hexdump(ct->perm_spch, ct->perm_spch_len));
 	return buf;
 }
 
+const char *gsm0808_channel_type_name(const struct gsm0808_channel_type *ct)
+{
+	static __thread char buf[128];
+	return gsm0808_channel_type_name_buf(buf, sizeof(buf), ct);
+}
+
+char *gsm0808_channel_type_name_c(const void *ctx, const struct gsm0808_channel_type *ct)
+{
+	char *buf = talloc_size(ctx, 128);
+	if (!buf)
+		return NULL;
+	return gsm0808_channel_type_name_buf(buf, 128, ct);
+}
+
 /*! @} */
