@@ -209,37 +209,76 @@ int rsl_dec_chan_nr(uint8_t chan_nr, uint8_t *type, uint8_t *subch, uint8_t *tim
 	} else if ((chan_nr & 0xf8) == RSL_CHAN_OSMO_PDCH) {
 		*type = RSL_CHAN_OSMO_PDCH;
 		*subch = 0;
+	} else if ((chan_nr & 0xf8) == RSL_CHAN_OSMO_CBCH4) {
+		*type = RSL_CHAN_OSMO_CBCH4;
+		*subch = 0;
+	} else if ((chan_nr & 0xf8) == RSL_CHAN_OSMO_CBCH8) {
+		*type = RSL_CHAN_OSMO_CBCH8;
+		*subch = 0;
 	} else
 		return -EINVAL;
 
 	return 0;
 }
 
-/*! Get human-readable string for RSL channel number */
-const char *rsl_chan_nr_str(uint8_t chan_nr)
+/*! Get human-readable string for RSL channel number, in caller-provided buffer.
+ *  \param[out] buf caller-provided output buffer
+ *  \param[in] buf_len size of buf in bytes
+ *  \param[in] chan_nr channel number to be stringified
+ *  \returns buf with string
+ */
+char *rsl_chan_nr_str_buf(char *buf, size_t buf_len, uint8_t chan_nr)
 {
-	static char str[20];
 	int ts = chan_nr & 7;
 	uint8_t cbits = chan_nr >> 3;
 
 	if (cbits == 0x01)
-		sprintf(str, "TCH/F on TS%d", ts);
+		snprintf(buf, buf_len, "TCH/F on TS%d", ts);
 	else if ((cbits & 0x1e) == 0x02)
-		sprintf(str, "TCH/H(%u) on TS%d", cbits & 0x01, ts);
+		snprintf(buf, buf_len, "TCH/H(%u) on TS%d", cbits & 0x01, ts);
 	else if ((cbits & 0x1c) == 0x04)
-		sprintf(str, "SDCCH/4(%u) on TS%d", cbits & 0x03, ts);
+		snprintf(buf, buf_len, "SDCCH/4(%u) on TS%d", cbits & 0x03, ts);
 	else if ((cbits & 0x18) == 0x08)
-		sprintf(str, "SDCCH/8(%u) on TS%d", cbits & 0x07, ts);
+		snprintf(buf, buf_len, "SDCCH/8(%u) on TS%d", cbits & 0x07, ts);
 	else if (cbits == 0x10)
-		sprintf(str, "BCCH on TS%d", ts);
+		snprintf(buf, buf_len, "BCCH on TS%d", ts);
 	else if (cbits == 0x11)
-		sprintf(str, "RACH on TS%d", ts);
+		snprintf(buf, buf_len, "RACH on TS%d", ts);
 	else if (cbits == 0x12)
-		sprintf(str, "PCH/AGCH on TS%d", ts);
+		snprintf(buf, buf_len, "PCH/AGCH on TS%d", ts);
+	else if (cbits == 0x18)
+		snprintf(buf, buf_len, "PDCH on TS%d", ts);
+	else if (cbits == 0x19)
+		snprintf(buf, buf_len, "CBCH(SDCCH/4) on TS%d", ts);
+	else if (cbits == 0x1a)
+		snprintf(buf, buf_len, "CBCH(SDCCH/8) on TS%d", ts);
 	else
-		sprintf(str, "UNKNOWN on TS%d", ts);
+		snprintf(buf, buf_len, "UNKNOWN on TS%d", ts);
 
-        return str;
+        return buf;
+}
+
+/*! Get human-readable string for RSL channel number, in static buffer.
+ *  \param[in] chan_nr channel number to be stringified
+ *  \returns buf with string
+ */
+const char *rsl_chan_nr_str(uint8_t chan_nr)
+{
+	static __thread char str[20];
+	return rsl_chan_nr_str_buf(str, sizeof(str), chan_nr);
+}
+
+/*! Get human-readable string for RSL channel number, in dynamically-allocated buffer.
+ *  \param[in] ctx talloc context from which to allocate output buffer
+ *  \param[in] chan_nr channel number to be stringified
+ *  \returns dynamically-allocated buffer with string representation
+ */
+char *rsl_chan_nr_str_c(const void *ctx, uint8_t chan_nr)
+{
+	char *str = talloc_size(ctx, 20);
+	if (!str)
+		return NULL;
+	return rsl_chan_nr_str_buf(str, 20, chan_nr);
 }
 
 static const struct value_string rsl_err_vals[] = {
@@ -256,6 +295,9 @@ static const struct value_string rsl_err_vals[] = {
 	{ RSL_ERR_CCCH_OVERLOAD,	"CCCH Overload" },
 	{ RSL_ERR_ACCH_OVERLOAD,	"ACCH Overload" },
 	{ RSL_ERR_PROCESSOR_OVERLOAD,	"Processor Overload" },
+	{ RSL_ERR_BTS_NOT_EQUIPPED,     "BTS not equipped" },
+	{ RSL_ERR_REMOTE_TRANSC_FAIL,   "Remote Transcoder Failure" },
+	{ RSL_ERR_NOTIFICATION_OVERFL,  "Notification Overflow" },
 	{ RSL_ERR_RES_UNAVAIL,		"Resource not available, unspecified" },
 	{ RSL_ERR_TRANSC_UNAVAIL,	"Transcoding not available" },
 	{ RSL_ERR_SERV_OPT_UNAVAIL,	"Service or Option not available" },
@@ -350,6 +392,7 @@ static const struct value_string rsl_msgt_names[] = {
 	{ RSL_MT_TFO_REP,		"TFO_REP" },
 	{ RSL_MT_TFO_MOD_REQ,		"TFO_MOD_REQ" },
 	{ RSL_MT_LOCATION_INFO,		"LOCATION_INFO" },
+	{ RSL_MT_OSMO_ETWS_CMD,		"OSMO_ETWS_CMD" },
 	{ 0,				NULL }
 };
 
