@@ -35,7 +35,6 @@
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
-#include <ctype.h>
 
 #ifdef HAVE_STRINGS_H
 #include <strings.h>
@@ -43,6 +42,7 @@
 #include <time.h>
 #include <sys/time.h>
 #include <errno.h>
+#include <pthread.h>
 
 #include <osmocom/core/talloc.h>
 #include <osmocom/core/utils.h>
@@ -61,13 +61,60 @@ osmo_static_assert(_LOG_FLT_COUNT <= 8*sizeof(((struct log_target*)NULL)->filter
 struct log_info *osmo_log_info;
 
 static struct log_context log_context;
-static void *tall_log_ctx = NULL;
+void *tall_log_ctx = NULL;
 LLIST_HEAD(osmo_log_target_list);
 
-#define LOGLEVEL_DEFS	6	/* Number of loglevels.*/
+#if (!EMBEDDED)
+/*! This mutex must be held while using osmo_log_target_list or any of its
+  log_targets in a multithread program. Prevents race conditions between threads
+  like producing unordered timestamps or VTY deleting a target while another
+  thread is writing to it */
+static pthread_mutex_t osmo_log_tgt_mutex;
+static bool osmo_log_tgt_mutex_on = false;
 
-static const struct value_string loglevel_strs[LOGLEVEL_DEFS+1] = {
-	{ 0,		"EVERYTHING" },
+/*! Enable multithread support (mutex) in libosmocore logging system.
+ * Must be called by processes willing to use logging subsystem from several
+ * threads. Once enabled, it's not possible to disable it again.
+ */
+void log_enable_multithread(void) {
+	if (osmo_log_tgt_mutex_on)
+		return;
+	pthread_mutex_init(&osmo_log_tgt_mutex, NULL);
+	osmo_log_tgt_mutex_on = true;
+}
+
+/*! Acquire the osmo_log_tgt_mutex. Don't use this function directly, always use
+ *  macro log_tgt_mutex_lock() instead.
+ */
+void log_tgt_mutex_lock_impl(void) {
+	/* These lines are useful to debug scenarios where there's only 1 thread
+	   and a double lock appears, for instance during startup and some
+	   unlock() missing somewhere:
+	if (osmo_log_tgt_mutex_on && pthread_mutex_trylock(&osmo_log_tgt_mutex) != 0)
+		osmo_panic("acquiring already locked mutex!\n");
+	return;
+	*/
+
+	if (osmo_log_tgt_mutex_on)
+		pthread_mutex_lock(&osmo_log_tgt_mutex);
+}
+
+/*! Release the osmo_log_tgt_mutex. Don't use this function directly, always use
+ *  macro log_tgt_mutex_unlock() instead.
+ */
+void log_tgt_mutex_unlock_impl(void) {
+	if (osmo_log_tgt_mutex_on)
+		pthread_mutex_unlock(&osmo_log_tgt_mutex);
+}
+
+#else /* if (!EMBEDDED) */
+#pragma message ("logging multithread support disabled in embedded build")
+void log_enable_multithread(void) {}
+void log_tgt_mutex_lock_impl(void) {}
+void log_tgt_mutex_unlock_impl(void) {}
+#endif /* if (!EMBEDDED) */
+
+const struct value_string loglevel_strs[] = {
 	{ LOGL_DEBUG,	"DEBUG" },
 	{ LOGL_INFO,	"INFO" },
 	{ LOGL_NOTICE,	"NOTICE" },
@@ -116,7 +163,7 @@ static const struct log_info_cat internal_cat[OSMO_NUM_DLIB] = {
 		.name = "DLSMS",
 		.description = "Layer3 Short Message Service (SMS)",
 		.enabled = 1, .loglevel = LOGL_NOTICE,
-		.color = "\033[1;38m",
+		.color = OSMO_LOGCOLOR_BRIGHTWHITE,
 	},
 	[INT2IDX(DLCTRL)] = {
 		.name = "DLCTRL",
@@ -173,21 +220,14 @@ static const struct log_info_cat internal_cat[OSMO_NUM_DLIB] = {
 		.description = "libosmo-netif Jitter Buffer",
 		.enabled = 1, .loglevel = LOGL_NOTICE,
 	},
+	[INT2IDX(DLRSPRO)] = {
+		.name = "DLRSPRO",
+		.description = "Remote SIM protocol",
+		.enabled = 1, .loglevel = LOGL_NOTICE,
+	},
 };
 
-/*! descriptive string for each log level */
-/* You have to keep this in sync with the structure loglevel_strs. */
-static const char *loglevel_descriptions[LOGLEVEL_DEFS+1] = {
-	"Don't use. It doesn't log anything",
-	"Log debug messages and higher levels",
-	"Log informational messages and higher levels",
-	"Log noticeable messages and higher levels",
-	"Log error messages and higher levels",
-	"Log only fatal messages",
-	NULL,
-};
-
-static void assert_loginfo(const char *src)
+void assert_loginfo(const char *src)
 {
 	if (!osmo_log_info) {
 		fprintf(stderr, "ERROR: osmo_log_info == NULL! "
@@ -203,7 +243,7 @@ static int subsys_lib2index(int subsys)
 }
 
 /*! Parse a human-readable log level into a numeric value
- *  \param lvl[in] zero-terminated string containing log level name
+ *  \param[in] lvl zero-terminated string containing log level name
  *  \returns numeric log level
  */
 int log_parse_level(const char *lvl)
@@ -212,7 +252,7 @@ int log_parse_level(const char *lvl)
 }
 
 /*! convert a numeric log level into human-readable string
- *  \param lvl[in] numeric log level
+ *  \param[in] lvl numeric log level
  *  \returns zero-terminated string (log level name)
  */
 const char *log_level_str(unsigned int lvl)
@@ -304,11 +344,11 @@ static const char* color(int subsys)
 }
 
 static const struct value_string level_colors[] = {
-	{ LOGL_DEBUG, "\033[1;34m" },
-	{ LOGL_INFO, "\033[1;32m" },
-	{ LOGL_NOTICE, "\033[1;33m" },
-	{ LOGL_ERROR, "\033[1;31m" },
-	{ LOGL_FATAL, "\033[1;31m" },
+	{ LOGL_DEBUG, OSMO_LOGCOLOR_BLUE },
+	{ LOGL_INFO, OSMO_LOGCOLOR_GREEN },
+	{ LOGL_NOTICE, OSMO_LOGCOLOR_YELLOW },
+	{ LOGL_ERROR, OSMO_LOGCOLOR_RED },
+	{ LOGL_FATAL, OSMO_LOGCOLOR_RED },
 	{ 0, NULL }
 };
 
@@ -370,21 +410,25 @@ static void _output(struct log_target *target, unsigned int subsys,
 			OSMO_SNPRINTF_RET(ret, rem, offset, len);
 #endif
 		} else if (target->print_timestamp) {
-			char *timestr;
 			time_t tm;
-			tm = time(NULL);
-			timestr = ctime(&tm);
-			timestr[strlen(timestr)-1] = '\0';
-			ret = snprintf(buf + offset, rem, "%s ", timestr);
-			if (ret < 0)
+			if ((tm = time(NULL)) == (time_t) -1)
 				goto err;
+			/* Get human-readable representation of time.
+			   man ctime: we need at least 26 bytes in buf */
+			if (rem < 26 || !ctime_r(&tm, buf + offset))
+				goto err;
+			ret = strlen(buf + offset);
+			if (ret <= 0)
+				goto err;
+			/* Get rid of useless final '\n' added by ctime_r. We want a space instead. */
+			buf[offset + ret - 1] = ' ';
 			OSMO_SNPRINTF_RET(ret, rem, offset, len);
 		}
 		if (target->print_category) {
 			ret = snprintf(buf + offset, rem, "%s%s%s%s ",
 				       target->use_color ? level_color(level) : "",
 				       log_category_name(subsys),
-				       target->use_color ? "\033[0;m" : "",
+				       target->use_color ? OSMO_LOGCOLOR_END : "",
 				       c_subsys ? c_subsys : "");
 			if (ret < 0)
 				goto err;
@@ -394,7 +438,7 @@ static void _output(struct log_target *target, unsigned int subsys,
 			ret = snprintf(buf + offset, rem, "%s%s%s%s ",
 				       target->use_color ? level_color(level) : "",
 				       log_level_str(level),
-				       target->use_color ? "\033[0;m" : "",
+				       target->use_color ? OSMO_LOGCOLOR_END : "",
 				       c_subsys ? c_subsys : "");
 			if (ret < 0)
 				goto err;
@@ -406,21 +450,24 @@ static void _output(struct log_target *target, unsigned int subsys,
 				goto err;
 			OSMO_SNPRINTF_RET(ret, rem, offset, len);
 		}
-		switch (target->print_filename2) {
-		case LOG_FILENAME_NONE:
-			break;
-		case LOG_FILENAME_PATH:
-			ret = snprintf(buf + offset, rem, "%s:%d ", file, line);
-			if (ret < 0)
-				goto err;
-			OSMO_SNPRINTF_RET(ret, rem, offset, len);
-			break;
-		case LOG_FILENAME_BASENAME:
-			ret = snprintf(buf + offset, rem, "%s:%d ", const_basename(file), line);
-			if (ret < 0)
-				goto err;
-			OSMO_SNPRINTF_RET(ret, rem, offset, len);
-			break;
+
+		if (target->print_filename_pos == LOG_FILENAME_POS_HEADER_END) {
+			switch (target->print_filename2) {
+			case LOG_FILENAME_NONE:
+				break;
+			case LOG_FILENAME_PATH:
+				ret = snprintf(buf + offset, rem, "%s:%d ", file, line);
+				if (ret < 0)
+					goto err;
+				OSMO_SNPRINTF_RET(ret, rem, offset, len);
+				break;
+			case LOG_FILENAME_BASENAME:
+				ret = snprintf(buf + offset, rem, "%s:%d ", const_basename(file), line);
+				if (ret < 0)
+					goto err;
+				OSMO_SNPRINTF_RET(ret, rem, offset, len);
+				break;
+			}
 		}
 	}
 	ret = vsnprintf(buf + offset, rem, format, ap);
@@ -428,8 +475,33 @@ static void _output(struct log_target *target, unsigned int subsys,
 		goto err;
 	OSMO_SNPRINTF_RET(ret, rem, offset, len);
 
+	/* For LOG_FILENAME_POS_LAST, print the source file info only when the caller ended the log
+	 * message in '\n'. If so, nip the last '\n' away, insert the source file info and re-append an
+	 * '\n'. All this to allow LOGP("start..."); LOGPC("...end\n") constructs. */
+	if (target->print_filename_pos == LOG_FILENAME_POS_LINE_END
+	    && offset > 0 && buf[offset-1] == '\n') {
+		switch (target->print_filename2) {
+		case LOG_FILENAME_NONE:
+			break;
+		case LOG_FILENAME_PATH:
+			offset --;
+			ret = snprintf(buf + offset, rem, " (%s:%d)\n", file, line);
+			if (ret < 0)
+				goto err;
+			OSMO_SNPRINTF_RET(ret, rem, offset, len);
+			break;
+		case LOG_FILENAME_BASENAME:
+			offset --;
+			ret = snprintf(buf + offset, rem, " (%s:%d)\n", const_basename(file), line);
+			if (ret < 0)
+				goto err;
+			OSMO_SNPRINTF_RET(ret, rem, offset, len);
+			break;
+		}
+	}
+
 	if (target->use_color) {
-		ret = snprintf(buf + offset, rem, "\033[0;m");
+		ret = snprintf(buf + offset, rem, OSMO_LOGCOLOR_END);
 		if (ret < 0)
 			goto err;
 		OSMO_SNPRINTF_RET(ret, rem, offset, len);
@@ -511,6 +583,8 @@ void osmo_vlogp(int subsys, int level, const char *file, int line,
 
 	subsys = map_subsys(subsys);
 
+	log_tgt_mutex_lock();
+
 	llist_for_each_entry(tar, &osmo_log_target_list, entry) {
 		va_list bp;
 
@@ -527,6 +601,8 @@ void osmo_vlogp(int subsys, int level, const char *file, int line,
 			_output(tar, subsys, level, file, line, cont, format, bp);
 		va_end(bp);
 	}
+
+	log_tgt_mutex_unlock();
 }
 
 /*! logging function used by DEBUGP() macro
@@ -668,7 +744,7 @@ void log_set_print_filename(struct log_target *target, int print_filename)
 
 /*! Enable or disable printing of the filename while logging.
  *  \param[in] target Log target to be affected.
- *  \param[in] print_filename An LOG_FILENAME_* enum value.
+ *  \param[in] lft An LOG_FILENAME_* enum value.
  * LOG_FILENAME_NONE omits the source file and line information from logs.
  * LOG_FILENAME_PATH prints the entire source file path as passed to LOGP macros.
  */
@@ -677,9 +753,20 @@ void log_set_print_filename2(struct log_target *target, enum log_filename_type l
 	target->print_filename2 = lft;
 }
 
+/*! Set the position where on a log line the source file info should be logged.
+ *  \param[in] target Log target to be affected.
+ *  \param[in] pos A LOG_FILENAME_POS_* enum value.
+ * LOG_FILENAME_POS_DEFAULT logs just before the caller supplied log message.
+ * LOG_FILENAME_POS_LAST logs only at the end of a log line, where the caller issued an '\n' to end the
+ */
+void log_set_print_filename_pos(struct log_target *target, enum log_filename_pos pos)
+{
+	target->print_filename_pos = pos;
+}
+
 /*! Enable or disable printing of the category name
  *  \param[in] target Log target to be affected
- *  \param[in] print_catname Enable (1) or disable (0) filenames
+ *  \param[in] print_category Enable (1) or disable (0) filenames
  *
  *  Print the category/subsys name in front of every log message.
  */
@@ -699,7 +786,7 @@ void log_set_print_category_hex(struct log_target *target, int print_category_he
 
 /*! Enable or disable printing of the log level name.
  *  \param[in] target Log target to be affected
- *  \param[in] print_catname Enable (1) or disable (0) filenames
+ *  \param[in] print_level Enable (1) or disable (0) log level name
  *
  *  Print the log level name in front of every log message.
  */
@@ -838,6 +925,7 @@ struct log_target *log_target_create_file(const char *fname)
  *  \param[in] type Log target type
  *  \param[in] fname File name
  *  \returns Log target (if found), NULL otherwise
+ *  Must be called with mutex osmo_log_tgt_mutex held, see log_tgt_mutex_lock.
  */
 struct log_target *log_target_find(int type, const char *fname)
 {
@@ -863,7 +951,7 @@ struct log_target *log_target_find(int type, const char *fname)
 }
 
 /*! Unregister, close and delete a log target
- *  \param target[in] log target to unregister, close and delete */
+ *  \param[in] target log target to unregister, close and delete */
 void log_target_destroy(struct log_target *target)
 {
 
@@ -910,6 +998,8 @@ int log_targets_reopen(void)
 	struct log_target *tar;
 	int rc = 0;
 
+	log_tgt_mutex_lock();
+
 	llist_for_each_entry(tar, &osmo_log_target_list, entry) {
 		switch (tar->type) {
 		case LOG_TGT_TYPE_FILE:
@@ -921,160 +1011,25 @@ int log_targets_reopen(void)
 		}
 	}
 
+	log_tgt_mutex_unlock();
+
 	return rc;
 }
 
-/*! Generates the logging command string for VTY
- *  \param[in] unused_info Deprecated parameter, no longer used!
- *  \returns vty command string for use by VTY command node
- */
-const char *log_vty_command_string()
-{
-	struct log_info *info = osmo_log_info;
-	int len = 0, offset = 0, ret, i, rem;
-	int size = strlen("logging level (all|) ()") + 1;
-	char *str;
-
-	assert_loginfo(__func__);
-
-	for (i = 0; i < info->num_cat; i++) {
-		if (info->cat[i].name == NULL)
-			continue;
-		size += strlen(info->cat[i].name) + 1;
-	}
-
-	for (i = 0; i < LOGLEVEL_DEFS; i++)
-		size += strlen(loglevel_strs[i].str) + 1;
-
-	rem = size;
-	str = talloc_zero_size(tall_log_ctx, size);
-	if (!str)
-		return NULL;
-
-	ret = snprintf(str + offset, rem, "logging level (all|");
-	if (ret < 0)
-		goto err;
-	OSMO_SNPRINTF_RET(ret, rem, offset, len);
-
-	for (i = 0; i < info->num_cat; i++) {
-		if (info->cat[i].name) {
-			int j, name_len = strlen(info->cat[i].name)+1;
-			char name[name_len];
-
-			for (j = 0; j < name_len; j++)
-				name[j] = tolower((unsigned char)info->cat[i].name[j]);
-
-			name[name_len-1] = '\0';
-			ret = snprintf(str + offset, rem, "%s|", name+1);
-			if (ret < 0)
-				goto err;
-			OSMO_SNPRINTF_RET(ret, rem, offset, len);
-		}
-	}
-	offset--;	/* to remove the trailing | */
-	rem++;
-
-	ret = snprintf(str + offset, rem, ") (");
-	if (ret < 0)
-		goto err;
-	OSMO_SNPRINTF_RET(ret, rem, offset, len);
-
-	for (i = 0; i < LOGLEVEL_DEFS; i++) {
-		int j, loglevel_str_len = strlen(loglevel_strs[i].str)+1;
-		char loglevel_str[loglevel_str_len];
-
-		for (j = 0; j < loglevel_str_len; j++)
-			loglevel_str[j] = tolower((unsigned char)loglevel_strs[i].str[j]);
-
-		loglevel_str[loglevel_str_len-1] = '\0';
-		ret = snprintf(str + offset, rem, "%s|", loglevel_str);
-		if (ret < 0)
-			goto err;
-		OSMO_SNPRINTF_RET(ret, rem, offset, len);
-	}
-	offset--;	/* to remove the trailing | */
-	rem++;
-
-	ret = snprintf(str + offset, rem, ")");
-	if (ret < 0)
-		goto err;
-	OSMO_SNPRINTF_RET(ret, rem, offset, len);
-err:
-	str[size-1] = '\0';
-	return str;
-}
-
-/*! Generates the logging command description for VTY
- *  \param[in] unused_info Deprecated parameter, no longer used!
- *  \returns logging command description for use by VTY command node
- */
-const char *log_vty_command_description()
-{
-	struct log_info *info = osmo_log_info;
-	char *str;
-	int i, ret, len = 0, offset = 0, rem;
-	unsigned int size =
-		strlen(LOGGING_STR
-		       "Set the log level for a specified category\n") + 1;
-
-	assert_loginfo(__func__);
-
-	for (i = 0; i < info->num_cat; i++) {
-		if (info->cat[i].name == NULL)
-			continue;
-		size += strlen(info->cat[i].description) + 1;
-	}
-
-	for (i = 0; i < LOGLEVEL_DEFS; i++)
-		size += strlen(loglevel_descriptions[i]) + 1;
-
-	size += strlen("Global setting for all subsystems") + 1;
-	rem = size;
-	str = talloc_zero_size(tall_log_ctx, size);
-	if (!str)
-		return NULL;
-
-	ret = snprintf(str + offset, rem, LOGGING_STR
-			"Set the log level for a specified category\n");
-	if (ret < 0)
-		goto err;
-	OSMO_SNPRINTF_RET(ret, rem, offset, len);
-
-	ret = snprintf(str + offset, rem,
-			"Global setting for all subsystems\n");
-	if (ret < 0)
-		goto err;
-	OSMO_SNPRINTF_RET(ret, rem, offset, len);
-
-	for (i = 0; i < info->num_cat; i++) {
-		if (info->cat[i].name == NULL)
-			continue;
-		ret = snprintf(str + offset, rem, "%s\n",
-				info->cat[i].description);
-		if (ret < 0)
-			goto err;
-		OSMO_SNPRINTF_RET(ret, rem, offset, len);
-	}
-	for (i = 0; i < LOGLEVEL_DEFS; i++) {
-		ret = snprintf(str + offset, rem, "%s\n",
-				loglevel_descriptions[i]);
-		if (ret < 0)
-			goto err;
-		OSMO_SNPRINTF_RET(ret, rem, offset, len);
-	}
-err:
-	str[size-1] = '\0';
-	return str;
-}
-
 /*! Initialize the Osmocom logging core
- *  \param[in] inf Information regarding logging categories
- *  \param[in] ctx \ref talloc context for logging allocations
+ *  \param[in] inf Information regarding logging categories, could be NULL
+ *  \param[in] ctx talloc context for logging allocations
  *  \returns 0 in case of success, negative in case of error
+ *
+ *  If inf is NULL then only library-internal categories are initialized.
  */
 int log_init(const struct log_info *inf, void *ctx)
 {
 	int i;
+	struct log_info_cat *cat_ptr;
+
+	/* Ensure that log_init is not called multiple times */
+	OSMO_ASSERT(tall_log_ctx == NULL)
 
 	tall_log_ctx = talloc_named_const(ctx, 1, "logging");
 	if (!tall_log_ctx)
@@ -1084,34 +1039,44 @@ int log_init(const struct log_info *inf, void *ctx)
 	if (!osmo_log_info)
 		return -ENOMEM;
 
-	osmo_log_info->filter_fn = inf->filter_fn;
-	osmo_log_info->num_cat_user = inf->num_cat;
-	/* total number = number of user cat + library cat */
-	osmo_log_info->num_cat = inf->num_cat + ARRAY_SIZE(internal_cat);
+	osmo_log_info->num_cat = ARRAY_SIZE(internal_cat);
 
-	osmo_log_info->cat = talloc_zero_array(osmo_log_info,
-					struct log_info_cat,
-					osmo_log_info->num_cat);
-	if (!osmo_log_info->cat) {
+	if (inf) {
+		osmo_log_info->filter_fn = inf->filter_fn;
+		osmo_log_info->num_cat_user = inf->num_cat;
+		osmo_log_info->num_cat += inf->num_cat;
+	}
+
+	cat_ptr = talloc_zero_array(osmo_log_info, struct log_info_cat,
+				    osmo_log_info->num_cat);
+	if (!cat_ptr) {
 		talloc_free(osmo_log_info);
 		osmo_log_info = NULL;
 		return -ENOMEM;
 	}
 
-	/* copy over the user part */
-	for (i = 0; i < inf->num_cat; i++) {
-		memcpy((struct log_info_cat *) &osmo_log_info->cat[i],
-			&inf->cat[i],
-			sizeof(struct log_info_cat));
+	/* copy over the user part and sanitize loglevel */
+	if (inf) {
+		for (i = 0; i < inf->num_cat; i++) {
+			memcpy(&cat_ptr[i], &inf->cat[i],
+			       sizeof(struct log_info_cat));
+
+			/* Make sure that the loglevel is set to NOTICE in case
+			 * no loglevel has been preset. */
+			if (!cat_ptr[i].loglevel) {
+				cat_ptr[i].loglevel = LOGL_NOTICE;
+			}
+		}
 	}
 
 	/* copy over the library part */
 	for (i = 0; i < ARRAY_SIZE(internal_cat); i++) {
 		unsigned int cn = osmo_log_info->num_cat_user + i;
-		memcpy((struct log_info_cat *) &osmo_log_info->cat[cn],
-			&internal_cat[i], sizeof(struct log_info_cat));
+		memcpy(&cat_ptr[cn], &internal_cat[i], sizeof(struct log_info_cat));
 	}
 
+	osmo_log_info->cat = cat_ptr;
+
 	return 0;
 }
 
@@ -1121,6 +1086,8 @@ void log_fini(void)
 {
 	struct log_target *tar, *tar2;
 
+	log_tgt_mutex_lock();
+
 	llist_for_each_entry_safe(tar, tar2, &osmo_log_target_list, entry)
 		log_target_destroy(tar);
 
@@ -1128,6 +1095,8 @@ void log_fini(void)
 	osmo_log_info = NULL;
 	talloc_free(tall_log_ctx);
 	tall_log_ctx = NULL;
+
+	log_tgt_mutex_unlock();
 }
 
 /*! Check whether a log entry will be generated.
@@ -1142,15 +1111,19 @@ int log_check_level(int subsys, unsigned int level)
 
 	/* TODO: The following could/should be cached (update on config) */
 
+	log_tgt_mutex_lock();
+
 	llist_for_each_entry(tar, &osmo_log_target_list, entry) {
 		if (!should_log_to_target(tar, subsys, level))
 			continue;
 
 		/* This might get logged (ignoring filters) */
+		log_tgt_mutex_unlock();
 		return 1;
 	}
 
 	/* We are sure, that this will not be logged. */
+	log_tgt_mutex_unlock();
 	return 0;
 }
 
