@@ -206,7 +206,7 @@ static unsigned int rate_ctr_get_unused_name_idx(const char *name)
 }
 
 /*! Allocate a new group of counters according to description
- *  \param[in] ctx \ref talloc context
+ *  \param[in] ctx parent talloc context
  *  \param[in] desc Rate counter group description
  *  \param[in] idx Index of new counter group
  */
@@ -255,7 +255,11 @@ struct rate_ctr_group *rate_ctr_group_alloc(void *ctx,
 /*! Free the memory for the specified group of counters */
 void rate_ctr_group_free(struct rate_ctr_group *grp)
 {
-	llist_del(&grp->list);
+	if (!grp)
+		return;
+
+	if (!llist_empty(&grp->list))
+		llist_del(&grp->list);
 	talloc_free(grp);
 }
 
@@ -381,9 +385,9 @@ const struct rate_ctr *rate_ctr_get_by_name(const struct rate_ctr_group *ctrg, c
 }
 
 /*! Iterate over each counter in group and call function
- *  \param[in] counter group over whose counter to iterate
+ *  \param[in] ctrg counter group over which to iterate
  *  \param[in] handle_counter function pointer
- *  \param[in] data Data to hand transparently to \ref handle_counter
+ *  \param[in] data Data to hand transparently to handle_counter()
  *  \returns 0 on success; negative otherwise
  */
 int rate_ctr_for_each_counter(struct rate_ctr_group *ctrg,
@@ -405,7 +409,7 @@ int rate_ctr_for_each_counter(struct rate_ctr_group *ctrg,
 
 /*! Iterate over all counter groups
  *  \param[in] handle_group function pointer of callback function
- *  \param[in] data Data to hand transparently to \ref handle_group
+ *  \param[in] data Data to hand transparently to handle_group()
  *  \returns 0 on success; negative otherwise
  */
 int rate_ctr_for_each_group(rate_ctr_group_handler_t handle_group, void *data)
@@ -422,4 +426,25 @@ int rate_ctr_for_each_group(rate_ctr_group_handler_t handle_group, void *data)
 	return rc;
 }
 
+/*! Reset a rate counter back to zero
+ *  \param[in] ctr counter to reset
+ */
+void rate_ctr_reset(struct rate_ctr *ctr)
+{
+        memset(ctr, 0, sizeof(*ctr));
+}
+
+/*! Reset all counters in a group
+ *  \param[in] ctrg counter group to reset
+ */
+void rate_ctr_group_reset(struct rate_ctr_group *ctrg)
+{
+	int i;
+
+	for (i = 0; i < ctrg->desc->num_ctr; i++) {
+		struct rate_ctr *ctr = &ctrg->ctr[i];
+                rate_ctr_reset(ctr);
+	}
+}
+
 /*! @} */
