@@ -35,6 +35,9 @@
 #include <osmocom/core/select.h>
 #include <osmocom/core/linuxlist.h>
 #include <osmocom/core/timer.h>
+#include <osmocom/core/logging.h>
+#include <osmocom/core/talloc.h>
+#include <osmocom/core/utils.h>
 
 #include "../config.h"
 
@@ -47,14 +50,16 @@
  *
  * \file select.c */
 
-static int maxfd = 0;
-static LLIST_HEAD(osmo_fds);
-static int unregistered_count;
+/* keep a set of file descriptors per-thread, so that each thread can have its own
+ * distinct set of file descriptors to interact with */
+static __thread int maxfd = 0;
+static __thread struct llist_head osmo_fds; /* TLS cannot use LLIST_HEAD() */
+static __thread int unregistered_count;
 
 /*! Set up an osmo-fd. Will not register it.
  *  \param[inout] ofd Osmo FD to be set-up
  *  \param[in] fd OS-level file descriptor number
- *  \param[in] when bit-mask of BSC_FD_{READ,WRITE,EXECEPT}
+ *  \param[in] when bit-mask of OSMO_FD_{READ,WRITE,EXECEPT}
  *  \param[in] cb Call-back function to be called
  *  \param[in] data Private context pointer
  *  \param[in] priv_nr Private number
@@ -116,7 +121,7 @@ int osmo_fd_register(struct osmo_fd *fd)
 	if (fd->fd > maxfd)
 		maxfd = fd->fd;
 
-#ifdef BSC_FD_CHECK
+#ifdef OSMO_FD_CHECK
 	if (osmo_fd_is_registered(fd)) {
 		fprintf(stderr, "Adding a osmo_fd that is already in the list.\n");
 		return 0;
@@ -170,13 +175,13 @@ inline int osmo_fd_fill_fds(void *_rset, void *_wset, void *_eset)
 	int highfd = 0;
 
 	llist_for_each_entry(ufd, &osmo_fds, list) {
-		if (ufd->when & BSC_FD_READ)
+		if (ufd->when & OSMO_FD_READ)
 			FD_SET(ufd->fd, readset);
 
-		if (ufd->when & BSC_FD_WRITE)
+		if (ufd->when & OSMO_FD_WRITE)
 			FD_SET(ufd->fd, writeset);
 
-		if (ufd->when & BSC_FD_EXCEPT)
+		if (ufd->when & OSMO_FD_EXCEPT)
 			FD_SET(ufd->fd, exceptset);
 
 		if (ufd->fd > highfd)
@@ -198,22 +203,27 @@ restart:
 		int flags = 0;
 
 		if (FD_ISSET(ufd->fd, readset)) {
-			flags |= BSC_FD_READ;
+			flags |= OSMO_FD_READ;
 			FD_CLR(ufd->fd, readset);
 		}
 
 		if (FD_ISSET(ufd->fd, writeset)) {
-			flags |= BSC_FD_WRITE;
+			flags |= OSMO_FD_WRITE;
 			FD_CLR(ufd->fd, writeset);
 		}
 
 		if (FD_ISSET(ufd->fd, exceptset)) {
-			flags |= BSC_FD_EXCEPT;
+			flags |= OSMO_FD_EXCEPT;
 			FD_CLR(ufd->fd, exceptset);
 		}
 
 		if (flags) {
 			work = 1;
+			/* make sure to clear any log context before processing the next incoming message
+			 * as part of some file descriptor callback.  This effectively prevents "context
+			 * leaking" from processing of one message into processing of the next message as part
+			 * of one iteration through the list of file descriptors here.  See OS#3813 */
+			log_reset_context();
 			ufd->cb(ufd, flags);
 		}
 		/* ugly, ugly hack. If more than one filedescriptor was
@@ -227,11 +237,7 @@ restart:
 	return work;
 }
 
-/*! select main loop integration
- *  \param[in] polling should we pollonly (1) or block on select (0)
- *  \returns 0 if no fd handled; 1 if fd handled; negative in case of error
- */
-int osmo_select_main(int polling)
+static int _osmo_select_main(int polling)
 {
 	fd_set readset, writeset, exceptset;
 	int rc;
@@ -253,10 +259,42 @@ int osmo_select_main(int polling)
 	/* fire timers */
 	osmo_timers_update();
 
+	OSMO_ASSERT(osmo_ctx->select);
+
 	/* call registered callback functions */
 	return osmo_fd_disp_fds(&readset, &writeset, &exceptset);
 }
 
+/*! select main loop integration
+ *  \param[in] polling should we pollonly (1) or block on select (0)
+ *  \returns 0 if no fd handled; 1 if fd handled; negative in case of error
+ */
+int osmo_select_main(int polling)
+{
+	int rc = _osmo_select_main(polling);
+#ifndef EMBEDDED
+	if (talloc_total_size(osmo_ctx->select) != 0) {
+		osmo_panic("You cannot use the 'select' volatile "
+			   "context if you don't use osmo_select_main_ctx()!\n");
+	}
+#endif
+	return rc;
+}
+
+#ifndef EMBEDDED
+/*! select main loop integration with temporary select-dispatch talloc context
+ *  \param[in] polling should we pollonly (1) or block on select (0)
+ *  \returns 0 if no fd handled; 1 if fd handled; negative in case of error
+ */
+int osmo_select_main_ctx(int polling)
+{
+	int rc = _osmo_select_main(polling);
+	/* free all the children of the volatile 'select' scope context */
+	talloc_free_children(osmo_ctx->select);
+	return rc;
+}
+#endif
+
 /*! find an osmo_fd based on the integer fd
  *  \param[in] fd file descriptor to use as search key
  *  \returns \ref osmo_fd for \ref fd; NULL in case it doesn't exist */
@@ -271,6 +309,18 @@ struct osmo_fd *osmo_fd_get_by_fd(int fd)
 	return NULL;
 }
 
+/*! initialize the osmocom select abstraction for the current thread */
+void osmo_select_init(void)
+{
+	INIT_LLIST_HEAD(&osmo_fds);
+}
+
+/* ensure main thread always has pre-initialized osmo_fds */
+static __attribute__((constructor)) void on_dso_load_select(void)
+{
+	osmo_select_init();
+}
+
 #ifdef HAVE_SYS_TIMERFD_H
 #include <sys/timerfd.h>
 
@@ -284,7 +334,7 @@ int osmo_timerfd_disable(struct osmo_fd *ofd)
 	return timerfd_settime(ofd->fd, 0, &its_null, NULL);
 }
 
-/*! schedule the osmcoom-wrapped timerfd to occur first at \a first, then periodically at \a interval
+/*! schedule the osmocom-wrapped timerfd to occur first at \a first, then periodically at \a interval
  *  \param[in] ofd Osmocom wrapped timerfd
  *  \param[in] first Relative time at which the timer should first execute (NULL = \a interval)
  *  \param[in] interval Time interval at which subsequent timer shall fire
@@ -321,20 +371,87 @@ int osmo_timerfd_setup(struct osmo_fd *ofd, int (*cb)(struct osmo_fd *, unsigned
 {
 	ofd->cb = cb;
 	ofd->data = data;
-	ofd->when = BSC_FD_READ;
+	ofd->when = OSMO_FD_READ;
 
 	if (ofd->fd < 0) {
+		int rc;
+
 		ofd->fd = timerfd_create(CLOCK_MONOTONIC, TFD_NONBLOCK);
 		if (ofd->fd < 0)
 			return ofd->fd;
 
-		osmo_fd_register(ofd);
+		rc = osmo_fd_register(ofd);
+		if (rc < 0) {
+			osmo_fd_unregister(ofd);
+			close(ofd->fd);
+			ofd->fd = -1;
+			return rc;
+		}
 	}
 	return 0;
 }
 
 #endif /* HAVE_SYS_TIMERFD_H */
 
+#ifdef HAVE_SYS_SIGNALFD_H
+#include <sys/signalfd.h>
+
+static int signalfd_callback(struct osmo_fd *ofd, unsigned int what)
+{
+	struct osmo_signalfd *osfd = ofd->data;
+	struct signalfd_siginfo fdsi;
+	int rc;
+
+	rc = read(ofd->fd, &fdsi, sizeof(fdsi));
+	if (rc < 0) {
+		osmo_fd_unregister(ofd);
+		close(ofd->fd);
+		ofd->fd = -1;
+		return rc;
+	}
+
+	osfd->cb(osfd, &fdsi);
+
+	return 0;
+};
+
+/*! create a signalfd and register it with osmocom select loop.
+ *  \param[in] ctx talloc context from which osmo_signalfd is to be allocated
+ *  \param[in] set of signals to be accept via this file descriptor
+ *  \param[in] cb call-back function to be called for each arriving signal
+ *  \param[in] data opaque user-provided data to pass to callback
+ *  \returns pointer to newly-allocated + registered osmo_signalfd; NULL on error */
+struct osmo_signalfd *
+osmo_signalfd_setup(void *ctx, sigset_t set, osmo_signalfd_cb *cb, void *data)
+{
+	struct osmo_signalfd *osfd = talloc_size(ctx, sizeof(*osfd));
+	int fd, rc;
+
+	if (!osfd)
+		return NULL;
+
+	osfd->data = data;
+	osfd->sigset = set;
+	osfd->cb = cb;
+
+	fd = signalfd(-1, &osfd->sigset, SFD_NONBLOCK);
+	if (fd < 0) {
+		talloc_free(osfd);
+		return NULL;
+	}
+
+	osmo_fd_setup(&osfd->ofd, fd, OSMO_FD_READ, signalfd_callback, osfd, 0);
+	rc = osmo_fd_register(&osfd->ofd);
+	if (rc < 0) {
+		close(fd);
+		talloc_free(osfd);
+		return NULL;
+	}
+
+	return osfd;
+}
+
+#endif /* HAVE_SYS_SIGNALFD_H */
 
 /*! @} */
 
