@@ -74,6 +74,7 @@
 #include <errno.h>
 #include <stdio.h>
 #include <sys/types.h>
+#include <inttypes.h>
 
 #ifdef HAVE_SYS_SOCKET_H
 #include <sys/socket.h>
@@ -85,7 +86,7 @@
 #include <osmocom/core/logging.h>
 #include <osmocom/core/rate_ctr.h>
 #include <osmocom/core/stat_item.h>
-#include <osmocom/core/timer.h>
+#include <osmocom/core/select.h>
 #include <osmocom/core/counter.h>
 #include <osmocom/core/msgb.h>
 
@@ -102,7 +103,7 @@ static struct osmo_stats_config s_stats_config = {
 };
 struct osmo_stats_config *osmo_stats_config = &s_stats_config;
 
-static struct osmo_timer_list osmo_stats_timer;
+static struct osmo_fd osmo_stats_timer = { .fd = -1 };
 
 static int osmo_stats_reporter_log_send_counter(struct osmo_stats_reporter *srep,
 	const struct rate_ctr_group *ctrg,
@@ -140,23 +141,61 @@ static int update_srep_config(struct osmo_stats_reporter *srep)
 	return rc;
 }
 
-static void osmo_stats_timer_cb(void *data)
+static int osmo_stats_timer_cb(struct osmo_fd *ofd, unsigned int what)
 {
-	int interval = osmo_stats_config->interval;
+	uint64_t expire_count;
+	int rc;
+
+	/* check that the timer has actually expired */
+	if (!(what & OSMO_FD_READ))
+		return 0;
+
+	/* read from timerfd: number of expirations of periodic timer */
+	rc = read(ofd->fd, (void *) &expire_count, sizeof(expire_count));
+	if (rc < 0 && errno == EAGAIN)
+		return 0;
+	OSMO_ASSERT(rc == sizeof(expire_count));
+
+	if (expire_count > 1)
+		LOGP(DLSTATS, LOGL_NOTICE, "Stats timer expire_count=%" PRIu64 ": We missed %" PRIu64 " timers\n",
+			expire_count, expire_count-1);
 
 	if (!llist_empty(&osmo_stats_reporter_list))
 		osmo_stats_report();
 
-	osmo_timer_schedule(&osmo_stats_timer, interval, 0);
+	return 0;
 }
 
 static int start_timer()
 {
+	int rc;
+	int interval = osmo_stats_config->interval;
+
 	if (!is_initialised)
 		return -ESRCH;
 
-	osmo_timer_setup(&osmo_stats_timer, osmo_stats_timer_cb, NULL);
-	osmo_timer_schedule(&osmo_stats_timer, 0, 1);
+	struct timespec ts_first = {.tv_sec=0, .tv_nsec=1000};
+	struct timespec ts_interval = {.tv_sec=interval, .tv_nsec=0};
+
+	rc = osmo_timerfd_setup(&osmo_stats_timer, osmo_stats_timer_cb, NULL);
+	if (rc < 0)
+		LOGP(DLSTATS, LOGL_ERROR, "Failed to setup the timer with error code %d (fd=%d)\n",
+		     rc, osmo_stats_timer.fd);
+
+        if (interval == 0) {
+		rc = osmo_timerfd_disable(&osmo_stats_timer);
+		if (rc < 0)
+			LOGP(DLSTATS, LOGL_ERROR, "Failed to disable the timer with error code %d (fd=%d)\n",
+			     rc, osmo_stats_timer.fd);
+        } else {
+
+		rc = osmo_timerfd_schedule(&osmo_stats_timer, &ts_first, &ts_interval);
+		if (rc < 0)
+			LOGP(DLSTATS, LOGL_ERROR, "Failed to schedule the timer with error code %d (fd=%d, interval %d sec)\n",
+			rc, osmo_stats_timer.fd, interval);
+
+		LOGP(DLSTATS, LOGL_INFO, "Stats timer started with interval %d sec\n", interval);
+	}
 
 	return 0;
 }
@@ -326,13 +365,12 @@ int osmo_stats_reporter_set_max_class(struct osmo_stats_reporter *srep,
 	return 0;
 }
 
-/*! Set the reporting interval of a given stats_reporter (in seconds).
- *  \param[in] srep stats_reporter whose remote address is to be set
+/*! Set the reporting interval (common for all reporters)
  *  \param[in] interval Reporting interval in seconds
  *  \returns 0 on success; negative on error */
 int osmo_stats_set_interval(int interval)
 {
-	if (interval <= 0)
+	if (interval < 0)
 		return -EINVAL;
 
 	osmo_stats_config->interval = interval;
@@ -342,6 +380,25 @@ int osmo_stats_set_interval(int interval)
 	return 0;
 }
 
+/*! Set the regular flush period for a given stats_reporter
+ *
+ * Send all stats even if they have not changed (i.e. force the flush)
+ * every N-th reporting interval. Set to 0 to disable regular flush,
+ * set to 1 to flush every time, set to 2 to flush every 2nd time, etc.
+ *  \param[in] srep stats_reporter to set flush period for
+ *  \param[in] period Reporting interval in seconds
+ *  \returns 0 on success; negative on error */
+int osmo_stats_reporter_set_flush_period(struct osmo_stats_reporter *srep, unsigned int period)
+{
+	srep->flush_period = period;
+	srep->flush_period_counter = 0;
+	/* force the flush now if it's not disabled by period=0 */
+	if (period > 0)
+		srep->force_single_flush = 1;
+
+	return 0;
+}
+
 /*! Set the name prefix of a given stats_reporter.
  *  \param[in] srep stats_reporter whose name prefix is to be set
  *  \param[in] prefix NAme perfix to pre-pend for any reported value
@@ -707,7 +764,17 @@ static void flush_all_reporters()
 			continue;
 
 		osmo_stats_reporter_send_buffer(srep);
+
+		/* reset force_single_flush first */
 		srep->force_single_flush = 0;
+		/* and schedule a new flush if it's time for it */
+		if (srep->flush_period > 0) {
+			srep->flush_period_counter++;
+			if (srep->flush_period_counter >= srep->flush_period) {
+				srep->force_single_flush = 1;
+				srep->flush_period_counter = 0;
+			}
+		}
 	}
 }
 
