@@ -99,24 +99,16 @@ static int osmo_stats_reporter_statsd_send(struct osmo_stats_reporter *srep,
 	int old_len = msgb_length(srep->buffer);
 
 	if (prefix) {
-		if (name1) {
-			if (index1 != 0)
-				fmt = "%1$s.%2$s.%6$u.%3$s:%4$d|%5$s";
-			else
-				fmt = "%1$s.%2$s.%3$s:%4$d|%5$s";
-		} else {
+		if (name1)
+			fmt = "%1$s.%2$s.%6$u.%3$s:%4$d|%5$s";
+		else
 			fmt = "%1$s.%2$0.0s%3$s:%4$d|%5$s";
-		}
 	} else {
 		prefix = "";
-		if (name1) {
-			if (index1 != 0)
-				fmt = "%1$s%2$s.%6$u.%3$s:%4$d|%5$s";
-			else
-				fmt = "%1$s%2$s.%3$s:%4$d|%5$s";
-		} else {
+		if (name1)
+			fmt = "%1$s%2$s.%6$u.%3$s:%4$d|%5$s";
+		else
 			fmt = "%1$s%2$0.0s%3$s:%4$d|%5$s";
-		}
 	}
 
 	if (srep->agg_enabled) {
@@ -184,20 +176,17 @@ static int osmo_stats_reporter_statsd_send_item(struct osmo_stats_reporter *srep
 	const struct osmo_stat_item_group *statg,
 	const struct osmo_stat_item_desc *desc, int64_t value)
 {
-	const char *unit = desc->unit;
-
-	if (unit == OSMO_STAT_ITEM_NO_UNIT) {
-		unit = "g";
-		if (value < 0)
-			osmo_stats_reporter_statsd_send(srep,
+	if (value < 0) {
+		return osmo_stats_reporter_statsd_send(srep,
 				statg->desc->group_name_prefix,
 				statg->idx,
-				desc->name, 0, unit);
+				desc->name, 0, "g");
+	} else {
+		return osmo_stats_reporter_statsd_send(srep,
+			statg->desc->group_name_prefix,
+			statg->idx,
+			desc->name, value, "g");
 	}
-	return osmo_stats_reporter_statsd_send(srep,
-		statg->desc->group_name_prefix,
-		statg->idx,
-		desc->name, value, unit);
 }
 #endif /* !EMBEDDED */
 
