@@ -40,10 +40,10 @@
 #include <osmocom/core/linuxlist.h>
 
 /* These store the amount of time that we wait until next timer expires. */
-static struct timeval nearest;
-static struct timeval *nearest_p;
+static __thread struct timeval nearest;
+static __thread struct timeval *nearest_p;
 
-static struct rb_root timer_root = RB_ROOT;
+static __thread struct rb_root timer_root = RB_ROOT;
 
 static void __add_timer(struct osmo_timer_list *timer)
 {
@@ -68,8 +68,8 @@ static void __add_timer(struct osmo_timer_list *timer)
 
 /*! set up timer callback and data
  *  \param[in] timer the timer that should be added
- *  \param[in] callback function to be called when timer expires
- *  \param[in] pointer to data that passed to the callback function
+ *  \param[in] cb function to be called when timer expires
+ *  \param[in] data pointer to data that passed to the callback function
  */
 void osmo_timer_setup(struct osmo_timer_list *timer, void (*cb)(void *data),
 		      void *data)
