@@ -37,6 +37,7 @@ Boston, MA  02110-1301, USA. */
 #include <unistd.h>
 #include <ctype.h>
 #include <time.h>
+#include <limits.h>
 #include <sys/time.h>
 #include <sys/stat.h>
 
@@ -47,6 +48,11 @@ Boston, MA  02110-1301, USA. */
 #include <osmocom/core/talloc.h>
 #include <osmocom/core/utils.h>
 
+#ifndef MAXPATHLEN
+  #define MAXPATHLEN 4096
+#endif
+
+
 /*! \addtogroup command
  *  @{
  *  VTY command handling
@@ -195,18 +201,6 @@ static int cmp_desc(const void *p, const void *q)
 	return strcmp(a->cmd, b->cmd);
 }
 
-static int is_config_child(struct vty *vty)
-{
-	if (vty->node <= CONFIG_NODE)
-		return 0;
-	else if (vty->node > CONFIG_NODE && vty->node < _LAST_OSMOVTY_NODE)
-		return 1;
-	else if (host.app_info->is_config_node)
-		return host.app_info->is_config_node(vty, vty->node);
-	else
-		return vty->node > CONFIG_NODE;
-}
-
 /*! Sort each node's command element according to command string. */
 void sort_node(void)
 {
@@ -380,6 +374,7 @@ static char *cmd_desc_str(const char **string)
 static vector cmd_make_descvec(const char *string, const char *descstr)
 {
 	int multiple = 0;
+	int optional_brace = 0;
 	const char *sp;
 	char *token;
 	int len;
@@ -401,6 +396,12 @@ static vector cmd_make_descvec(const char *string, const char *descstr)
 		while (isspace((int)*cp) && *cp != '\0')
 			cp++;
 
+		/* Explicitly detect optional multi-choice braces like [(one|two)]. */
+		if (cp[0] == '[' && cp[1] == '(') {
+			optional_brace = 1;
+			cp++;
+		}
+
 		if (*cp == '(') {
 			multiple = 1;
 			cp++;
@@ -408,6 +409,9 @@ static vector cmd_make_descvec(const char *string, const char *descstr)
 		if (*cp == ')') {
 			multiple = 0;
 			cp++;
+			if (*cp == ']')
+				cp++;
+			optional_brace = 0;
 		}
 		if (*cp == '|') {
 			OSMO_ASSERT(multiple);
@@ -434,9 +438,17 @@ static vector cmd_make_descvec(const char *string, const char *descstr)
 
 		len = cp - sp;
 
-		token = _talloc_zero(tall_vty_cmd_ctx, len + 1, "cmd_make_descvec");
-		memcpy(token, sp, len);
-		*(token + len) = '\0';
+		token = _talloc_zero(tall_vty_cmd_ctx, len + (optional_brace? 2 : 0) + 1, "cmd_make_descvec");
+		if (optional_brace) {
+			/* Place each individual multi-choice token in its own square braces */
+			token[0] = '[';
+			memcpy(token + 1, sp, len);
+			token[1 + len] = ']';
+			token[2 + len] = '\0';
+		} else {
+			memcpy(token, sp, len);
+			*(token + len) = '\0';
+		}
 
 		desc = talloc_zero(tall_vty_cmd_ctx, struct desc);
 		desc->cmd = token;
@@ -467,7 +479,7 @@ static int cmd_cmdsize(vector strvec)
 
 	for (i = 0; i < vector_active(strvec); i++)
 		if ((descvec = vector_slot(strvec, i)) != NULL) {
-			if ((vector_active(descvec)) == 1
+			if ((vector_active(descvec)) >= 1
 			    && (desc = vector_slot(descvec, 0)) != NULL) {
 				if (desc->cmd == NULL || CMD_OPTION(desc->cmd))
 					return size;
@@ -497,7 +509,7 @@ char *osmo_asciidoc_escape(const char *inp)
 {
 	int _strlen;
 	char *out, *out_ptr;
-	int len = 0, i, j;
+	int len = 0, i;
 
 	if (!inp)
 		return NULL;
@@ -520,23 +532,18 @@ char *osmo_asciidoc_escape(const char *inp)
 
 	out_ptr = out;
 
-#define ADD(out, str) \
-	for (j = 0; j < strlen(str); ++j) \
-		*(out++) = str[j];
-
 	for (i = 0; i < _strlen; ++i) {
 		switch (inp[i]) {
 		case '|':
-			ADD(out_ptr, "\\|");
-			break;
+		/* Prepend escape character "\": */
+			*(out_ptr++) = '\\';
+			/* fall through */
 		default:
 			*(out_ptr++) = inp[i];
 			break;
 		}
 	}
 
-#undef ADD
-
 	out_ptr[0] = '\0';
 	return out;
 }
@@ -613,15 +620,17 @@ static char *xml_escape(const char *inp)
 	return out;
 }
 
+typedef int (*print_func_t)(void *data, const char *fmt, ...);
+
 /*
- * Write one cmd_element as XML to the given VTY.
+ * Write one cmd_element as XML via a print_func_t.
  */
-static int vty_dump_element(struct cmd_element *cmd, struct vty *vty)
+static int vty_dump_element(struct cmd_element *cmd, print_func_t print_func, void *data, const char *newline)
 {
 	char *xml_string = xml_escape(cmd->string);
 
-	vty_out(vty, "    <command id='%s'>%s", xml_string, VTY_NEWLINE);
-	vty_out(vty, "      <params>%s", VTY_NEWLINE);
+	print_func(data, "    <command id='%s'>%s", xml_string, newline);
+	print_func(data, "      <params>%s", newline);
 
 	int j;
 	for (j = 0; j < vector_count(cmd->strvec); ++j) {
@@ -635,15 +644,15 @@ static int vty_dump_element(struct cmd_element *cmd, struct vty *vty)
 
 			xml_param = xml_escape(desc->cmd);
 			xml_doc = xml_escape(desc->str);
-			vty_out(vty, "        <param name='%s' doc='%s' />%s",
-				xml_param, xml_doc, VTY_NEWLINE);
+			print_func(data, "        <param name='%s' doc='%s' />%s",
+				xml_param, xml_doc, newline);
 			talloc_free(xml_param);
 			talloc_free(xml_doc);
 		}
 	}
 
-	vty_out(vty, "      </params>%s", VTY_NEWLINE);
-	vty_out(vty, "    </command>%s", VTY_NEWLINE);
+	print_func(data, "      </params>%s", newline);
+	print_func(data, "    </command>%s", newline);
 
 	talloc_free(xml_string);
 	return 0;
@@ -652,20 +661,20 @@ static int vty_dump_element(struct cmd_element *cmd, struct vty *vty)
 static bool vty_command_is_common(struct cmd_element *cmd);
 
 /*
- * Dump all nodes and commands associated with a given node as XML to the VTY.
+ * Dump all nodes and commands associated with a given node as XML via a print_func_t.
  */
-static int vty_dump_nodes(struct vty *vty)
+static int vty_dump_nodes(print_func_t print_func, void *data, const char *newline)
 {
 	int i, j;
 	int same_name_count;
 
-	vty_out(vty, "<vtydoc xmlns='urn:osmocom:xml:libosmocore:vty:doc:1.0'>%s", VTY_NEWLINE);
+	print_func(data, "<vtydoc xmlns='urn:osmocom:xml:libosmocore:vty:doc:1.0'>%s", newline);
 
 	/* Only once, list all common node commands. Use the CONFIG node to find common node commands. */
-	vty_out(vty, "  <node id='_common_cmds_'>%s", VTY_NEWLINE);
-	vty_out(vty, "    <name>Common Commands</name>%s", VTY_NEWLINE);
-	vty_out(vty, "    <description>These commands are available on all VTY nodes. They are listed"
-		" here only once, to unclutter the VTY reference.</description>%s", VTY_NEWLINE);
+	print_func(data, "  <node id='_common_cmds_'>%s", newline);
+	print_func(data, "    <name>Common Commands</name>%s", newline);
+	print_func(data, "    <description>These commands are available on all VTY nodes. They are listed"
+		" here only once, to unclutter the VTY reference.</description>%s", newline);
 	for (i = 0; i < vector_active(cmdvec); ++i) {
 		struct cmd_node *cnode;
 		cnode = vector_slot(cmdvec, i);
@@ -679,11 +688,11 @@ static int vty_dump_nodes(struct vty *vty)
 			elem = vector_slot(cnode->cmd_vector, j);
 			if (!vty_command_is_common(elem))
 				continue;
-			if (!elem->attr & CMD_ATTR_DEPRECATED)
-				vty_dump_element(elem, vty);
+			if (!(elem->attr & (CMD_ATTR_DEPRECATED | CMD_ATTR_HIDDEN)))
+				vty_dump_element(elem, print_func, data, newline);
 		}
 	}
-	vty_out(vty, "  </node>%s", VTY_NEWLINE);
+	print_func(data, "  </node>%s", newline);
 
 	for (i = 0; i < vector_active(cmdvec); ++i) {
 		struct cmd_node *cnode;
@@ -706,29 +715,62 @@ static int vty_dump_nodes(struct vty *vty)
 				same_name_count ++;
 		}
 
-		vty_out(vty, "  <node id='%s", cnode->name);
+		print_func(data, "  <node id='%s", cnode->name);
 		if (same_name_count > 1 || !*cnode->name)
-			vty_out(vty, "_%d", same_name_count);
-		vty_out(vty, "'>%s", VTY_NEWLINE);
-		vty_out(vty, "    <name>%s</name>%s", cnode->name, VTY_NEWLINE);
+			print_func(data, "_%d", same_name_count);
+		print_func(data, "'>%s", newline);
+		print_func(data, "    <name>%s</name>%s", cnode->name, newline);
 
 		for (j = 0; j < vector_active(cnode->cmd_vector); ++j) {
 			struct cmd_element *elem;
 			elem = vector_slot(cnode->cmd_vector, j);
 			if (vty_command_is_common(elem))
 				continue;
-			if (!elem->attr & CMD_ATTR_DEPRECATED)
-				vty_dump_element(elem, vty);
+			if (!(elem->attr & (CMD_ATTR_DEPRECATED | CMD_ATTR_HIDDEN)))
+				vty_dump_element(elem, print_func, data, newline);
 		}
 
-		vty_out(vty, "  </node>%s", VTY_NEWLINE);
+		print_func(data, "  </node>%s", newline);
 	}
 
-	vty_out(vty, "</vtydoc>%s", VTY_NEWLINE);
+	print_func(data, "</vtydoc>%s", newline);
 
 	return 0;
 }
 
+static int print_func_vty(void *data, const char *format, ...)
+{
+	struct vty *vty = data;
+	va_list args;
+	int rc;
+	va_start(args, format);
+	rc = vty_out_va(vty, format, args);
+	va_end(args);
+	return rc;
+}
+
+static int vty_dump_xml_ref_to_vty(struct vty *vty)
+{
+	return vty_dump_nodes(print_func_vty, vty, VTY_NEWLINE);
+}
+
+static int print_func_stream(void *data, const char *format, ...)
+{
+	va_list args;
+	int rc;
+	va_start(args, format);
+	rc = vfprintf((FILE*)data, format, args);
+	va_end(args);
+	return rc;
+}
+
+/*! Print the XML reference of all VTY nodes to the given stream.
+ */
+int vty_dump_xml_ref(FILE *stream)
+{
+	return vty_dump_nodes(print_func_stream, stream, "\n");
+}
+
 /* Check if a command with given string exists at given node */
 static int check_element_exists(struct cmd_node *cnode, const char *cmdstring)
 {
@@ -851,17 +893,17 @@ static vector cmd_node_vector(vector v, enum node_type ntype)
 
 /* Completion match types. */
 enum match_type {
-	no_match = 0,
-	any_match,
-	extend_match,
-	ipv4_prefix_match,
-	ipv4_match,
-	ipv6_prefix_match,
-	ipv6_match,
-	range_match,
-	vararg_match,
-	partly_match,
-	exact_match,
+	NO_MATCH = 0,
+	ANY_MATCH,
+	EXTEND_MATCH,
+	IPV4_PREFIX_MATCH,
+	IPV4_MATCH,
+	IPV6_PREFIX_MATCH,
+	IPV6_MATCH,
+	RANGE_MATCH,
+	VARARG_MATCH,
+	PARTLY_MATCH,
+	EXACT_MATCH,
 };
 
 static enum match_type cmd_ipv4_match(const char *str)
@@ -871,7 +913,7 @@ static enum match_type cmd_ipv4_match(const char *str)
 	char buf[4];
 
 	if (str == NULL)
-		return partly_match;
+		return PARTLY_MATCH;
 
 	for (;;) {
 		memset(buf, 0, sizeof(buf));
@@ -879,29 +921,29 @@ static enum match_type cmd_ipv4_match(const char *str)
 		while (*str != '\0') {
 			if (*str == '.') {
 				if (dots >= 3)
-					return no_match;
+					return NO_MATCH;
 
 				if (*(str + 1) == '.')
-					return no_match;
+					return NO_MATCH;
 
 				if (*(str + 1) == '\0')
-					return partly_match;
+					return PARTLY_MATCH;
 
 				dots++;
 				break;
 			}
 			if (!isdigit((int)*str))
-				return no_match;
+				return NO_MATCH;
 
 			str++;
 		}
 
 		if (str - sp > 3)
-			return no_match;
+			return NO_MATCH;
 
 		strncpy(buf, sp, str - sp);
 		if (atoi(buf) > 255)
-			return no_match;
+			return NO_MATCH;
 
 		nums++;
 
@@ -912,9 +954,9 @@ static enum match_type cmd_ipv4_match(const char *str)
 	}
 
 	if (nums < 4)
-		return partly_match;
+		return PARTLY_MATCH;
 
-	return exact_match;
+	return EXACT_MATCH;
 }
 
 static enum match_type cmd_ipv4_prefix_match(const char *str)
@@ -924,7 +966,7 @@ static enum match_type cmd_ipv4_prefix_match(const char *str)
 	char buf[4];
 
 	if (str == NULL)
-		return partly_match;
+		return PARTLY_MATCH;
 
 	for (;;) {
 		memset(buf, 0, sizeof(buf));
@@ -932,44 +974,44 @@ static enum match_type cmd_ipv4_prefix_match(const char *str)
 		while (*str != '\0' && *str != '/') {
 			if (*str == '.') {
 				if (dots == 3)
-					return no_match;
+					return NO_MATCH;
 
 				if (*(str + 1) == '.' || *(str + 1) == '/')
-					return no_match;
+					return NO_MATCH;
 
 				if (*(str + 1) == '\0')
-					return partly_match;
+					return PARTLY_MATCH;
 
 				dots++;
 				break;
 			}
 
 			if (!isdigit((int)*str))
-				return no_match;
+				return NO_MATCH;
 
 			str++;
 		}
 
 		if (str - sp > 3)
-			return no_match;
+			return NO_MATCH;
 
 		strncpy(buf, sp, str - sp);
 		if (atoi(buf) > 255)
-			return no_match;
+			return NO_MATCH;
 
 		if (dots == 3) {
 			if (*str == '/') {
 				if (*(str + 1) == '\0')
-					return partly_match;
+					return PARTLY_MATCH;
 
 				str++;
 				break;
 			} else if (*str == '\0')
-				return partly_match;
+				return PARTLY_MATCH;
 		}
 
 		if (*str == '\0')
-			return partly_match;
+			return PARTLY_MATCH;
 
 		str++;
 	}
@@ -977,15 +1019,15 @@ static enum match_type cmd_ipv4_prefix_match(const char *str)
 	sp = str;
 	while (*str != '\0') {
 		if (!isdigit((int)*str))
-			return no_match;
+			return NO_MATCH;
 
 		str++;
 	}
 
 	if (atoi(sp) > 32)
-		return no_match;
+		return NO_MATCH;
 
-	return exact_match;
+	return EXACT_MATCH;
 }
 
 #define IPV6_ADDR_STR		"0123456789abcdefABCDEF:.%"
@@ -1009,10 +1051,10 @@ static enum match_type cmd_ipv6_match(const char *str)
 	int ret;
 
 	if (str == NULL)
-		return partly_match;
+		return PARTLY_MATCH;
 
 	if (strspn(str, IPV6_ADDR_STR) != strlen(str))
-		return no_match;
+		return NO_MATCH;
 
 	/* use inet_pton that has a better support,
 	 * for example inet_pton can support the automatic addresses:
@@ -1021,14 +1063,14 @@ static enum match_type cmd_ipv6_match(const char *str)
 	ret = inet_pton(AF_INET6, str, &sin6_dummy.sin6_addr);
 
 	if (ret == 1)
-		return exact_match;
+		return EXACT_MATCH;
 
 	while (*str != '\0') {
 		switch (state) {
 		case STATE_START:
 			if (*str == ':') {
 				if (*(str + 1) != ':' && *(str + 1) != '\0')
-					return no_match;
+					return NO_MATCH;
 				colons--;
 				state = STATE_COLON;
 			} else {
@@ -1048,10 +1090,10 @@ static enum match_type cmd_ipv6_match(const char *str)
 			break;
 		case STATE_DOUBLE:
 			if (double_colon)
-				return no_match;
+				return NO_MATCH;
 
 			if (*(str + 1) == ':')
-				return no_match;
+				return NO_MATCH;
 			else {
 				if (*(str + 1) != '\0')
 					colons++;
@@ -1065,7 +1107,7 @@ static enum match_type cmd_ipv6_match(const char *str)
 		case STATE_ADDR:
 			if (*(str + 1) == ':' || *(str + 1) == '\0') {
 				if (str - sp > 3)
-					return no_match;
+					return NO_MATCH;
 
 				nums++;
 				state = STATE_COLON;
@@ -1081,20 +1123,20 @@ static enum match_type cmd_ipv6_match(const char *str)
 		}
 
 		if (nums > 8)
-			return no_match;
+			return NO_MATCH;
 
 		if (colons > 7)
-			return no_match;
+			return NO_MATCH;
 
 		str++;
 	}
 
 #if 0
 	if (nums < 11)
-		return partly_match;
+		return PARTLY_MATCH;
 #endif				/* 0 */
 
-	return exact_match;
+	return EXACT_MATCH;
 }
 
 static enum match_type cmd_ipv6_prefix_match(const char *str)
@@ -1106,17 +1148,17 @@ static enum match_type cmd_ipv6_prefix_match(const char *str)
 	char *endptr = NULL;
 
 	if (str == NULL)
-		return partly_match;
+		return PARTLY_MATCH;
 
 	if (strspn(str, IPV6_PREFIX_STR) != strlen(str))
-		return no_match;
+		return NO_MATCH;
 
 	while (*str != '\0' && state != STATE_MASK) {
 		switch (state) {
 		case STATE_START:
 			if (*str == ':') {
 				if (*(str + 1) != ':' && *(str + 1) != '\0')
-					return no_match;
+					return NO_MATCH;
 				colons--;
 				state = STATE_COLON;
 			} else {
@@ -1128,7 +1170,7 @@ static enum match_type cmd_ipv6_prefix_match(const char *str)
 		case STATE_COLON:
 			colons++;
 			if (*(str + 1) == '/')
-				return no_match;
+				return NO_MATCH;
 			else if (*(str + 1) == ':')
 				state = STATE_DOUBLE;
 			else {
@@ -1138,10 +1180,10 @@ static enum match_type cmd_ipv6_prefix_match(const char *str)
 			break;
 		case STATE_DOUBLE:
 			if (double_colon)
-				return no_match;
+				return NO_MATCH;
 
 			if (*(str + 1) == ':')
-				return no_match;
+				return NO_MATCH;
 			else {
 				if (*(str + 1) != '\0' && *(str + 1) != '/')
 					colons++;
@@ -1160,11 +1202,11 @@ static enum match_type cmd_ipv6_prefix_match(const char *str)
 			if (*(str + 1) == ':' || *(str + 1) == '.'
 			    || *(str + 1) == '\0' || *(str + 1) == '/') {
 				if (str - sp > 3)
-					return no_match;
+					return NO_MATCH;
 
 				for (; sp <= str; sp++)
 					if (*sp == '/')
-						return no_match;
+						return NO_MATCH;
 
 				nums++;
 
@@ -1181,7 +1223,7 @@ static enum match_type cmd_ipv6_prefix_match(const char *str)
 			break;
 		case STATE_SLASH:
 			if (*(str + 1) == '\0')
-				return partly_match;
+				return PARTLY_MATCH;
 
 			state = STATE_MASK;
 			break;
@@ -1190,43 +1232,58 @@ static enum match_type cmd_ipv6_prefix_match(const char *str)
 		}
 
 		if (nums > 11)
-			return no_match;
+			return NO_MATCH;
 
 		if (colons > 7)
-			return no_match;
+			return NO_MATCH;
 
 		str++;
 	}
 
 	if (state < STATE_MASK)
-		return partly_match;
+		return PARTLY_MATCH;
 
 	mask = strtol(str, &endptr, 10);
 	if (*endptr != '\0')
-		return no_match;
+		return NO_MATCH;
 
 	if (mask < 0 || mask > 128)
-		return no_match;
+		return NO_MATCH;
 
 /* I don't know why mask < 13 makes command match partly.
    Forgive me to make this comments. I Want to set static default route
    because of lack of function to originate default in ospf6d; sorry
        yasu
   if (mask < 13)
-    return partly_match;
+    return PARTLY_MATCH;
 */
 
-	return exact_match;
+	return EXACT_MATCH;
 }
 
 #endif				/* HAVE_IPV6  */
 
-#define DECIMAL_STRLEN_MAX 10
+
+#if ULONG_MAX == 18446744073709551615UL
+#define DECIMAL_STRLEN_MAX_UNSIGNED 20
+#elif ULONG_MAX == 4294967295UL
+#define DECIMAL_STRLEN_MAX_UNSIGNED 10
+#else
+#error "ULONG_MAX not defined!"
+#endif
+
+#if LONG_MAX == 9223372036854775807L
+#define DECIMAL_STRLEN_MAX_SIGNED 19
+#elif LONG_MAX == 2147483647L
+#define DECIMAL_STRLEN_MAX_SIGNED 10
+#else
+#error "LONG_MAX not defined!"
+#endif
 
 static int cmd_range_match(const char *range, const char *str)
 {
 	char *p;
-	char buf[DECIMAL_STRLEN_MAX + 1];
+	char buf[DECIMAL_STRLEN_MAX_UNSIGNED + 1];
 	char *endptr = NULL;
 
 	if (str == NULL)
@@ -1243,7 +1300,7 @@ static int cmd_range_match(const char *range, const char *str)
 		p = strchr(range, '-');
 		if (p == NULL)
 			return 0;
-		if (p - range > DECIMAL_STRLEN_MAX)
+		if (p - range > DECIMAL_STRLEN_MAX_SIGNED)
 			return 0;
 		strncpy(buf, range, p - range);
 		buf[p - range] = '\0';
@@ -1255,7 +1312,7 @@ static int cmd_range_match(const char *range, const char *str)
 		p = strchr(range, '>');
 		if (p == NULL)
 			return 0;
-		if (p - range > DECIMAL_STRLEN_MAX)
+		if (p - range > DECIMAL_STRLEN_MAX_SIGNED)
 			return 0;
 		strncpy(buf, range, p - range);
 		buf[p - range] = '\0';
@@ -1268,6 +1325,9 @@ static int cmd_range_match(const char *range, const char *str)
 	} else {
 		unsigned long min, max, val;
 
+		if (str[0] == '-')
+			return 0;
+
 		val = strtoul(str, &endptr, 10);
 		if (*endptr != '\0')
 			return 0;
@@ -1276,7 +1336,7 @@ static int cmd_range_match(const char *range, const char *str)
 		p = strchr(range, '-');
 		if (p == NULL)
 			return 0;
-		if (p - range > DECIMAL_STRLEN_MAX)
+		if (p - range > DECIMAL_STRLEN_MAX_UNSIGNED)
 			return 0;
 		strncpy(buf, range, p - range);
 		buf[p - range] = '\0';
@@ -1288,7 +1348,7 @@ static int cmd_range_match(const char *range, const char *str)
 		p = strchr(range, '>');
 		if (p == NULL)
 			return 0;
-		if (p - range > DECIMAL_STRLEN_MAX)
+		if (p - range > DECIMAL_STRLEN_MAX_UNSIGNED)
 			return 0;
 		strncpy(buf, range, p - range);
 		buf[p - range] = '\0';
@@ -1331,7 +1391,7 @@ cmd_match(const char *str, const char *command,
 		 * as it we only discover it if a user tries to run it
 		 */
 		if (tmp == NULL)
-			return no_match;
+			return NO_MATCH;
 
 		ret = cmd_match(tmp, command, min, false);
 
@@ -1340,45 +1400,45 @@ cmd_match(const char *str, const char *command,
 		return ret;
 	}
 	else if (CMD_VARARG(str))
-		return vararg_match;
+		return VARARG_MATCH;
 	else if (CMD_RANGE(str))
 	{
 		if (cmd_range_match(str, command))
-			return range_match;
+			return RANGE_MATCH;
 	}
 #ifdef HAVE_IPV6
 	else if (CMD_IPV6(str))
 	{
 		if (cmd_ipv6_match(command) >= min)
-			return ipv6_match;
+			return IPV6_MATCH;
 	}
 	else if (CMD_IPV6_PREFIX(str))
 	{
 		if (cmd_ipv6_prefix_match(command) >= min)
-			return ipv6_prefix_match;
+			return IPV6_PREFIX_MATCH;
 	}
 #endif /* HAVE_IPV6  */
 	else if (CMD_IPV4(str))
 	{
 		if (cmd_ipv4_match(command) >= min)
-			return ipv4_match;
+			return IPV4_MATCH;
 	}
 	else if (CMD_IPV4_PREFIX(str))
 	{
 		if (cmd_ipv4_prefix_match(command) >= min)
-			return ipv4_prefix_match;
+			return IPV4_PREFIX_MATCH;
 	}
 	else if (CMD_VARIABLE(str))
-		return extend_match;
+		return EXTEND_MATCH;
 	else if (strncmp(command, str, strlen(command)) == 0)
 	{
 		if (strcmp(command, str) == 0)
-			return  exact_match;
-		else if (partly_match >= min)
-			return partly_match;
+			return  EXACT_MATCH;
+		else if (PARTLY_MATCH >= min)
+			return PARTLY_MATCH;
 	}
 
-	return no_match;
+	return NO_MATCH;
 }
 
 /* Filter vector at the specified index and by the given command string, to
@@ -1394,7 +1454,7 @@ cmd_filter(char *command, vector v, unsigned int index, enum match_type level)
 	vector descvec;
 	struct desc *desc;
 
-	match_type = no_match;
+	match_type = NO_MATCH;
 
 	/* If command and cmd_element string does not match set NULL to vector */
 	for (i = 0; i < vector_active(v); i++)
@@ -1414,7 +1474,7 @@ cmd_filter(char *command, vector v, unsigned int index, enum match_type level)
 
 						ret = cmd_match (desc->cmd, command, level, true);
 
-						if (ret != no_match)
+						if (ret != NO_MATCH)
 							matched++;
 
 						if (match_type < ret)
@@ -1425,8 +1485,8 @@ cmd_filter(char *command, vector v, unsigned int index, enum match_type level)
 			}
 		}
 
-	if (match_type == no_match)
-		return no_match;
+	if (match_type == NO_MATCH)
+		return NO_MATCH;
 
 	/* 2nd pass: We now know the 'strongest' match type for the index, so we
 	 * go again and filter out commands whose argument (at this index) is
@@ -1436,7 +1496,7 @@ cmd_filter(char *command, vector v, unsigned int index, enum match_type level)
 	 *   foo bar BLAH
 	 *
 	 * and the command string is 'foo bar 10', then we will get here with with
-	 * 'range_match' being the strongest match.  However, if 'BLAH' came
+	 * 'RANGE_MATCH' being the strongest match.  However, if 'BLAH' came
 	 * earlier, it won't have been filtered out (as a CMD_VARIABLE allows "10").
 	 *
 	 * If we don't do a 2nd pass and filter it out, the higher-layers will
@@ -1457,7 +1517,7 @@ cmd_filter(char *command, vector v, unsigned int index, enum match_type level)
 					if ((desc = vector_slot(descvec, j))) {
 						enum match_type ret;
 
-						ret = cmd_match(desc->cmd, command, any_match, true);
+						ret = cmd_match(desc->cmd, command, ANY_MATCH, true);
 
 						if (ret >= match_type)
 							matched++;
@@ -1492,106 +1552,112 @@ is_cmd_ambiguous(char *command, vector v, int index, enum match_type type)
 	 * or ambiguities to cause a noticeable memory footprint from keeping all allocations. */
 	void *cmd_deopt_ctx = NULL;
 
-	for (i = 0; i < vector_active(v); i++)
-		if ((cmd_element = vector_slot(v, i)) != NULL) {
-			int match = 0;
+	for (i = 0; i < vector_active(v); i++) {
+		cmd_element = vector_slot(v, i);
+		if (!cmd_element)
+			continue;
 
-			descvec = vector_slot(cmd_element->strvec, index);
+		int match = 0;
 
-			for (j = 0; j < vector_active(descvec); j++)
-				if ((desc = vector_slot(descvec, j))) {
-					enum match_type ret;
-					const char *str = desc->cmd;
+		descvec = vector_slot(cmd_element->strvec, index);
 
-					if (CMD_OPTION(str)) {
-						if (!cmd_deopt_ctx)
-							cmd_deopt_ctx =
-								talloc_named_const(tall_vty_cmd_ctx, 0,
-										   __func__);
-						str = cmd_deopt(cmd_deopt_ctx, str);
-						if (str == NULL)
-							continue;
-					}
+		for (j = 0; j < vector_active(descvec); j++) {
+			desc = vector_slot(descvec, j);
+			if (!desc)
+				continue;
 
-					switch (type) {
-					case exact_match:
-						if (!(CMD_VARIABLE (str))
-						   && strcmp(command, str) == 0)
-							match++;
-						break;
-					case partly_match:
-						if (!(CMD_VARIABLE (str))
-						   && strncmp(command, str, strlen (command)) == 0)
-						{
-							if (matched
-							    && strcmp(matched,
-								      str) != 0) {
-								ret = 1; /* There is ambiguous match. */
-								goto free_and_return;
-							} else
-								matched = str;
-							match++;
-						}
-						break;
-					case range_match:
-						if (cmd_range_match
-						    (str, command)) {
-							if (matched
-							    && strcmp(matched,
-								      str) != 0) {
-								ret = 1;
-								goto free_and_return;
-							} else
-								matched = str;
-							match++;
-						}
-						break;
-#ifdef HAVE_IPV6
-					case ipv6_match:
-						if (CMD_IPV6(str))
-							match++;
-						break;
-					case ipv6_prefix_match:
-						if ((ret =
-						     cmd_ipv6_prefix_match
-						     (command)) != no_match) {
-							if (ret == partly_match) {
-								ret = 2;	/* There is incomplete match. */
-								goto free_and_return;
-							}
+			enum match_type mtype;
+			const char *str = desc->cmd;
 
-							match++;
-						}
-						break;
-#endif				/* HAVE_IPV6 */
-					case ipv4_match:
-						if (CMD_IPV4(str))
-							match++;
-						break;
-					case ipv4_prefix_match:
-						if ((ret =
-						     cmd_ipv4_prefix_match
-						     (command)) != no_match) {
-							if (ret == partly_match) {
-								ret = 2;	/* There is incomplete match. */
-								goto free_and_return;
-							}
+			if (CMD_OPTION(str)) {
+				if (!cmd_deopt_ctx)
+					cmd_deopt_ctx =
+						talloc_named_const(tall_vty_cmd_ctx, 0,
+								   __func__);
+				str = cmd_deopt(cmd_deopt_ctx, str);
+				if (str == NULL)
+					continue;
+			}
 
-							match++;
-						}
-						break;
-					case extend_match:
-						if (CMD_VARIABLE (str))
-							match++;
-						break;
-					case no_match:
-					default:
-						break;
-					}
+			switch (type) {
+			case EXACT_MATCH:
+				if (!(CMD_VARIABLE (str))
+				   && strcmp(command, str) == 0)
+					match++;
+				break;
+			case PARTLY_MATCH:
+				if (!(CMD_VARIABLE (str))
+				   && strncmp(command, str, strlen (command)) == 0)
+				{
+					if (matched
+					    && strcmp(matched,
+						      str) != 0) {
+						ret = 1; /* There is ambiguous match. */
+						goto free_and_return;
+					} else
+						matched = str;
+					match++;
 				}
-			if (!match)
-				vector_slot(v, i) = NULL;
+				break;
+			case RANGE_MATCH:
+				if (cmd_range_match
+				    (str, command)) {
+					if (matched
+					    && strcmp(matched,
+						      str) != 0) {
+						ret = 1;
+						goto free_and_return;
+					} else
+						matched = str;
+					match++;
+				}
+				break;
+#ifdef HAVE_IPV6
+			case IPV6_MATCH:
+				if (CMD_IPV6(str))
+					match++;
+				break;
+			case IPV6_PREFIX_MATCH:
+				if ((mtype =
+				     cmd_ipv6_prefix_match
+				     (command)) != NO_MATCH) {
+					if (mtype == PARTLY_MATCH) {
+						ret = 2;	/* There is incomplete match. */
+						goto free_and_return;
+					}
+
+					match++;
+				}
+				break;
+#endif				/* HAVE_IPV6 */
+			case IPV4_MATCH:
+				if (CMD_IPV4(str))
+					match++;
+				break;
+			case IPV4_PREFIX_MATCH:
+				if ((mtype =
+				     cmd_ipv4_prefix_match
+				     (command)) != NO_MATCH) {
+					if (mtype == PARTLY_MATCH) {
+						ret = 2;	/* There is incomplete match. */
+						goto free_and_return;
+					}
+
+					match++;
+				}
+				break;
+			case EXTEND_MATCH:
+				if (CMD_VARIABLE (str))
+					match++;
+				break;
+			case NO_MATCH:
+			default:
+				break;
+			}
 		}
+		if (!match)
+			vector_slot(v, i) = NULL;
+	}
 
 free_and_return:
 	if (cmd_deopt_ctx)
@@ -1750,9 +1816,9 @@ cmd_describe_command_real(vector vline, struct vty *vty, int *status)
 		if (!command)
 			continue;
 
-		match = cmd_filter(command, cmd_vector, i, any_match);
+		match = cmd_filter(command, cmd_vector, i, ANY_MATCH);
 
-		if (match == vararg_match) {
+		if (match == VARARG_MATCH) {
 			struct cmd_element *cmd_element;
 			vector descvec;
 			unsigned int j, k;
@@ -1805,7 +1871,7 @@ cmd_describe_command_real(vector vline, struct vty *vty, int *status)
 	/* Make sure that cmd_vector is filtered based on current word */
 	command = vector_slot(vline, index);
 	if (command)
-		cmd_filter(command, cmd_vector, index, any_match);
+		cmd_filter(command, cmd_vector, index, ANY_MATCH);
 
 	/* Make description vector. */
 	for (i = 0; i < vector_active(cmd_vector); i++) {
@@ -1953,7 +2019,7 @@ static char **cmd_complete_command_real(vector vline, struct vty *vty,
 
 			/* First try completion match, if there is exactly match return 1 */
 			match =
-			    cmd_filter(command, cmd_vector, i, any_match);
+			    cmd_filter(command, cmd_vector, i, ANY_MATCH);
 
 			/* If there is exact match then filter ambiguous match else check
 			   ambiguousness. */
@@ -1992,9 +2058,18 @@ static char **cmd_complete_command_real(vector vline, struct vty *vty,
 				descvec = vector_slot(strvec, index);
 				for (j = 0; j < vector_active(descvec); j++)
 					if ((desc = vector_slot(descvec, j))) {
-						if ((string = cmd_entry_function(vector_slot(vline, index), desc->cmd)))
+						const char *cmd = desc->cmd;
+						char *tmp = NULL;
+
+						if (CMD_OPTION(desc->cmd)) {
+							tmp = cmd_deopt(tall_vty_cmd_ctx, desc->cmd);
+							cmd = tmp;
+						}
+						if ((string = cmd_entry_function(vector_slot(vline, index), cmd)))
 							if (cmd_unique_string (matchvec, string))
 								vector_set (matchvec, talloc_strdup(tall_vty_cmd_ctx, string));
+						if (tmp)
+							talloc_free(tmp);
 					}
 			}
 		}
@@ -2154,25 +2229,10 @@ int vty_go_parent(struct vty *vty)
 			vty_clear_parents(vty);
 			break;
 
-		case CFG_LOG_NODE:
-		case VTY_NODE:
-			vty->node = CONFIG_NODE;
-			vty_clear_parents(vty);
-			break;
-
 		default:
-			if (host.app_info->go_parent_cb) {
+			if (host.app_info->go_parent_cb)
 				host.app_info->go_parent_cb(vty);
-				vty_pop_parent(vty);
-			}
-			else if (is_config_child(vty)) {
-				vty->node = CONFIG_NODE;
-				vty_clear_parents(vty);
-			}
-			else {
-				vty->node = VIEW_NODE;
-				vty_clear_parents(vty);
-			}
+			vty_pop_parent(vty);
 			break;
 	}
 
@@ -2184,7 +2244,7 @@ static int
 cmd_execute_command_real(vector vline, struct vty *vty,
 			 struct cmd_element **cmd)
 {
-	unsigned int i;
+	unsigned int i, j;
 	unsigned int index;
 	vector cmd_vector;
 	struct cmd_element *cmd_element;
@@ -2195,18 +2255,22 @@ cmd_execute_command_real(vector vline, struct vty *vty,
 	enum match_type match = 0;
 	int varflag;
 	char *command;
+	int rc;
+	/* Used for temporary storage of cmd_deopt() allocated arguments during
+	   argv[] generation */
+	void *cmd_deopt_ctx = NULL;
 
 	/* Make copy of command elements. */
 	cmd_vector = vector_copy(cmd_node_vector(cmdvec, vty->node));
 
-	for (index = 0; index < vector_active(vline); index++)
+	for (index = 0; index < vector_active(vline); index++) {
 		if ((command = vector_slot(vline, index))) {
 			int ret;
 
 			match = cmd_filter(command, cmd_vector, index,
-			                   any_match);
+			                   ANY_MATCH);
 
-			if (match == vararg_match)
+			if (match == VARARG_MATCH)
 				break;
 
 			ret =
@@ -2220,15 +2284,16 @@ cmd_execute_command_real(vector vline, struct vty *vty,
 				return CMD_ERR_NO_MATCH;
 			}
 		}
+	}
 
 	/* Check matched count. */
 	matched_element = NULL;
 	matched_count = 0;
 	incomplete_count = 0;
 
-	for (i = 0; i < vector_active(cmd_vector); i++)
+	for (i = 0; i < vector_active(cmd_vector); i++) {
 		if ((cmd_element = vector_slot(cmd_vector, i))) {
-			if (match == vararg_match
+			if (match == VARARG_MATCH
 			    || index >= cmd_element->cmdsize) {
 				matched_element = cmd_element;
 #if 0
@@ -2239,6 +2304,7 @@ cmd_execute_command_real(vector vline, struct vty *vty,
 				incomplete_count++;
 			}
 		}
+	}
 
 	/* Finish of using cmd_vector. */
 	vector_free(cmd_vector);
@@ -2258,28 +2324,66 @@ cmd_execute_command_real(vector vline, struct vty *vty,
 	varflag = 0;
 	argc = 0;
 
+	cmd_deopt_ctx = talloc_named_const(tall_vty_cmd_ctx, 0, __func__);
+
 	for (i = 0; i < vector_active(vline); i++) {
-		if (varflag)
+		if (argc == CMD_ARGC_MAX) {
+			rc = CMD_ERR_EXEED_ARGC_MAX;
+			goto rc_free_deopt_ctx;
+		}
+		if (varflag) {
 			argv[argc++] = vector_slot(vline, i);
-		else {
-			vector descvec =
-			    vector_slot(matched_element->strvec, i);
-
-			if (vector_active(descvec) == 1) {
-				struct desc *desc = vector_slot(descvec, 0);
-
-				if (CMD_VARARG(desc->cmd))
-					varflag = 1;
-
-				if (varflag || CMD_VARIABLE(desc->cmd)
-				    || CMD_OPTION(desc->cmd))
-					argv[argc++] = vector_slot(vline, i);
-			} else
-				argv[argc++] = vector_slot(vline, i);
+			continue;
 		}
 
-		if (argc >= CMD_ARGC_MAX)
-			return CMD_ERR_EXEED_ARGC_MAX;
+		vector descvec = vector_slot(matched_element->strvec, i);
+		const char *tmp_cmd;
+
+		if (vector_active(descvec) == 1) {
+			/* Single coice argument, no "(...|...)". */
+			struct desc *desc = vector_slot(descvec, 0);
+
+			if (CMD_OPTION(desc->cmd)) {
+				/* we need to first remove the [] chars, then check to see what's inside (var or token) */
+				tmp_cmd = cmd_deopt(cmd_deopt_ctx, desc->cmd);
+			} else {
+				tmp_cmd = desc->cmd;
+			}
+
+			if (CMD_VARARG(tmp_cmd))
+				varflag = 1;
+			if (varflag || CMD_VARIABLE(tmp_cmd))
+				argv[argc++] = vector_slot(vline, i);
+			else if (CMD_OPTION(desc->cmd))
+				argv[argc++] = tmp_cmd;
+			/* else : we don't want to add non-opt single-choice static args in argv[] */
+		} else {
+			/* multi choice argument. look up which choice
+			   the user meant (can only be one after
+			   filtering and checking for ambigous). For instance,
+			   if user typed "th" for "(two|three)" arg, we
+			   want to pass "three" in argv[]. */
+			for (j = 0; j < vector_active(descvec); j++) {
+				struct desc *desc = vector_slot(descvec, j);
+				if (!desc)
+					continue;
+				if (cmd_match(desc->cmd, vector_slot(vline, i), ANY_MATCH, true) == NO_MATCH)
+					continue;
+				if (CMD_OPTION(desc->cmd)) {
+					/* we need to first remove the [] chars, then check to see what's inside (var or token) */
+					tmp_cmd = cmd_deopt(cmd_deopt_ctx, desc->cmd);
+				} else {
+					tmp_cmd = desc->cmd;
+				}
+
+				if(CMD_VARIABLE(tmp_cmd)) {
+					argv[argc++] = vector_slot(vline, i);
+				} else {
+					argv[argc++] = tmp_cmd;
+				}
+				break;
+			}
+		}
 	}
 
 	/* For vtysh execution. */
@@ -2287,10 +2391,35 @@ cmd_execute_command_real(vector vline, struct vty *vty,
 		*cmd = matched_element;
 
 	if (matched_element->daemon)
-		return CMD_SUCCESS_DAEMON;
+		rc = CMD_SUCCESS_DAEMON;
+	else {
+		/* Execute matched command. */
+		struct vty_parent_node this_node = {
+				.node = vty->node,
+				.priv = vty->priv,
+				.indent = vty->indent,
+			};
+		struct vty_parent_node *parent = vty_parent(vty);
+		rc = (*matched_element->func) (matched_element, vty, argc, argv);
 
-	/* Execute matched command. */
-	return (*matched_element->func) (matched_element, vty, argc, argv);
+		/* If we have stepped down into a child node, push a parent frame.
+		 * The causality is such: we don't expect every single node entry implementation to push
+		 * a parent node entry onto vty->parent_nodes. Instead we expect vty_go_parent() to *pop*
+		 * a parent node. Hence if the node changed without the parent node changing, we must
+		 * have stepped into a child node. */
+		if (vty->node != this_node.node && parent == vty_parent(vty)
+		    && vty->node > CONFIG_NODE) {
+			/* Push the parent node. */
+			parent = talloc_zero(vty, struct vty_parent_node);
+			*parent = this_node;
+			llist_add(&parent->entry, &vty->parent_nodes);
+		}
+	}
+
+rc_free_deopt_ctx:
+	/* Now after we called the command func, we can free temporary strings */
+	talloc_free(cmd_deopt_ctx);
+	return rc;
 }
 
 int
@@ -2351,10 +2480,10 @@ cmd_execute_command_strict(vector vline, struct vty *vty,
 			int ret;
 
 			match = cmd_filter(vector_slot(vline, index),
-			                   cmd_vector, index, exact_match);
+			                   cmd_vector, index, EXACT_MATCH);
 
 			/* If command meets '.VARARG' then finish matching. */
-			if (match == vararg_match)
+			if (match == VARARG_MATCH)
 				break;
 
 			ret =
@@ -2377,7 +2506,7 @@ cmd_execute_command_strict(vector vline, struct vty *vty,
 		if (vector_slot(cmd_vector, i) != NULL) {
 			cmd_element = vector_slot(cmd_vector, i);
 
-			if (match == vararg_match
+			if (match == VARARG_MATCH
 			    || index >= cmd_element->cmdsize) {
 				matched_element = cmd_element;
 				matched_count++;
@@ -2404,27 +2533,27 @@ cmd_execute_command_strict(vector vline, struct vty *vty,
 	argc = 0;
 
 	for (i = 0; i < vector_active(vline); i++) {
-		if (varflag)
+		if (argc == CMD_ARGC_MAX)
+			return CMD_ERR_EXEED_ARGC_MAX;
+		if (varflag) {
 			argv[argc++] = vector_slot(vline, i);
-		else {
-			vector descvec =
-			    vector_slot(matched_element->strvec, i);
-
-			if (vector_active(descvec) == 1) {
-				struct desc *desc = vector_slot(descvec, 0);
-
-				if (CMD_VARARG(desc->cmd))
-					varflag = 1;
-
-				if (varflag || CMD_VARIABLE(desc->cmd)
-				    || CMD_OPTION(desc->cmd))
-					argv[argc++] = vector_slot(vline, i);
-			} else
-				argv[argc++] = vector_slot(vline, i);
+			continue;
 		}
 
-		if (argc >= CMD_ARGC_MAX)
-			return CMD_ERR_EXEED_ARGC_MAX;
+		vector descvec = vector_slot(matched_element->strvec, i);
+
+		if (vector_active(descvec) == 1) {
+			struct desc *desc = vector_slot(descvec, 0);
+
+			if (CMD_VARARG(desc->cmd))
+				varflag = 1;
+
+			if (varflag || CMD_VARIABLE(desc->cmd)
+			    || CMD_OPTION(desc->cmd))
+				argv[argc++] = vector_slot(vline, i);
+		} else {
+			argv[argc++] = vector_slot(vline, i);
+		}
 	}
 
 	/* For vtysh execution. */
@@ -2550,8 +2679,7 @@ int config_from_file(struct vty *vty, FILE * fp)
 		ret = cmd_execute_command_strict(vline, vty, NULL);
 		cmd_free_strvec(vline);
 
-		if (ret != CMD_SUCCESS && ret != CMD_WARNING
-		    && ret != CMD_ERR_NOTHING_TODO) {
+		if (ret != CMD_SUCCESS && ret != CMD_ERR_NOTHING_TODO) {
 			if (indent) {
 				talloc_free(indent);
 				indent = NULL;
@@ -2581,6 +2709,10 @@ int config_from_file(struct vty *vty, FILE * fp)
 			indent = NULL;
 		}
 	}
+	/* Make sure we call go_parent_cb for all remaining indent levels at the end of file */
+	while (vty_parent(vty))
+		vty_go_parent(vty);
+
 	return CMD_SUCCESS;
 
 return_invalid_indent:
@@ -2694,7 +2826,7 @@ DEFUN(show_version,
 DEFUN(show_online_help,
       show_online_help_cmd, "show online-help", SHOW_STR "Online help\n")
 {
-	vty_dump_nodes(vty);
+	vty_dump_xml_ref_to_vty(vty);
 	return CMD_SUCCESS;
 }
 
@@ -2728,8 +2860,7 @@ gDEFUN(config_list, config_list_cmd, "list", "Print command list\n")
 
 	for (i = 0; i < vector_active(cnode->cmd_vector); i++)
 		if ((cmd = vector_slot(cnode->cmd_vector, i)) != NULL
-		    && !(cmd->attr == CMD_ATTR_DEPRECATED
-			 || cmd->attr == CMD_ATTR_HIDDEN))
+		    && !(cmd->attr & (CMD_ATTR_DEPRECATED | CMD_ATTR_HIDDEN)))
 			vty_out(vty, "  %s%s", cmd->string, VTY_NEWLINE);
 	return CMD_SUCCESS;
 }
@@ -2847,9 +2978,10 @@ static int write_config_file(const char *config_file, char **outpath)
 /* Write current configuration into file. */
 DEFUN(config_write_file,
       config_write_file_cmd,
-      "write file",
+      "write file [PATH]",
       "Write running configuration to memory, network, or terminal\n"
-      "Write to configuration file\n")
+      "Write to configuration file\n"
+      "Set file path to store the config, or replace if already exists\n")
 {
 	char *failed_file;
 	int rc;
@@ -2863,6 +2995,9 @@ DEFUN(config_write_file,
 		}
 	}
 
+	if (argc == 1)
+		host_config_set(argv[0]);
+
 	if (host.config == NULL) {
 		vty_out(vty, "Can't save to configuration file, using vtysh.%s",
 			VTY_NEWLINE);
