@@ -59,7 +59,7 @@ static void *tall_telnet_ctx;
 static int telnet_new_connection(struct osmo_fd *fd, unsigned int what);
 
 static struct osmo_fd server_socket = {
-	.when	    = BSC_FD_READ,
+	.when	    = OSMO_FD_READ,
 	.cb	    = telnet_new_connection,
 	.priv_nr    = 0,
 };
@@ -67,7 +67,7 @@ static struct osmo_fd server_socket = {
 /*! Initialize telnet based VTY interface listening to 127.0.0.1
  *  \param[in] tall_ctx \ref talloc context
  *  \param[in] priv private data to be passed to callback
- *  \param[in] port UDP port number
+ *  \param[in] port TCP port number to bind to
  */
 int telnet_init(void *tall_ctx, void *priv, int port)
 {
@@ -78,7 +78,7 @@ int telnet_init(void *tall_ctx, void *priv, int port)
  *  \param[in] tall_ctx \ref talloc context
  *  \param[in] priv private data to be passed to callback
  *  \param[in] ip IP to listen to ('::1' for localhost, '::0' for all, ...)
- *  \param[in] port UDP port number
+ *  \param[in] port TCP port number to bind to
  */
 int telnet_init_dynif(void *tall_ctx, void *priv, const char *ip, int port)
 {
@@ -101,16 +101,35 @@ int telnet_init_dynif(void *tall_ctx, void *priv, const char *ip, int port)
 		return -1;
 	}
 
-	LOGP(DLGLOBAL, LOGL_NOTICE, "telnet at %s %d\n", ip, port);
+	LOGP(DLGLOBAL, LOGL_NOTICE, "Available via telnet %s %d\n", ip, port);
 	return 0;
 }
 
+/*! Initializes telnet based VTY interface using the configured bind addr/port.
+ *  \param[in] tall_ctx \ref talloc context
+ *  \param[in] priv private data to be passed to callback
+ *  \param[in] default_port TCP port number to bind to if not explicitely configured
+ */
+int telnet_init_default(void *tall_ctx, void *priv, int default_port)
+{
+	return telnet_init_dynif(tall_ctx, priv, vty_get_bind_addr(),
+				 vty_get_bind_port(default_port));
+}
+
+
 extern struct host host;
 
 /*! close a telnet connection */
 int telnet_close_client(struct osmo_fd *fd)
 {
 	struct telnet_connection *conn = (struct telnet_connection*)fd->data;
+	char sock_name_buf[OSMO_SOCK_NAME_MAXLEN];
+	int rc;
+
+	/* FIXME: getsockname() always fails: "Bad file descriptor" */
+	rc = osmo_sock_get_name_buf(sock_name_buf, OSMO_SOCK_NAME_MAXLEN, fd->fd);
+	LOGP(DLGLOBAL, LOGL_INFO, "Closing telnet connection %s\n",
+	     (rc <= 0) ? "r=NULL<->l=NULL" : sock_name_buf);
 
 	close(fd->fd);
 	osmo_fd_unregister(fd);
@@ -130,8 +149,8 @@ static int client_data(struct osmo_fd *fd, unsigned int what)
 	struct telnet_connection *conn = fd->data;
 	int rc = 0;
 
-	if (what & BSC_FD_READ) {
-		conn->fd.when &= ~BSC_FD_READ;
+	if (what & OSMO_FD_READ) {
+		conn->fd.when &= ~OSMO_FD_READ;
 		rc = vty_read(conn->vty);
 	}
 
@@ -139,10 +158,10 @@ static int client_data(struct osmo_fd *fd, unsigned int what)
 	if (rc == -EBADF)
 		return rc;
 
-	if (what & BSC_FD_WRITE) {
+	if (what & OSMO_FD_WRITE) {
 		rc = buffer_flush_all(conn->vty->obuf, fd->fd);
 		if (rc == BUFFER_EMPTY)
-			conn->fd.when &= ~BSC_FD_WRITE;
+			conn->fd.when &= ~OSMO_FD_WRITE;
 	}
 
 	return rc;
@@ -154,18 +173,23 @@ static int telnet_new_connection(struct osmo_fd *fd, unsigned int what)
 	struct sockaddr_in sockaddr;
 	socklen_t len = sizeof(sockaddr);
 	int new_connection = accept(fd->fd, (struct sockaddr*)&sockaddr, &len);
+	char sock_name_buf[OSMO_SOCK_NAME_MAXLEN];
 	int rc;
 
 	if (new_connection < 0) {
-		LOGP(0, LOGL_ERROR, "telnet accept failed\n");
+		LOGP(DLGLOBAL, LOGL_ERROR, "telnet accept failed\n");
 		return new_connection;
 	}
 
+	rc = osmo_sock_get_name_buf(sock_name_buf, OSMO_SOCK_NAME_MAXLEN, new_connection);
+	LOGP(DLGLOBAL, LOGL_INFO, "Accept()ed new telnet connection %s\n",
+	     (rc <= 0) ? "r=NULL<->l=NULL" : sock_name_buf);
+
 	connection = talloc_zero(tall_telnet_ctx, struct telnet_connection);
 	connection->priv = fd->data;
 	connection->fd.data = connection;
 	connection->fd.fd = new_connection;
-	connection->fd.when = BSC_FD_READ;
+	connection->fd.when = OSMO_FD_READ;
 	connection->fd.cb = client_data;
 	rc = osmo_fd_register(&connection->fd);
 	if (rc < 0) {
@@ -176,8 +200,8 @@ static int telnet_new_connection(struct osmo_fd *fd, unsigned int what)
 
 	connection->vty = vty_create(new_connection, connection);
 	if (!connection->vty) {
-		LOGP(0, LOGL_ERROR, "couldn't create VTY\n");
-		close(new_connection);
+		LOGP(DLGLOBAL, LOGL_ERROR, "couldn't create VTY\n");
+		/* vty_create() is already closing the fd if it returns NULL */
 		talloc_free(connection);
 		return -1;
 	}
@@ -185,10 +209,20 @@ static int telnet_new_connection(struct osmo_fd *fd, unsigned int what)
 	return 0;
 }
 
+bool vty_is_active(struct vty *vty)
+{
+	struct telnet_connection *connection;
+	llist_for_each_entry(connection, &active_connections, entry) {
+		if (connection->vty == vty)
+			return true;
+	}
+	return false;
+}
+
 /*! callback from core VTY code about VTY related events */
 void vty_event(enum event event, int sock, struct vty *vty)
 {
-	struct vty_signal_data sig_data = { 0, };
+	struct vty_signal_data sig_data;
 	struct telnet_connection *connection = vty->priv;
 	struct osmo_fd *bfd;
 
@@ -207,10 +241,10 @@ void vty_event(enum event event, int sock, struct vty *vty)
 
 	switch (event) {
 	case VTY_READ:
-		bfd->when |= BSC_FD_READ;
+		bfd->when |= OSMO_FD_READ;
 		break;
 	case VTY_WRITE:
-		bfd->when |= BSC_FD_WRITE;
+		bfd->when |= OSMO_FD_WRITE;
 		break;
 	case VTY_CLOSED:
 		/* vty layer is about to free() vty */
